using System;
using System.Collections;
using System.Diagnostics;
using System.Reflection;
using AspectLib;
using AspectNet.AOP;
using AspectNet.Transformations;
using Rail.Reflect;

namespace AspectNet
{
	/// <summary>
	/// Summary description for AspectWeaver.
	/// </summary>
	public class AspectWeaver
	{
		private Hashtable	_transformations	= new Hashtable();
		private Hashtable	_aspectsCache		= new Hashtable();
		private SortedList	_aspectAssemblies	= new SortedList();

		private string		_outputFolder		= "..\\";
		private string		_assemblyName		= string.Empty;


		/// <summary>
		/// Default constructor to register the different transformations
		/// </summary>
		public AspectWeaver()
		{
			RegisterTransformation( new WeaveAfterTrans() );
			RegisterTransformation( new WeaveBeforeTrans() );
			RegisterTransformation( new WeaveAroundTrans() );
		}

		public void Reset()
		{
			_aspectsCache.Clear();
			_aspectAssemblies.Clear();
		}

		/// <summary>
		/// Returns the assemblies that have been processed
		/// (the ones that contained aspects)
		/// </summary>
		public ICollection AdjustedAssemblies
		{
			get { return _aspectAssemblies.Values; }
		}

		/// <summary>
		/// Sets the folder which will be used to copy the woven files to
		/// </summary>
		public string OutputFolder
		{
			get { return _outputFolder; }
			set { _outputFolder = value; }
		}

		/// <summary>
		/// Registers a transformation within the weaver
		/// </summary>
		public void RegisterTransformation(ITransformation trans)
		{
			if ( !_transformations.ContainsKey( trans.Binding ) )
			{
				_transformations.Add( trans.Binding, trans );
			}
			else
			{
				throw new ApplicationException(
					string.Format( "Transformation (tag: {0}) already registered!", trans.Binding ) );
			}
		}

		/// <summary>
		/// This method loads all aspects located in the given assembly
		/// </summary>
		public void LoadAspects(string assemblyName)
		{
			_assemblyName = assemblyName;
			
			// make sure that only the name of the assembly 
			// (without path) is passed
			if ( assemblyName.IndexOf("\\") != -1 )
			{
				assemblyName = assemblyName.Substring( assemblyName.LastIndexOf("\\") + 1 )	;
			}

			// if we have processed an assembly with the same name (i.e. the same assembly)
			// then it's time to go :)
			if ( !_aspectAssemblies.ContainsKey( assemblyName ) )
			{
				_aspectAssemblies.Add( assemblyName, _outputFolder + "\\" + assemblyName );
			}
			else
			{
				return;
			}

			// load the assembly
			RAssemblyDef asm = RAssemblyDef.LoadAssembly( _assemblyName );

			// get all define types
			RType [] allTypes = asm.RModuleDef.GetTypes();

			ArrayList types = new ArrayList();

			// for each type
			foreach ( RTypeDef currType in allTypes )
			{
				// get the aspec attributes
				RCustomAttrib [] attributes = 
					currType.GetCustomAttributes( typeof(AspectAttribute).FullName, false );

				// if this is an aspect class and we don't have it in our cache
				// we have to store it
				if ( attributes != null		&& 
					 attributes.Length > 0  &&
					 !_aspectsCache.ContainsKey( currType.Name ) )
				{
					// adjust the aspect type (replace all calls to ControlFlow.Resume
					// with the respective dynamic invoke of the delegate in the context)
					RailHelper.AdjustAspect( asm, currType );

					types.Add( currType );

					//_aspectsCache[currType.Name] = new Aspect( asm, asm.RModuleDef, currType );
				}
			}

			asm.SaveAssembly( assemblyName, _outputFolder );

			if ( types.Count > 0 )
			{
				asm = RAssemblyDef.LoadAssembly( _outputFolder + "\\" + assemblyName );

				foreach( RTypeDef rtd in types )
				{
					RTypeDef rt = (RTypeDef) asm.RModuleDef.GetType( rtd.NameWithNameSpace );
					_aspectsCache[rt.Name] = new Aspect( asm, asm.RModuleDef, rt );
				}
				
			}
#if DEBUG
//			RailHelper.ValidateAssembly( _outputFolder + "\\" + assemblyName );
#endif
		}

		/// <summary>
		/// This method weaves the base code, represented here with the
		/// point cuts collection, with the aspect code
		/// </summary>
		public void ProcessJoinPoints(ArrayList pointCuts)
		{
			if ( _aspectsCache.Count == 0 )
			{
				throw new ApplicationException("There are no loaded aspects!");
			}

			bool bSuccess = true;

			// this sorted list is used to store the assemblies which contain
			// the join points that will be transformed. the list is needed
			// because later on we'll need to save those assemblies
			SortedList assemblies = new SortedList( pointCuts.Count );

			// traverse all the join points
			foreach ( PointCut pcut in pointCuts )
			{
				RAssemblyDef pAsm = pcut.ParentAssembly;
				
				// check if we have this assembly already stored
				if ( !assemblies.ContainsKey( pAsm.Name ) )
				{
					assemblies.Add( pAsm.Name, pAsm );
				}

				foreach ( JoinPoint point in pcut )
				{
					// get the transformation for the respective joinpoint
					ITransformation transformation = 
						_transformations[point.PointCut.Binding] as ITransformation;

					// get the advice and transform the join point
					if ( false == transformation.TransformJoinPoint( point, GetAdvice(point) ) )
					{
						bSuccess = false;

						Trace.WriteLine( string.Format( "Failed transforming {0} with {1}", 
														point.ToString(), 
														transformation.ToString() ) );
					}
				}

				// if everything went smoothly we should save the changes
				if ( bSuccess )
				{
					IDictionaryEnumerator e = assemblies.GetEnumerator();

					// so we iterate over all assemblies
					while ( e.MoveNext() )
					{
						// we get the short name of the assembly (no path included)
						RAssemblyDef asm = (RAssemblyDef) e.Value;
						string asmName = asm.FileName;
						
						if ( asmName.IndexOf("\\") != -1 )
						{
							asmName = asmName.Substring( asmName.LastIndexOf("\\") + 1 );
						}

						// and voila - assembly saved :)
						asm.SaveAssembly( asmName, _outputFolder );
					}
				}
			}
		}

		/// <summary>
		/// This method retrieves the advice required by the respective
		/// join point
		/// </summary>
		private Advice GetAdvice(JoinPoint point)
		{
			Advice result = null;

			if ( _aspectsCache.ContainsKey( point.PointCut.AspectName ) )
			{
				result = ((Aspect)_aspectsCache[point.PointCut.AspectName])[point.PointCut.AdviceName];
			}

			return result;
		}
	}
}
