using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Reflection.Emit;
using AspectLib;
using Rail.MSIL;
using Rail.Reflect;

namespace AspectNet.Transformations
{
	/// <summary>
	/// Summary description for RailHelper.
	/// </summary>
	public class RailHelper
	{
		static RAssemblyDef libAssembly = null;

		public static string AspectLibPath
		{
			set
			{
				libAssembly = RAssemblyDef.LoadAssembly( value );	
			}			
		}


		// modification nested class
		class Modification
		{
			public enum Operation
			{
				None,
				Insert,
				Delete
			}

			public Operation operation		= Operation.None;
			public Instruction instruction	= null;
			public int position				= 0;

			public Modification(int pos, Instruction instr, Operation op)
			{
				position = pos;
				instruction = instr;
				operation = op;
			}

			public Modification(int pos, Instruction instr)
			{
				position = pos;
				instruction = instr;
				operation = Operation.Insert;
			}
		}


		#region private helper methods
		/// <summary>
		/// Helper method (2 overloads) - returns a method from a type, given its name
		/// </summary>
		static RMethodDef GetMethodByName(RTypeDef type, string name)
		{
			RMethodDef method = null;

			if ( type != null )
			{
				foreach(RMethodDef m in type.GetMethods())
				{
					if ( m.Name == name )
					{
						method = m;
						break;
					}
				}				
			}

			return method;
		}

		static RMethodRef GetMethodByName(RTypeRef type, string name)
		{
			RMethodRef method = null;

			if ( type != null )
			{
				foreach(RMethodRef m in type.GetMethods())
				{
					if ( m.Name == name )
					{
						method = m;
						break;
					}
				}
			}

			return method;
		}

		
		static string GetBindFieldName( RMethod point, string advName )
		{
			return "_" + point.Name + advName;
		}

		static string GetAspectInstanceFieldName( RType aspect )
		{
			return aspect.Name + "_Instance";
		}

		static string GetDefImplName( RMethod point )
		{
			return point.Name + "DefImpl";
		}

		static string GetDelegateName( RMethod point, bool fullName )
		{
			string str = point.Name + "Delegate";
			
			if ( fullName )
			{
				str = point.DeclaringType.Assembly.Name.Name + "." + point.Name + "Delegate";	
			}
			
			return str;
		}

		static string GetEventName( RMethod point )
		{
			return point.Name + "Event";
		}

		static string GetDefHandlerName( RMethod point  )
		{
			return point.Name + "DefHandler";
		}

		static string GetAspectHandlerName( RMethod point, RType aspect, RMethod advice  )
		{
			return point.Name + advice.Name + aspect.Name + "Handler";
		}

		static string GetInitAspectsMethodName()
		{
			return "InitAspects";
		}

		#endregion

		#region RailHelper methods

		public static RMethodDef CopyMethod( RTypeDef declType, 
												RMethodDef method )
		{
			string defImplName = GetDefImplName( method );

			RMethodDef newMethod = GetMethodByName( declType, defImplName );
			
			if ( newMethod == null )
			{
				// copy the method definition
				newMethod = declType.CopyMethod( method, defImplName );
				
				// adjust the attributes
				newMethod.Attributes |= MethodAttributes.Private;

				if ( ( method.Attributes & MethodAttributes.Public ) != 0 )
				{
					newMethod.Attributes ^= MethodAttributes.Public;	
				}

				// define new body
				MethodBody mb = newMethod.DefineMethodBody();

				// this hash contains the mapping between old and new instructions
				Hashtable replacementTargets = new Hashtable();
				
				// 1st copy phase: copy each instruction (by creating clones)
				foreach ( Instruction instr in method.MethodBody.Code )
				{
					instr.CopyTo( method, newMethod, replacementTargets );
				}

				// 2nd copy phase: adjust targets
				foreach ( Instruction instr in newMethod.MethodBody.Code )
				{
					if ( instr is ILBranch ) 
					{
						Instruction oldTarget = ((ILBranch)instr).Target;
						
						((ILBranch)instr).Target = (Instruction)replacementTargets[oldTarget];
					} 
					else if ( instr is ILSwitch )
					{
						ILSwitch sw = instr as ILSwitch;
						
						Instruction[] switchTargets = new Instruction[sw.Targets.Length];

						for ( int i = 0; i < switchTargets.Length; i++ ) 
						{
							switchTargets[i] = (Instruction)replacementTargets[sw.Targets[i]];
						}

						sw.Targets = switchTargets;
					}
				}

				// adjust the indices
				newMethod.MethodBody.Code.UpdateInstructionIndex();
			}

			return newMethod;
		}

		/// <summary>
		/// This method defines a local constant which defines the binding
		/// of the aspect to the joinpoint. To accomplish this, the name of the
		/// variable is generated: _ + pointname + advicename + aspectname
		/// </summary>
		public static bool DefineVariables( RAssemblyDef declAssembly,
											RTypeDef type, 
											RMethodDef method, 
											string advName, 
											ref RFieldDef aspectVar,
											RType aspectType )
		{
			bool bAlreadyBoundOnce = true;

			string aspVarName = GetAspectInstanceFieldName( aspectType );

			aspectVar = (RFieldDef)type.GetField( aspVarName );

			if (  aspectVar == null )
			{
				aspectVar = type.DefineField( aspVarName, aspectType );
				aspectVar.Attributes = FieldAttributes.Private | FieldAttributes.Static;
				bAlreadyBoundOnce = false;
			}

			return bAlreadyBoundOnce;
		}

		/// <summary>
		/// This method adds a reference to the AspectLib assembly and
		/// particularly a TypeRef to the MethodContext
		/// </summary>
		public static void AdjustAssemblyReferences( RAssemblyDef declAssembly, RType aspectType )
		{
			// add a reference to the AspectLib assembly
			if ( null == declAssembly.GetReferencedAssembly( libAssembly.Name.FullName ) )
			{
				declAssembly.AddReferencedAssembly( libAssembly.Name.FullName );

				FileStream f = File.OpenRead(libAssembly.FileName);
				byte[] b = new byte[f.Length];
				f.Read(b,0,(int)f.Length);
				f.Close();

				Assembly a = Assembly.Load( b );

				Type t = a.GetType( typeof(MethodContext).FullName );
				declAssembly.DefineTypeRef( t );
			}

			// check and add a reference to the assembly containing the aspect
			if ( declAssembly.Name.FullName != aspectType.Assembly.Name.FullName &&
				 null == declAssembly.GetReferencedAssembly( aspectType.Assembly.Name.FullName ) )
			{
				declAssembly.AddReferencedAssembly( aspectType.Assembly.Name.FullName );
				
				FileStream f = File.OpenRead(aspectType.Assembly.FileName);
				byte[] b = new byte[f.Length];
				f.Read(b,0,(int)f.Length);
				f.Close();

				Assembly a = Assembly.Load( b );
				Type t = a.GetType( aspectType.NameWithNameSpace );
				declAssembly.DefineTypeRef( t );
			}
		}

		
		/// <summary>
		/// This method will adjust all current constructors to first call the
		/// static initializing method for the aspects, before anything else
		/// </summary>
		public static void AdjustConstructors( RAssemblyDef ass, 
												RTypeDef type, 
												RMethodDef initMethod )
		{
			// iterate over the different constructors	
			foreach( RConstructorDef ctor in type.GetConstructors())
			{
				Code code = ctor.MethodBody.Code;
				
				// the index of the last call to the base class'es constructor
				int lastIdx = 0;

				// now iterate over all instructions and find the last call to the
				// base constructors
				for(int i = 0; i < code.InstructionCount; i++)
				{
					Instruction instr = code[i];
					
					// if this is a CALL instruction and the instruction is a method
					if ( instr.OpCode.Value == OpCodes.Call.Value && instr is ILMethod )
					{
						ILMethod m = instr as ILMethod;
						
						// and if the method is actually constructor to our parent class'es
						// constructor, then we store the index
						if ( m.Method is RConstructor &&
							(m.Method as RConstructor).DeclaringType == type.BaseType)
						{
							lastIdx = i+1;
						}
					}
					
				}

				// right after the last call to the base constructor
				// (i.e. the very first instruction in the constructor's body)
				// insert a call to the private initializing method
				code.Insert(lastIdx, ILFactory.Call(false, initMethod));
			}
		}


		/// <summary>
		/// This method adjusts the original method so that:
		///		1. creates and initializes the MethodContext
		///		2. raises the method interceptor's event
		/// </summary>
		public static void AdjustOriginalMethod(RAssemblyDef declAssembly,
												RMethodDef method, 
												RTypeDef type,
												RTypeDef delegateType, 
												REventDef newEvent,
												RMethodDef defHandler,
												string adviceName )
		{
			// adjust the max stack property
			method.MethodBody.MaxStack += 1;

			Code code = method.MethodBody.Code;
			
			// clear the existing implementation
			code.Clear();

			// get all parameters
			RParameter[] parameters = method.GetParameters();
			
			// get the method context class
			RType ctxType = declAssembly.GetType(typeof(MethodContext).FullName);

			
			// **********************************************************************************
			// Step 1 - Create and fill the context structure


			// create two new local variables
			// object[] that will contain the delegate parameters (needed for dynamic call)
			LocalVariable l1 = method.MethodBody.DefineLocalVariable( libAssembly.GetType(typeof(object[]).FullName) );
			
			// the context class
			LocalVariable l2 = method.MethodBody.DefineLocalVariable( ctxType );


			// now create and assign the context structure
			code.Add( ILFactory.Newobj( ctxType.GetConstructors()[0] ) );
			code.Add( ILFactory.Stloc( l2 ) );

			// create the array and store it in the local variable
			code.Add( ILFactory.Ldc( parameters.Length ) );
			code.Add( ILFactory.Newarr( libAssembly.GetType(typeof(object).FullName) ) );
			code.Add( ILFactory.Stloc( l1 ) );

			int i = 0;
			
			// now fill in all parameters in the array
			foreach (RParameter par in parameters)
			{
				// load the local variable
				code.Add( ILFactory.Ldloc( l1 ) );
				
				// load the position in the array
				code.Add( ILFactory.Ldc( i++ ) );
				
				// load the respective param
				code.Add( ILFactory.Ldarg( par ) );
				
				// if needed we should box the parameter
				if ( par.ParameterType.IsValueType )
				{
					code.Add( ILFactory.Box( par.ParameterType ) );
				}

				// store the parameter in the array at the right position
				code.Add( ILFactory.Stelem( typeof(object) ) );
			}
		
			// first assign the local array to the respective field in the context
			code.Add( ILFactory.Ldloc( l2 ) );
			code.Add( ILFactory.Ldloc( l1 ) );
			code.Add( ILFactory.Stfld( ctxType.GetField("ctxParameters") ) );

			// now create the callback delegate 
			code.Add( ILFactory.Ldloc( l2 ) );
			
			// if we're static just load null, otherwise the this
			if ( (method.Attributes & MethodAttributes.Static) == 0 )
				code.Add( ILFactory.Ldarg( new RParameter( 0, null ) ) );
			else
				code.Add( ILFactory.Ldnull() );

			// get a pointer to the default handler
			code.Add( ILFactory.Ldftn( defHandler ) );
			// create the delegate
			code.Add( ILFactory.Newobj( delegateType.GetConstructors()[0] ) );
			// store the delegate in the respective field in the context
			code.Add( ILFactory.Stfld( ctxType.GetField("ctxOriginalMethod") ) );

			// now store the this pointer in the context
			code.Add( ILFactory.Ldloc( l2 ) );
			code.Add( ILFactory.Ldarg( new RParameter( 0, null ) ) );
			code.Add( ILFactory.Stfld( ctxType.GetField("ctxObject") ) );


			// **********************************************************************************
			// Step 2 - Raise the event
			
			// load the event field
			code.Add( ILFactory.Ldsfld( type.GetField(newEvent.Name) ) );

			// load the last parameter - the context stucture
			code.Add( ILFactory.Ldloc( l2 ) );

			// call the invoke method of the event
			code.Add( ILFactory.Call(true, GetMethodByName( delegateType, "Invoke" ) ) );
			

			// if the original method is Void, then we should remove the 
			// result (NULL) from the stack
			if ( method.ReturnType.ReturnType.NameWithNameSpace ==
				libAssembly.GetType(typeof(void).FullName).NameWithNameSpace )
			{
				code.Add( ILFactory.Pop() );
			}
			else if ( method.ReturnType.ReturnType.IsValueType )
			{
				code.Add( ILFactory.Unbox( method.ReturnType.ReturnType ) );
				code.Add( ILFactory.Ldind( Type.GetType(method.ReturnType.ReturnType.FullName) ) );				
			}

			// return result
			code.Add( ILFactory.Ret() );
		}

		
		/// <summary>
		/// This method inspects the static aspect initializer method
		/// and adds code for binding the defHandler param to the newEvent
		/// event
		/// </summary>
		public static void BindAspectHandler( RMethodDef initMethod, 
												REventDef newEvent, 
												RMethodDef adviceHandler,
												RType aspectType,
												RTypeDef delegateType,
												RField newAspectVar, 
												bool bAspectDefined )
		{	
			Code code = initMethod.MethodBody.Code;
			
			// remove the Ret instruction
			code.Remove( code.InstructionCount - 1 );

			// bind the handler to the event
			if ( (adviceHandler.Attributes & MethodAttributes.Static) != 0 )
			{
				code.Add( ILFactory.Ldnull() );	
			}
			
			code.Add( ILFactory.Ldftn( adviceHandler ) );
			code.Add( ILFactory.Newobj( delegateType.GetConstructors()[0] ) );
			code.Add( ILFactory.Call( false, newEvent.AddOnMethod ) );
		
			if ( !bAspectDefined )
			{
				code.Add( ILFactory.Newobj( aspectType.GetConstructors()[0] ) );
				code.Add( ILFactory.Stsfld( newAspectVar ) );
			}

			// restore the Ret instruction
			code.Add( ILFactory.Ret() );
		}
		
		/// <summary>
		/// This method checks if a method with the given name already exists
		/// and if it does - returns it. If not - creates an empty method.
		/// </summary>
		public static RMethodDef CreateAspectInitMethod( RAssemblyDef declAssembly, 
															RTypeDef type,
															string methodName,
															ref bool created)
		{
			// get the method
			RMethodDef method = GetMethodByName(type, methodName);
			
			// if it's not present
			if ( method == null )
			{
				created = true;

				// create the method
				method = type.DefineMethod( methodName, 
											libAssembly.GetType(typeof(void).FullName),
											new RParameter[]{},
											MethodAttributes.HideBySig |
											MethodAttributes.Static |
											MethodAttributes.Private);

				// define body and maxstack
				method.DefineMethodBody();
				method.MethodBody.MaxStack = 4;
				method.MethodBody.Code.Add( ILFactory.Ret() );
			}
			
			return method;
		}

		
		/// <summary>
		/// This method creates an event (private and static) compliant 
		/// with the delegate type passed.
		/// </summary>
		public static REventDef CreateEvent( RAssemblyDef declAssembly, 
												RTypeDef type, 
												RTypeDef delegateType,
												RMethodDef method )
		{
			REventDef eventDef = (REventDef) type.GetEvent( GetEventName( method ) );

			if ( eventDef == null )
			{
				// create the event itself
				eventDef = type.DefineEvent( GetEventName( method ), delegateType );				

				// create a private static field of the delegate type
				RFieldDef field = type.DefineField( GetEventName( method ), delegateType );
				field.Attributes = FieldAttributes.Private | FieldAttributes.Static;

				// create and assign the Add and Remove methods of the event
				eventDef.AddOnMethod = CreateAddMethod(declAssembly, type, delegateType, method, field);
				eventDef.RemoveOnMethod = CreateRemoveMethod(declAssembly, type, delegateType, method, field);
			}
			
			return eventDef;
		}

		
		/// <summary>
		/// This method creates the Add method for an event
		/// </summary>
		private static RMethodDef CreateAddMethod( RAssemblyDef declAssembly,
													RTypeDef type,
													RTypeDef delegateType,
													RMethodDef method,
													RFieldDef field)
		{
			// first define the method itself
			RMethodDef addMethod = type.DefineMethod( "add_" + method.Name + "Event",
														libAssembly.GetType(typeof(void).FullName),
														new RParameter[]{ new RParameter(0, delegateType )},
														MethodAttributes.HideBySig |
														MethodAttributes.SpecialName |
														MethodAttributes.Static |
														MethodAttributes.Private);

			addMethod.MethodImplementationFlags = MethodImplAttributes.Synchronized | 
				MethodImplAttributes.Managed;

			// get the combine method
			RMethodRef combineMethod = 
				GetMethodByName( libAssembly.GetType( typeof(Delegate).FullName ), 
								"Combine" );
			
			// define the body and the maxstack
			Code code = addMethod.DefineMethodBody().Code;
			addMethod.MethodBody.MaxStack = 2;

			// load arguments for the combine (the delegate field and the this pointer)
			code.Add( ILFactory.Ldsfld(field) );
			code.Add( ILFactory.Ldarg(new RParameter(0, null)) );
			
			// call the combine method
			code.Add( ILFactory.Call(false, combineMethod) );
			
			// cast back the result and store the static field back
			code.Add( ILFactory.Castclass(delegateType) );
			code.Add( ILFactory.Stsfld( field ) );
			
			code.Add( ILFactory.Ret() );

			return addMethod;
		}

		
		/// <summary>
		/// This method creates the Remove method for an event
		/// </summary>
		private static RMethodDef CreateRemoveMethod( RAssemblyDef declAssembly,
														RTypeDef type,
														RTypeDef delegateType,
														RMethodDef method,
														RFieldDef field)
		{
			// define the method itself
			RMethodDef removeMethod = type.DefineMethod( "remove_" + method.Name + "Event",
															libAssembly.GetType(typeof(void).FullName),
															new RParameter[]{ new RParameter(0, delegateType )},
															MethodAttributes.HideBySig |
															MethodAttributes.SpecialName |
															MethodAttributes.Static |
															MethodAttributes.Private);

			removeMethod.MethodImplementationFlags = MethodImplAttributes.Synchronized | 
				MethodImplAttributes.Managed;

			// get the remove method of the System.Delegate
			RMethodRef removeDelegateMethod = 
				GetMethodByName( libAssembly.GetType( typeof(Delegate).FullName ), 
								"Remove" );
				
			// define the method body and the maxstack
			Code code = removeMethod.DefineMethodBody().Code;
			removeMethod.MethodBody.MaxStack = 2;

			// load the arguments for the Remove method
			code.Add( ILFactory.Ldsfld(field) );
			code.Add( ILFactory.Ldarg(new RParameter(0, null)) );
			
			// call the Remove method
			code.Add( ILFactory.Call(false, removeDelegateMethod) );
			
			// cast back the result and store the static field back
			code.Add( ILFactory.Castclass(delegateType) );
			code.Add( ILFactory.Stsfld( field ) );
			
			code.Add( ILFactory.Ret() );

			return removeMethod;
		}

		/// <summary>
		/// This method sets the binding variable to the context parameter and
		/// invokes the aspect's advice
		/// </summary>
		public static RMethodDef CreateAspectHandler( RAssemblyDef decAssembly, 
														RTypeDef type, 
														RType aspType,
														RMethodDef defHandlerMethod,
														RMethod advice,
														RMethodDef point,
														RFieldDef aspectInstance,
														
														int bindValue )
		{

			// copy the method definition
			RMethodDef newMethod = type.CopyMethod( defHandlerMethod, GetAspectHandlerName( point, aspType, advice ) );
			
			RType ctxType = decAssembly.GetType( typeof(MethodContext).FullName );			
			
			// define new body
			MethodBody mb = newMethod.DefineMethodBody();
			Code code = mb.Code;	

			// define local variable of type context
			LocalVariable varContext = mb.DefineLocalVariable( ctxType );

			RParameter p = newMethod.GetParameters()[0];
			
			// cast and assign the last parameter of the method to the context local var
			code.Add( ILFactory.Ldarg( p ) );
			code.Add( ILFactory.Castclass( ctxType ) );
			code.Add( ILFactory.Stloc( varContext ) );
			
			// now set the binding flag
			code.Add( ILFactory.Ldloc( varContext ) );
			code.Add( ILFactory.Ldc( bindValue ) );
			code.Add( ILFactory.Stfld( ctxType.GetField("ctxBindOption") ) );

			// load the aspect instance, which method we'll call
			code.Add( ILFactory.Ldsfld( aspectInstance ) );

			// put parameters on stack
			foreach ( RParameter param in newMethod.GetParameters() )
			{
				code.Add( ILFactory.Ldarg( param ) );
			}
			
			code.Add( ILFactory.Call( true, advice ) );

			// maintain result
			code.Add( ILFactory.Ret() );

			return newMethod;
		}

		/// <summary>
		/// This method creates a default proxy method for the method 
		/// given as a parameter
		/// </summary>
		public static RMethodDef CreateDefaultHandler( RAssemblyDef decAssembly, 
														RTypeDef type, 
														RMethodDef method )
		{
			RMethodDef defaultMethod = GetMethodByName( type, GetDefHandlerName(method) );
			
			if ( defaultMethod != null )
			{
				return defaultMethod;
			}

			// create a parameters array for all method's parameters + 1 new (the context)
			RParameter [] delegateParams = new RParameter[1];
			
			// fill in the above array
			// fill in the last parameter (type of Object, named context)
			delegateParams[0] = new RParameter( 1, libAssembly.GetType(typeof(object).FullName) );
			delegateParams[0].Name = "context";

			// now create the Invoke method of the new delegate type
			defaultMethod = type.DefineMethod(	GetDefHandlerName(method), 
																	libAssembly.GetType(typeof(object).FullName),
																	delegateParams,
																	method.Attributes		| 
																	MethodAttributes.Static | 
																	MethodAttributes.Private);

			defaultMethod.MethodImplementationFlags = MethodImplAttributes.Managed;
			
			// define the body and the max stack
			defaultMethod.DefineMethodBody();
			defaultMethod.MethodBody.MaxStack = delegateParams.Length - 1;
			
			Code code = defaultMethod.MethodBody.Code;	
			
			RType ctxType = decAssembly.GetType( typeof(MethodContext).FullName );
	
			// define local variable of type context
			LocalVariable varContext = defaultMethod.MethodBody.DefineLocalVariable( ctxType );

			LocalVariable varResult = defaultMethod.MethodBody.DefineLocalVariable( libAssembly.GetType(typeof(object).FullName) );

			Instruction retInstr = ILFactory.Ldloc( varResult );

			// store the context in the local variable
			code.Add( ILFactory.Ldarg( delegateParams[0] ) );
			code.Add( ILFactory.Castclass( ctxType ) );
			code.Add( ILFactory.Stloc( varContext ) );

			code.Add(ILFactory.Ldloc( varContext ) );
			code.Add(ILFactory.Ldfld( ctxType.GetField( "ctxOriginalCalled" ) ) );
			code.Add(ILFactory.Ldc( 0 ) );
			code.Add(ILFactory.Bne_Un(retInstr) );

			// if the method is not static we should load the this pointer,
			// which is stored in the last parameter of the delegate
			if ( (method.Attributes & MethodAttributes.Static ) == 0 )
			{
				code.Add( ILFactory.Ldloc( varContext ) );
				code.Add( ILFactory.Ldfld( ctxType.GetField( "ctxObject" ) ) ) ;
				code.Add( ILFactory.Castclass( type ) );
			}

			// load all arguments for the following call to the default implementation
			for ( int i = 0; i < method.ParametersCount; ++i )
			{
				RParameter param = 
					(method.Attributes & MethodAttributes.Static ) == 0 ?
					method.GetParameter(i + 1) :
					method.GetParameter(i);

				code.Add( ILFactory.Ldloc( varContext ) );
				code.Add( ILFactory.Ldfld( ctxType.GetField( "ctxParameters" ) ) );
				code.Add( ILFactory.Ldc( i ) );
				code.Add( ILFactory.Ldelem( typeof(object) ) );
				
				if ( param.ParameterType.IsValueType )
				{
					code.Add( ILFactory.Unbox( param.ParameterType ) );
					code.Add( ILFactory.Ldind( Type.GetType(param.ParameterType.FullName) ) );
				}
				else
				{
					code.Add( ILFactory.Castclass( param.ParameterType ) );	
				}
			}

			// the call itself
			code.Add( ILFactory.Call( true, method ) );

			//if the return value is value type, we should box it
			if ( method.ReturnType.ReturnType.NameWithNameSpace !=
					libAssembly.GetType(typeof(void).FullName).NameWithNameSpace )
			{
				if ( method.ReturnType.ReturnType.IsValueType )
				{
					code.Add(ILFactory.Box( method.ReturnType.ReturnType ));
				}
	
				code.Add( ILFactory.Stloc( varResult ) );
			}
			else
			{
				code.Add( ILFactory.Ldnull() );
				code.Add( ILFactory.Stloc( varResult ) );
			}
			
			code.Add( retInstr );

			// maintain result
			code.Add( ILFactory.Ret() );				

			code.UpdateInstructionIndex();

			return defaultMethod;
		}

		
		/// <summary>
		/// This method creates a delegate type which has the same signature
		/// as the given method + an additional parameter of type Object (the context)
		/// </summary>
		public static RTypeDef CreateDelegateType(  RAssemblyDef decAssembly,			
													RModuleDef declModule, 
													RTypeDef declType,
													RMethodDef method )
		{
			RTypeDef delegateType = (RTypeDef) declModule.GetType( GetDelegateName(method, true) );
			
			if ( delegateType == null )
			{
				// define the type itself
				delegateType = declModule.DefineType( GetDelegateName(method, false), declType.NameSpace );
				delegateType.BaseType = libAssembly.GetType(typeof(MulticastDelegate).FullName);
				delegateType.Attributes |=  TypeAttributes.Class		| 
											TypeAttributes.Public		| 
											TypeAttributes.Sealed		| 
											TypeAttributes.AnsiClass	| 
											TypeAttributes.AutoClass;

			
				// create and initialize the constructor
				RConstructorDef ctor = 
					delegateType.DefineConstructor( new RParameter[]{ new RParameter(1, libAssembly.GetType(typeof(object).FullName) ),
																		new RParameter(2, libAssembly.GetType(typeof(IntPtr).FullName) ), 
																	} );
			
				ctor.Attributes = MethodAttributes.RTSpecialName | 
									MethodAttributes.HideBySig | 
									MethodAttributes.Public;
			
				ctor.CallingConventions = CallingConventions.Standard;
				ctor.MethodImplementationFlags = MethodImplAttributes.Runtime | MethodImplAttributes.Managed;

				RParameter [] delegateParams = new RParameter[1];
			
				delegateParams[0] = new RParameter( 1, libAssembly.GetType(typeof(object).FullName) );

				// now create the Invoke method of the new delegate type
				RMethodDef invokeMethod = delegateType.DefineMethod(	"Invoke", 
																		libAssembly.GetType(typeof(object).FullName),
																		//method.ReturnType.ReturnType,
																		delegateParams,
																		MethodAttributes.Public | 
																		MethodAttributes.HideBySig | 
																		MethodAttributes.NewSlot | 
																		MethodAttributes.Virtual );
																
				invokeMethod.MethodImplementationFlags = MethodImplAttributes.Runtime | 
					MethodImplAttributes.Managed;
				
			}

			return delegateType;
		}

		#endregion
		
		/// <summary>
		/// This method adjusts all advices of an aspect
		/// </summary>
		public static void AdjustAspect( RAssemblyDef declAssembly, RTypeDef aspect )
		{
			// loop over all methods
			foreach ( RMethodDef advice in aspect.GetMethods() )
			{
				// get the custom attributes of type AdviceAttribute
				RCustomAttrib[] adviceAttributes = 
					advice.GetCustomAttributes( typeof( AdviceAttribute ).FullName, false );

				// check if this method is an advice (has AdviceAttribute defined)
				if ( adviceAttributes != null && adviceAttributes.Length > 0 )
				{
					Code code = advice.MethodBody.Code;
					
					ArrayList newInstructions = new ArrayList();

					RType ctxType = declAssembly.GetType( typeof(MethodContext).FullName );

					// define local variable of type context
					LocalVariable varContext = advice.MethodBody.DefineLocalVariable( ctxType );
					
					// local variable for the method's return value
					LocalVariable varOrigResult = 
						advice.MethodBody.DefineLocalVariable( libAssembly.GetType(typeof(object).FullName ) );

					LocalVariable varArr = advice.MethodBody.DefineLocalVariable( libAssembly.GetType(typeof(object[]).FullName) );
			
					RParameter p = advice.GetParameters()[0];
					// cast and assign the last parameter of the method to the context local var
					newInstructions.Add(new Modification( 0, ILFactory.Ldarg( p ) ) );
					newInstructions.Add(new Modification( 1, ILFactory.Castclass( ctxType ) ) );
					newInstructions.Add(new Modification( 2, ILFactory.Stloc( varContext ) ) );

					int i = 3;

					// check and call the original method first if the binding is AFTER
					CheckCallOriginalMethod( declAssembly,
												code, 
												BindingOpt.bindAfter, 
												ref i, 
												varOrigResult,
												varContext,
												varArr,
												newInstructions,
												code[0] );

					// now iterate all instructions
					foreach ( Instruction instr in code )
					{						
						if ( instr.OpCode.Value == OpCodes.Call.Value )
						{
							if ( instr is ILMethod )
							{
								ILMethod m = instr as ILMethod;
							
								RType r1 = libAssembly.RModuleDef.GetType( typeof(ControlFlow).FullName );

								// now check if this call is to the ControlFlow.Resume() method
								if ( m.Method.DeclaringType.NameWithNameSpace == r1.NameWithNameSpace &&
									m.Method.Name == "Resume" )
								{
									// remove the call instruction from the code
									newInstructions.Add( new Modification( i, instr, Modification.Operation.Delete ) );
									
									int j = newInstructions.Count; //store current index

									// now replace the call with the dynamic invoke of the delegate
									// passed in the context
									InsertDynamicInvoke( declAssembly, newInstructions, ref i, varContext, varArr );

									foreach( Modification mod in newInstructions )
									{
										if ( mod.instruction is ITargeter &&
											 (mod.instruction as ITargeter).ContainsTarget( instr ) )
										{
											Instruction newTarget = (newInstructions[j] as Modification).instruction;
											ITargeter iTargeter = mod.instruction as ITargeter;

											iTargeter.ReplaceTarget( instr, newTarget );

											newTarget.AddTargeter( iTargeter );
										}
									}
								}
								else
								{
									i++;
								}
							}
							else
							{
								i++;
							}
						} else 
						if ( instr.OpCode.Value == OpCodes.Ret.Value )
						{
							Instruction nextInstr = ILFactory.Ldloc( varOrigResult );
							
							// to ignore it we just store the value in the local variable
							newInstructions.Add( new Modification( i++, ILFactory.Stloc( varOrigResult ) ) );

							// insert the conditional call to the original method
							CheckCallOriginalMethod( declAssembly,
														code,
														BindingOpt.bindBefore,
														ref i,
														varOrigResult,
														varContext,
														varArr,
														newInstructions,
														nextInstr );


							// to ignore it we just store the value in the local variable
							newInstructions.Add( new Modification( i++, nextInstr ) );
						}
						else
						{
							++i;
						}
					}

					foreach (Modification m in newInstructions)
					{
						switch(m.operation)
						{
							case(Modification.Operation.Insert):
								code.Insert( m.position, m.instruction );
								break;
							case(Modification.Operation.Delete):
								code.Remove( m.position );
								break;
							default:
								break;
								// exception
						}
					}

					code.UpdateInstructionIndex();
				}
			}
		}

		/// <summary>
		/// This method calls the original method (via the delegate)
		/// in case the binding option in the context is the same as
		/// the one passed as parameter
		/// </summary>
		public static void CheckCallOriginalMethod( RAssemblyDef declAssembly,
													Code code, 
		                                            BindingOpt bindOption,
													ref int iPos,
													LocalVariable tmpResultVar,
													LocalVariable contextVar,
													LocalVariable varArr,
													ArrayList newInstructions,
													Instruction nextInstr )
		{
			RType ctxType = declAssembly.GetType( typeof(MethodContext).FullName );

			// compare the context's binding option with the one passed as argument
			// and if they're different continue with the next instruction
			newInstructions.Add(new Modification( iPos++, ILFactory.Ldloc( contextVar ) ) );
			newInstructions.Add(new Modification( iPos++, ILFactory.Ldfld( ctxType.GetField( "ctxBindOption" ) ) ) );
			newInstructions.Add(new Modification( iPos++, ILFactory.Ldc( (int)bindOption ) ) );
			newInstructions.Add(new Modification( iPos++, ILFactory.Bne_Un(nextInstr) ) );

			// *** within the IF statement ******************************
			// call the original method
			InsertDynamicInvoke( declAssembly, newInstructions, ref iPos, contextVar, varArr );

			// if the method is returning a value - store it
			newInstructions.Add(new Modification( iPos++, ILFactory.Stloc( tmpResultVar ) ) );

			newInstructions.Add(new Modification( iPos++, ILFactory.Ldloc( contextVar ) ) );
			newInstructions.Add(new Modification( iPos++, ILFactory.Ldc(1) ) );
			newInstructions.Add(new Modification( iPos++, ILFactory.Stfld( ctxType.GetField("ctxOriginalCalled") ) ) );
			
			// *** end of the IF statement ******************************
		}

		/// <summary>
		/// This method creates the IL code necessary for calling the DynamicInvoke
		/// method of a delegate. Used to adjust the aspects
		/// </summary>
		public static void InsertDynamicInvoke( RAssemblyDef declAssembly, 
												ArrayList newInstructions, 
												ref int iPos,
												LocalVariable ctxVar,
												LocalVariable varArr )
		{
			RType ctxType = declAssembly.GetType( typeof(MethodContext).FullName );
			RMethod dynamicInvoke = 
				GetMethodByName( libAssembly.GetType( typeof(System.Delegate).FullName ),
								"DynamicInvoke");

			newInstructions.Add( new Modification( iPos++, ILFactory.Ldc(1) ) );
			newInstructions.Add( new Modification( iPos++, ILFactory.Newarr( libAssembly.GetType(typeof(object).FullName) ) ) );
			newInstructions.Add( new Modification( iPos++, ILFactory.Stloc( varArr ) ) );
			
			newInstructions.Add( new Modification( iPos++, ILFactory.Ldloc( varArr ) ) );
			newInstructions.Add( new Modification( iPos++, ILFactory.Ldc(0) ) );
			newInstructions.Add( new Modification( iPos++, ILFactory.Ldloc( ctxVar ) ) );
			newInstructions.Add( new Modification( iPos++, ILFactory.Stelem( typeof(object) ) ) );

			newInstructions.Add(new Modification( iPos++, ILFactory.Ldloc( ctxVar ) ) );
			newInstructions.Add(new Modification( iPos++, ILFactory.Ldfld( ctxType.GetField( "ctxOriginalMethod" ) ) ) );
			newInstructions.Add(new Modification( iPos++, ILFactory.Ldloc( varArr ) ) );

			newInstructions.Add(new Modification( iPos++, ILFactory.Call( true, dynamicInvoke ) ) );
		}

		/// <summary>
		/// This method runs the PEVERIFY tool to validate the assembly.
		/// The result of the check is then written in the Output window
		/// </summary>
		public static void ValidateAssembly(string fileName )
		{
			string cmdLine = "C:\\Program Files\\Microsoft Visual Studio .NET 2003\\SDK\\v1.1\\Bin\\peverify.exe";

			DirectoryInfo inf = System.IO.Directory.GetParent( System.IO.Directory.GetCurrentDirectory() );
			
			ProcessStartInfo pInfo = new ProcessStartInfo( cmdLine, fileName );
			pInfo.RedirectStandardOutput = true;
			pInfo.UseShellExecute = false;

			Process p = Process.Start( pInfo );
			
			string output = p.StandardOutput.ReadToEnd();
			output = output.Replace("\0", "");
			output = output.Replace("\r\n", "\n");
			
			output = output.Replace( inf.ToString().ToLower() + "\\", "" );

			p.WaitForExit( );

			string delim = "\n";

			while ( output != string.Empty )
			{
				int idx = output.IndexOf(delim);

				if ( idx >= 0 )
				{
					string tmp = output.Substring(0, idx );
					System.Diagnostics.Trace.WriteLine(tmp);
				}
				else
					System.Diagnostics.Trace.WriteLine(output.Substring(0));
				
				if ( output.Length >= idx + delim.Length )
					output = output.Substring( idx + delim.Length );
			}

			System.Diagnostics.Trace.Write( output );
		}
	}
}