using System;
using System.Collections;
using System.Reflection;
using System.Xml.XPath;
using Rail.Reflect;

namespace AspectNet
{
	/// <summary>
	/// Summary description for MethodMatcher.
	/// </summary>
	public class MethodJPBuilder : IJPBuilder
	{
		#region internal classes
		/// <summary>
		/// Internal representation of parameter information
		/// </summary>
		private class ParamContraints
		{
			private string _paramName = null;
			private string _paramType = null;

			/// <summary>
			/// ctor
			/// </summary>
			public ParamContraints(string paramName, string paramType)
			{
                this._paramName = paramName;
				this._paramType = paramType;
			}

			/// <summary>
			/// Returns the name of parameter
			/// </summary>
			public string Name
			{
				get { return _paramName; }
			}

			/// <summary>
			/// Returns the type of the parameter
			/// </summary>
			public string Type
			{
				get { return _paramType; }
			}
		}
		#endregion

		private int			_constrAccess		= (int)XmlConstants.AccModifier.accAll;

		private string		_constrClassName	= XmlConstants.MATCH_ALL;
		private string		_constrName			= XmlConstants.MATCH_ALL;
		private string		_constrRetType		= XmlConstants.MATCH_ALL;

		private ArrayList	_constrParameters	= new ArrayList();
		
		// the prefix tree, used for caching the names of the methods
		// in the assembly. facilitates fast prefix-based searching
		private PrefixTree	_cacheMethodsPTree = new PrefixTree();

		#region IJPBuilder Members

		/// <summary>
		/// Returns the token to which this builder corresponds
		/// </summary>
		public string XmlTag
		{
			get
			{
				return XsdConstants.ELEMENT_MATCHMETHOD;
			}
		}

		/// <summary>
		/// Loads the assembly and precaches its contents for
		/// faster matching of pointcuts
		/// </summary>
		public void PreCacheAssembly(string asmPath)
		{
			BuildMethodsList(RAssemblyDef.LoadAssembly( asmPath ));
		}

		/// <summary>
		/// Load the builder's configuration data 
		/// from the XML file containing the mapping
		/// </summary>
		public void InitializeState(System.Xml.XPath.XPathNavigator node)
		{
			// reset current state
			Clear();

			// get iterator to all child elements
			XPathNodeIterator it = node.SelectChildren( XPathNodeType.Element );

			while ( it.MoveNext() )
			{
				XPathNavigator currNode = it.Current;
				string nodeName = currNode.Name;

				switch( nodeName )
				{
					case(XsdConstants.ELEMENT_CSTR_ACC_MODIFIER):
						GetAccessValue(currNode.GetAttribute(XsdConstants.ATTRIB_VALUE, XsdConstants.DEF_NAMESPACE));
						break;
					case(XsdConstants.ELEMENT_CSTR_CLNAME):
						_constrClassName = currNode.GetAttribute(XsdConstants.ATTRIB_VALUE, XsdConstants.DEF_NAMESPACE);
						break;
					case(XsdConstants.ELEMENT_CSTR_NAME):
						_constrName = currNode.GetAttribute(XsdConstants.ATTRIB_VALUE, XsdConstants.DEF_NAMESPACE);
						break;
					case(XsdConstants.ELEMENT_CSTR_RET_TYPE):
						_constrRetType = currNode.GetAttribute(XsdConstants.ATTRIB_VALUE, XsdConstants.DEF_NAMESPACE);
						break;
					case(XsdConstants.ELEMENT_CSTR_PARAMETERS):
						LoadParameters( currNode );
						break;
					default:
						throw new ApplicationException( "Invalid XML file! Unknown element:" + nodeName );
				}
			}
		}

		/// <summary>
		/// Summary description for IJPBuilder.
		/// </summary>
		public void MatchJoinPoints(PointCut pCut)
		{
			ArrayList methodDefinitions = new ArrayList();
			
			// first match all methods by their name (if no name
			// restriction is set, then all methods are returned here)
			FilterByName(methodDefinitions);

			// then filter by class name
			FilterByClass( ref methodDefinitions );

			// then filter by access modifier
			FilterByAccess(ref methodDefinitions);
			
			// then filter by return type
			FilterByReturnType(ref methodDefinitions);
			
			// then filter by method parameters
			FilterByParameters(ref methodDefinitions);

			// if there're any left, build up the join points list
			foreach ( RMethodDef mdef in methodDefinitions )
			{
				pCut.Add( new JoinPoint(pCut, mdef, typeof(MethodJPBuilder) ) );
			}
		}

		public void Reset()
		{
			Clear();
		}

		#endregion

		#region Filtering Methods

		/// <summary>
		/// Filter the list passed as arguments via
		/// the parameters constraints
		/// </summary>
		private void FilterByParameters(ref ArrayList list)
		{
			// we don't filter if we don't have filtering criteria
			if ( _constrParameters.Count == 0 ||
				 list.Count == 0 )
			{
				return;
			}

			ArrayList filtered = new ArrayList(list.Count / 2);

			// we check each method against the parameter criteria
			foreach ( RMethodDef md in list )
			{
				// flagged when all parameter criteria is satisfied
				bool bConstrSatisfied = true;

				// for each criterion
				foreach ( ParamContraints pc in _constrParameters )
				{
					// flagged when the current constraint is satisfied
					bool bLocConstrSatisfied = false;

					// iterate over all method's parameters
					foreach( RParameter rp in md.GetParameters() )
					{
						// we match by name if the filter is *, or
						// the names match
						bool match1 =	pc.Name == XmlConstants.MATCH_ALL ||
										pc.Name == rp.Name;

						// we match by type if the filter is *, or
						// the types match
						bool match2 =	pc.Type == XmlConstants.MATCH_ALL ||
										pc.Type == rp.ParameterType.NameWithNameSpace;

						// whether the local constraint is satisfied depends
						// on the conjunction of the two match flags
						bLocConstrSatisfied = match1 && match2;
	 
						// if we matched completely this parameter, then we should break
						// because we don't need to verify the rest of the method's parameters
						if ( bLocConstrSatisfied )
						{
							break;
						}
					}

					// if we failed for one of the constraints, 
					// we fail for the whole method
					if ( !bLocConstrSatisfied )
					{
						bConstrSatisfied = false;
						break;
					}
				}

				// if we didn't fail for any of the parameters, then the
				// method satisfies the filter
				if ( bConstrSatisfied )
				{
					filtered.Add( md );
				}
			}

			list = filtered;
		}

		/// <summary>
		/// Filter the list passed as arguments via
		/// the value in the return type member variable
		/// </summary>
		private void FilterByReturnType(ref ArrayList list)
		{
			// we don't filter if we don't have filtering criteria
			if ( _constrRetType.Length > 0 &&
				 _constrRetType != XmlConstants.MATCH_ALL )
			{
				ArrayList filtered = new ArrayList( list.Count / 2 );

				foreach ( RMethodDef md in list )
				{
					if ( md.ReturnType.ReturnType.NameWithNameSpace == _constrRetType )
					{
						filtered.Add( md );
					}
				}

				list = filtered;
			}
		}

		/// <summary>
		/// Filter the list passed as arguments via
		/// the value in the constraint class name variable
		/// </summary>
		private void FilterByClass( ref ArrayList methodDefinitions )
		{
			if ( _constrClassName != XmlConstants.MATCH_ALL && 
				 methodDefinitions.Count > 0 )
			{
				ArrayList filtered = new ArrayList( methodDefinitions.Count / 2 );

				foreach( RMethodDef md in methodDefinitions )
				{
					// if we have an asterisk
					int pos = _constrClassName.IndexOf( XmlConstants.MATCH_ALL );

					string name2Check = 
						_constrClassName.IndexOf( XsdConstants.NAMESPACE_DELIM ) > 0 ? 
												md.DeclaringType.NameWithNameSpace :
												md.DeclaringType.Name;

					if ( pos > 0 )
					{
						// we check for substring

						if ( name2Check.StartsWith( _constrClassName.Substring( 0, pos ) ) )
						{
							filtered.Add( md );
						}
					}
					else
					{
						// we check equality
						if ( name2Check == _constrClassName )
						{
							filtered.Add( md );
						}
					}
				}

				methodDefinitions = filtered;
			}
		}

		/// <summary>
		/// Filter the list passed as arguments via
		/// the value in the constraint access member variable
		/// </summary>
		private void FilterByAccess(ref ArrayList list)
		{
			// we don't filter if we don't have filtering criteria
			if ( _constrAccess != (int)XmlConstants.AccModifier.accAll &&
				list.Count > 0 )
			{
				ArrayList filtered = new ArrayList( list.Count / 2 );

				foreach ( RMethodDef md in list )
				{
					if ( ( md.Attributes & MethodAttributes.Abstract ) != 0 &&
						 ( _constrAccess & (int)XmlConstants.AccModifier.accAbstract ) != 0 )
					{
						filtered.Add(md);
					}
				
					if ( ( md.Attributes & MethodAttributes.Final ) != 0 &&
						 ( _constrAccess & (int)XmlConstants.AccModifier.accFinal ) != 0 )
					{
						filtered.Add(md);
					}

					if ( ( md.Attributes & MethodAttributes.Private ) != 0 &&
						 ( _constrAccess & (int)XmlConstants.AccModifier.accPrivate )!= 0  )
					{
						filtered.Add(md);
					}
				
					if ( ( md.Attributes & MethodAttributes.Public ) != 0 &&
						 ( _constrAccess & (int)XmlConstants.AccModifier.accPublic )!= 0  )
					{
						filtered.Add(md);
					}
				
					if ( ( md.Attributes & MethodAttributes.Family ) != 0 &&
						 ( _constrAccess & (int)XmlConstants.AccModifier.accProtected )!= 0  )
					{
						filtered.Add(md);
					}
				}

				list = filtered;
			}
		}

		private void FilterByName(ArrayList list)
		{
			_cacheMethodsPTree.GetNodesData( _constrName, list );
		}

		#endregion

		#region Private Methods
		/// <summary>
		/// Resets the state in between calls to the Initialize() method
		/// </summary>
		private void Clear()
		{
			_constrAccess		= (int) XmlConstants.AccModifier.accAll;
			_constrClassName	= XmlConstants.MATCH_ALL;
			_constrName			= XmlConstants.MATCH_ALL;
			_constrRetType		= XmlConstants.MATCH_ALL;
			
			_constrParameters.Clear();
		}

		/// <summary>
		/// Caches all methods within the current assembly
		/// </summary>
		private void BuildMethodsList(RAssemblyDef assembly)
		{
			_cacheMethodsPTree.Clear();

			// iterate over all types and methods and build up
			// the prefix tree data structure
			foreach ( RTypeDef type in assembly.RModuleDef.GetTypes() )
			{
				foreach ( RMethodDef meth in type.GetMethods() )
				{
					_cacheMethodsPTree.AddString( meth.Name, meth );
				}
			}
		}

		/// <summary>
		/// Determines the actual value that should be used for searching
		/// </summary>
		private void GetAccessValue(string strData)
		{
			_constrAccess = (int)XmlConstants.AccModifier.accNone;

			if ( strData.IndexOf( XmlConstants.ACC_ABSTRACT ) != -1 )
			{
				_constrAccess |= (int)XmlConstants.AccModifier.accAbstract;
			}

			if ( strData.IndexOf( XmlConstants.ACC_FINAL ) != -1 )
			{
				_constrAccess |= (int)XmlConstants.AccModifier.accFinal;
			}

			if ( strData.IndexOf( XmlConstants.ACC_PRIVATE ) != -1 )
			{
				_constrAccess |= (int)XmlConstants.AccModifier.accPrivate;
			}

			if ( strData.IndexOf( XmlConstants.ACC_PROTECTED ) != -1 )
			{
				_constrAccess |= (int)XmlConstants.AccModifier.accProtected;
			}

			if ( strData.IndexOf( XmlConstants.ACC_PUBLIC ) != -1 )
			{
				_constrAccess |= (int)XmlConstants.AccModifier.accPublic;
			}
		}

		/// <summary>
		/// Loads the parameters information
		/// </summary>
		private void LoadParameters(System.Xml.XPath.XPathNavigator node)
		{
			XPathNodeIterator itParams = node.SelectChildren( XPathNodeType.Element );

			while ( itParams.MoveNext() )
			{
				_constrParameters.Add( new ParamContraints( itParams.Current.GetAttribute( XsdConstants.ATTRIB_NAME, 
					XsdConstants.DEF_NAMESPACE ), 
					itParams.Current.GetAttribute( XsdConstants.ATTRIB_TYPE,
					XsdConstants.DEF_NAMESPACE ) ) );
			}
		}
		#endregion
	}
}
