using System;
using AspectNet.Transformations;
using Rail.Reflect;

namespace AspectNet
{
	/// <summary>
	/// Summary description for MethodJPWeaver.
	/// </summary>
	public class MethodJPWeaver : IJPWeaver
	{
		protected WeaveContext _currentContext = null;		
		protected const string ASPECTINITMETHOD = "AspectInit";

		#region IJPWeaver Members

		public bool TransformJoinPoint()
		{
			bool bResult = true;
			try
			{
				bool aspectInitCreated = false;

				RFieldDef aspInstanceVar = null;

				AdjustAssemblyReferences();

				// define the two variables:
				//	1. the binding variable
				//  2. the variable holding the aspect's instance
				bool bAspectDefined = DefineVariables( Context.binding, ref aspInstanceVar );

				// first create the DefImpl method in the class
				RMethodDef defImpl = RailHelper.CopyMethod( Context.pointType, Context.pointMethod );

				// create the delegate type in the aspect type
				RTypeDef delegateType = CreateDelegateType();
			
				// create a default handler method wich is a proxy to the DefImpl
				RMethodDef defHandler = CreateDefaultHandler( defImpl );

				// create the aspect handler
				RMethodDef aspectCaller = CreateAspectHandler( defHandler, aspInstanceVar );

				// now define an event of the delegate type within the class,
				// holding the joinppoint
				REventDef interceptEvent = CreateEvent( delegateType, Context.pointMethod );

				// adjust the original method so that the actual body is replaced with:
				// 1. context structure is created and initialized
				// 2. raise the newly created event
				AdjustOriginalMethod( delegateType, interceptEvent, defHandler );
			
				// check and create a static init method that's used for binding the
				// aspects to the interceptor events
				RMethodDef initMethod = CreateAspectInitMethod( ref aspectInitCreated );

				// if the method was missing, then adjust all constructors so that the
				// new static method is called
				if ( aspectInitCreated )
				{
					AdjustConstructors( initMethod  );				
				}

				// last we should bind the aspect's advice to the newly created
				// interceptor event
				BindAspectHandler( initMethod, interceptEvent, delegateType, aspInstanceVar, bAspectDefined, aspectCaller ); 
			}
			catch
			{
				bResult = false;
			}

			return bResult;
		}

		public WeaveContext Context
		{
			set { _currentContext = value; }
			get { return _currentContext; }
		}

		public void Reset()
		{
			_currentContext = null;
		}

		#endregion

		#region proxy methods
		/// <summary>
		/// This method creates a delegate type which has the same signature
		/// as the given method + an additional parameter of type Object (the context)
		/// </summary>
		protected virtual RTypeDef CreateDelegateType( )
		{
			return RailHelper.CreateDelegateType( Context.pointAssembly,
													Context.pointModule,
													Context.pointType,
													Context.pointMethod );
		}

		/// <summary>
		/// This method adds a reference to the AspectLib assembly and
		/// particularly a TypeRef to the MethodContext
		/// </summary>
		protected virtual void AdjustAssemblyReferences()
		{
			RailHelper.AdjustAssemblyReferences( Context.pointAssembly, Context.targetAdvice.Aspect.AspectType );
		}

		/// <summary>
		/// This method creates a default proxy method for the method 
		/// given as a parameter
		/// </summary>
		protected virtual RMethodDef CreateDefaultHandler( RMethodDef method )
		{
			return RailHelper.CreateDefaultHandler( Context.pointAssembly, Context.pointType, method );
		}

		/// <summary>
		/// This method sets the binding variable to the context parameter and
		/// invokes the aspect's advice
		/// </summary>
		protected virtual RMethodDef CreateAspectHandler(  RMethodDef defHandler,
			RFieldDef aspectInstance )
		{
			return RailHelper.CreateAspectHandler( Context.pointAssembly, 
													Context.pointType, 
													Context.AspectType,
													defHandler, 
													Context.AdviceMethod,
													Context.pointMethod,
													aspectInstance, 
													Context.binding );
		}

		/// <summary>
		/// This method creates an event (private and static) compliant 
		/// with the delegate type passed.
		/// </summary>
		protected virtual REventDef CreateEvent( RTypeDef delegateType, RMethodDef method )
		{
			return RailHelper.CreateEvent( Context.pointAssembly, Context.pointType, delegateType, method  );
		}

		/// <summary>
		/// This method checks if a method with the given name already exists
		/// and if it does - returns it. If not - creates an empty method.
		/// </summary>
		
		protected virtual RMethodDef CreateAspectInitMethod( ref bool create )
		{
			return RailHelper.CreateAspectInitMethod( Context.pointAssembly, Context.pointType, ASPECTINITMETHOD, ref create );
		}

		/// <summary>
		/// This method inspects the static aspect initializer method
		/// and adds code for binding the defHandler param to the newEvent
		/// event
		/// </summary>
		protected virtual void BindAspectHandler( RMethodDef initMethod, 
													REventDef newEvent, 
													RTypeDef delegateType,
													RField aspInstanceVar, 
													bool bAspectDefined,
													RMethodDef aspectCaller)
		{
			RailHelper.BindAspectHandler( initMethod,
											newEvent,
											aspectCaller,
											Context.AspectType,
											delegateType,
											aspInstanceVar,
											bAspectDefined);
		}

		/// <summary>
		/// This method adjusts the original method so that:
		///		1. creates and initializes the MethodContext
		///		2. raises the method interceptor's event
		/// </summary>
		protected virtual void AdjustOriginalMethod( RTypeDef delegateType, 
			REventDef newEvent,
			RMethodDef defHandler )
		{
			RailHelper.AdjustOriginalMethod( Context.pointAssembly,
												Context.pointMethod,
												Context.pointType,
												delegateType,
												newEvent,
												defHandler,
												Context.AdviceMethod.Name );
		}

		/// <summary>
		/// This method will adjust all current constructors to first call the
		/// static initializing method for the aspects, before anything else
		/// </summary>
		protected virtual void AdjustConstructors( RMethodDef initMethod )
		{
			RailHelper.AdjustConstructors( Context.pointAssembly, Context.pointType, initMethod );
		}

		/// <summary>
		/// This method defines a local constant which defines the binding
		/// of the aspect to the joinpoint. To accomplish this, the name of the
		/// variable is generated: _ + pointname + advicename + aspectname
		/// </summary>
		protected virtual bool DefineVariables( int bindOpt, ref RFieldDef aspectVar )
		{
			return RailHelper.DefineVariables( Context.pointAssembly, 
												Context.pointType, 
												Context.pointMethod, 
												Context.AdviceMethod.Name,
												ref aspectVar,
												Context.AspectType );
		}

		#endregion
	}
}
