using System;
using System.Collections;
using System.Xml;
using System.Xml.Schema;
using System.Xml.XPath;
using AspectLib;

namespace AspectNet
{
	/// <summary>
	/// Summary description for Class1.
	/// </summary>
	public class PointCutManager
	{
		private ArrayList		_assemblies2Process = new ArrayList();

		private Hashtable		_pointCuts			= new Hashtable();

		private string			_currMappingFName	= null;
		private string			_schemaFName		= null;

		private XmlSchema		_compiledSchema		= null;
		private XPathDocument	_currentDocument	= null;

		public event XMLErrorHandler Error;


		/// <summary>
		/// Default constructor
		/// </summary>
		public PointCutManager(string currSchema)
		{
			_schemaFName = currSchema;
	
		}

		public string XmlMappingPath
		{
			set { _currMappingFName = value; }
		}

		/// <summary>
		/// Loads the schema and the XML mapping, performs validation
		/// </summary>
		private void LoadXmlMapping()
		{
			try
			{
				// load the schema
				if ( _compiledSchema == null )
				{
					_compiledSchema = XmlSchema.Read( new XmlTextReader(_schemaFName ), new ValidationEventHandler (SchemaValidationHandler) );
					_compiledSchema.Compile(new ValidationEventHandler (SchemaValidationHandler));
				}

				// create a validating reader and load set the schema
				XmlTextReader tr = new XmlTextReader(_currMappingFName);
				XmlValidatingReader vr = new XmlValidatingReader(tr);
				vr.ValidationType = ValidationType.Schema;
				vr.ValidationEventHandler += new ValidationEventHandler (MappingValidationHandler);
			
				vr.Schemas.Add( _compiledSchema );

				_currentDocument = new XPathDocument(vr);
			}
			catch(Exception e)
			{
				if ( Error != null )
				{
					Error(this, new ErrorEventArgs(e));
				}	
			}
		}

		/// <summary>
		/// Stores the path to the assembly for further processing
		/// </summary>
		public void ProcessAssembly( string asmFName )
		{
			_assemblies2Process.Add( asmFName );
		}

		/// <summary>
		/// This method should be called before invoking the BuildJoinPoints method.
		/// The idea is that in case we have aspects and basic functionality all in one
		/// assembly, then here we should update the already stored assembly paths. 
		/// 
		/// This is needed because when adjusting the aspects, we actually modify the 
		/// assemblies that contain aspects, so that they are afterwards stored in the 
		/// output folder.
		/// </summary>
		public void UpdateAssemblyPaths( ICollection processedAssemblies )
		{
			// create a hash table, to make keys matching easier
			Hashtable hTable = new Hashtable();

			// add the old assemblies which will be processed
			foreach ( string oldAsmPath in _assemblies2Process )
			{
				// the key is the assembly name (no path)
				string key = ( oldAsmPath.IndexOf("\\") != -1 ) ?
					oldAsmPath.Substring( oldAsmPath.LastIndexOf("\\") + 1 ) : oldAsmPath;

				// the value we store is the full path to the assembly
				hTable.Add( key, oldAsmPath );
			}

			// now update the paths with the assemblies processed by the AspectWeaver
			foreach ( string newAsmPath in processedAssemblies )
			{
				string key = ( newAsmPath.IndexOf("\\") != -1 ) ?
					newAsmPath.Substring( newAsmPath.LastIndexOf("\\") + 1 ) : newAsmPath;

				hTable[key] = newAsmPath;
			}

			// now, when we have all the new paths stored in the hash table,
			// it's time to update the _assemblies2Process array list
			_assemblies2Process.Clear();

			foreach( string asmPath in hTable.Values )
			{
				_assemblies2Process.Add( asmPath );
			}
		}

		/// <summary>
		/// Returns the point cuts that were found during the last
		/// call to the BuildJoinPoints method
		/// </summary>
		public ArrayList PointCuts
		{
			get
			{
				ArrayList res = new ArrayList();

				foreach( PointCut pCut in _pointCuts.Values )
				{
					res.Add( pCut );
				}

				return res;
			}
		}

		#region Error handling
		
		public void SchemaValidationHandler(object sender, ValidationEventArgs args)
		{
			if ( Error != null )
			{
				Error(this, new ErrorEventArgs(ErrorEventArgs.ErrKind.ErrSchema, sender, args));
			}
		}

		public void MappingValidationHandler(object sender, ValidationEventArgs args)
		{
			if ( Error != null )
			{
				Error(this, new ErrorEventArgs(ErrorEventArgs.ErrKind.ErrMapping, sender, args));
			}
		}
		#endregion

		/// <summary>
		/// Starts the actual processing of the assemblies (the scanning for the
		/// join points, matching the pointcuts)
		/// </summary>
		public void BuildJoinPoints()
		{
			try
			{
				Clear();

				// load the XML mapping
				LoadXmlMapping();

				ArrayList jpBuildersUsed = new ArrayList();

				// get the join point builders, used in the XML
				IdentifyBuilders(jpBuildersUsed);

				// for each assembly to be processed
				foreach( string asmPath in _assemblies2Process )
				{
					// make the builders used in the mapping XML file,
					// pre-cache the assemblies, so that the join point idenditification
					// is accomplished quick enough
					foreach ( IJPBuilder builder in jpBuildersUsed )
					{
						builder.PreCacheAssembly( asmPath );
					}
				
					// now process the mapping and get the joinpoints
					ProcessAssemblyInt( jpBuildersUsed );
				}

				if ( _currentDocument != null &&
					 _pointCuts.Count == 0 )
				{
					Error(this, new ErrorEventArgs("No join points have been found to match the XML mapping"));
				}

			}
			catch(Exception e)
			{
				if ( Error != null )
				{
					Error(this, new ErrorEventArgs(e));
				}	
			}
		}

		public void Reset()
		{
			_assemblies2Process.Clear();
			Clear();
		}

		/// <summary>
		/// Clears the class'es state
		/// </summary>
		private void Clear()
		{
			_currentDocument = null;
			_pointCuts.Clear();
		}

		/// <summary>
		/// Starts the actual processing of the assemblies (the scanning for the
		/// join points, matching the pointcuts)
		/// </summary>
		private void ProcessAssemblyInt( ArrayList jpBuilders )
		{
			// create a navigator for the XML mapping document
			XPathNavigator nav = _currentDocument.CreateNavigator();
		
			// create and compile an XPath expression
			XPathExpression expr = nav.Compile( XsdConstants.XPATH_DELIM + XsdConstants.ELEMENT_WEAVE );

			// get iterator
			XPathNodeIterator it = nav.Select( expr );

			// for each weaving configuration
			while ( it.MoveNext() )
			{
				XPathNavigator nodeWeave = it.Current;
			
				// get the pointcut ID
				string pointCutID	= nodeWeave.GetAttribute( XsdConstants.ATTRIB_POINTCUT, 
																XsdConstants.DEF_NAMESPACE );
				
				// build the query to match the respective pointcut definition
				string query = string.Format("//{0}[@{1}='{2}']", XsdConstants.ELEMENT_POINTCUT,
																	XsdConstants.ATTRIB_NAME,
																	pointCutID );

				// get the pointcut definition now
				XPathNodeIterator pcutIterator = nav.Select( query );
				
				if ( pcutIterator.Count > 0 )
				{
					// get the bindOptions
					string bindOptions	= nodeWeave.GetAttribute( XsdConstants.ATTRIB_BINDOPTIONS, 
																	XsdConstants.DEF_NAMESPACE );

					// get either existing pointcut OR create new one
					PointCut pCut = GetPointCut(pointCutID, 
												nodeWeave.GetAttribute( XsdConstants.ATTRIB_ASPECT, 
																		XsdConstants.DEF_NAMESPACE ),
												nodeWeave.GetAttribute( XsdConstants.ATTRIB_ADVICE , 
																		XsdConstants.DEF_NAMESPACE ),
					                            (BindingOpt)BindingOpt.Parse(typeof(BindingOpt), bindOptions) );

					// process the pointcut
					ProcessPointcut( pCut, pcutIterator );

					if ( pCut.Count == 0 )
					{
						_pointCuts.Remove( pointCutID );
					}
				}
			}
		}

		/// <summary>
		/// Checks for the existence of a pointcut with pCutID and if present 
		/// returns it, otherwise create a new one
		/// </summary>
		private PointCut GetPointCut( string pCutID, 
										string aspect, 
										string advice, 
										BindingOpt bOpts )
		{
			PointCut res = null;

			if ( _pointCuts.ContainsKey(pCutID) )
			{
				res = (PointCut)_pointCuts[pCutID];
			}
			else
			{
				res = new PointCut( aspect, advice, bOpts );
				_pointCuts[pCutID] = res;
			}

			return res;
		}

		/// <summary>
		/// Starts the actual processing of the assemblies (the scanning for the
		/// join points, matching the pointcuts)
		/// </summary>
		private void IdentifyBuilders(ArrayList jpBuildersUsed)
		{
			// create a navigator for the XML mapping document
			XPathNavigator nav = _currentDocument.CreateNavigator();

			// now check for the presence of each of the builders
			foreach ( IJPBuilder builder in JPFactory.Instance.JPBuilders.Values )
			{
				// create and compile an XPath expression
				XPathExpression expr = nav.Compile( XsdConstants.XPATH_DELIM + builder.XmlTag );

				// get iterator
				XPathNodeIterator it = nav.Select( expr );

				// if the node is present, then the builder should be used
				if ( it != null && it.Count > 0 )
				{
					jpBuildersUsed.Add( builder );
				}
			}
		}

		/// <summary>
		/// Process the xml mapping for a pointcut definition
		/// </summary>
		private void ProcessPointcut(PointCut cut, XPathNodeIterator pCutIterator)
		{
			pCutIterator.MoveNext();
			
			// create a navigator for the XML mapping document
			XPathNavigator nav = pCutIterator.Current;
			
			// move to the combine node
			nav.MoveToFirstChild();
			
			// now get all predicates that are within the match_all group
			XPathNodeIterator itPredicates = nav.SelectChildren( XPathNodeType.Element );

			while ( itPredicates.MoveNext() )
			{
				// get a builder for the xml tag
				IJPBuilder jpBuilder = JPFactory.Instance.GetBuilder(itPredicates.Current.Name);
				
				// make it load its configuration from the XML node
				jpBuilder.InitializeState( itPredicates.Current );

				// now search for the matching join points and add them to
				// the pointcut
				jpBuilder.MatchJoinPoints( cut );
			}
        }
	}
}
