using System;
using System.Collections;

namespace AspectNet
{
	/// <summary>
	/// Summary description for PrefixTree.
	/// </summary>
	public class PrefixTree
	{
		// the root node of the prefix tree
		private Node _rootNode = new Node();

		/// <summary>
		/// ctor
		/// </summary>
		public PrefixTree()
		{
		}

		/// <summary>
		/// This method takes care for adding the required nodes, 
		/// needed to represent a particular string within the prefix tree
		/// </summary>
		public void AddString(string str, object data)
		{
			_rootNode.AddString( str, data );
		}

		/// <summary>
		/// This method returns the data objects, which match a particular
		/// search string. 
		/// 
		/// The search string should be in the form: [0...n characters]*
		/// The asterisk means to match all subsequent words, prefixed
		/// by the sequence [0...n characters.
		/// 
		/// Example: the string "GetD*" will match all of the following methods:
		///		- GetD
		///		- GetDate
		///		- GetData
		///		- GetDataEx
		///	
		///	Using only asterisk will result in returning all the data
		///	stored at all nodes of the prefix tree
		/// </summary>
		public void GetNodesData(string srchString, ArrayList nodesFound)
		{
			_rootNode.GetData( srchString, nodesFound );
		}

		/// <summary>
		/// This method clears the root node and forces the GC to collect,
		/// thus presuming the old prefix tree nodes will be collected
		/// </summary>
		public void Clear()
		{
			_rootNode = new Node();
			System.GC.Collect();
		}

		#region nested classes

		/// <summary>
		/// Internal representation of a node of the prefix tree
		/// </summary>
		private class Node
		{
			// static counter of the size of the tree
			private int _instanceCount = 0;

			// hash table containing the child nodes
			private Hashtable	_childNodes		= new Hashtable();
			
			// flag whether there're whole words stored in this node
			private bool		_hasWords		= false;
			
			// the data that's stored on the node level
			private ArrayList	_nodeData		= new ArrayList();

			/// <summary>
			/// ctor
			/// </summary>
			public Node()
			{
				
			}

			/// <summary>
			/// This method takes care for adding the required nodes, 
			/// needed to represent a particular string within the prefix tree
			/// </summary>
			public void AddString(string remainingStr, object data)
			{
				bool bEndProcessing = false;
				Node currentNode = this;

				// until the string is empty
				while ( !bEndProcessing )
				{
					// if this is the last character, we should add the data
					// to the node and mark that it contains at least one complete word
					if ( remainingStr.Length == 0 )
					{
						currentNode._hasWords = true;
						currentNode._nodeData.Add( data );
						bEndProcessing = true;
					}
						// if we have a route to the rest of the string
					else if ( currentNode._childNodes.ContainsKey( remainingStr[0] )  )
					{
						// determine the next node
						currentNode = (Node)currentNode._childNodes[remainingStr[0]];
						// adjust the remaining string
						remainingStr = remainingStr.Substring(1);
					}
					else
					{
						// otherwise we should create a new node
						Node newNode = new Node();
						
						_instanceCount++;

						// add a route to the new node
						currentNode._childNodes.Add( remainingStr[0], newNode );
						
						// mark the current node as the one to be processed
						currentNode = newNode;
						
						// adjust the remaining string
						remainingStr = remainingStr.Substring(1);
					}					
				}
			}

			
			/// <summary>
			/// This method returns the data objects, which match a particular
			/// search string. 
			/// 
			/// The search string should be in the form: [0...n characters]*
			/// The asterisk means to match all subsequent words, prefixed
			/// by the sequence [0...n characters.
			/// 
			/// Example: the string "GetD*" will match all of the following methods:
			///		- GetD
			///		- GetDate
			///		- GetData
			///		- GetDataEx
			///	
			///	Using only asterisk will result in returning all the data
			///	stored at all nodes of the prefix tree
			/// </summary>
			public void GetData(string searchString, ArrayList result)
			{
				bool bEndProcessing = false;

				Node currentNode = this;

				// until the processing is over
				while ( !bEndProcessing )
				{
					// by default we assume that we have to stop processing at this point
					bEndProcessing = true;

					if ( searchString.Length == 0 )
					{
						// if we've reached the word and it contains any
						// data in its node - return it
						if ( currentNode._hasWords == true )
						{
							result.AddRange(currentNode._nodeData);
						}
					}
						// if we reached the asterisk
					else if ( searchString == XmlConstants.MATCH_ALL )
					{
						// create a queue to be used during the tree traversal
						// thus the traversal is breadth-first
						Queue nodes2Process = new Queue();
						
						// enqueue the current node as the root of the subtree
						nodes2Process.Enqueue( currentNode );

						// while there're nodes to be processed
						while ( nodes2Process.Count > 0 )
						{
							// get the head of the queue
							Node cNode = (Node)nodes2Process.Dequeue();

							// process the head
							if ( cNode._hasWords )
							{
								result.AddRange( cNode._nodeData );
							}

							// append all its children to the queue
							foreach ( Node n in cNode._childNodes.Values)
							{
								nodes2Process.Enqueue( n );
							}
						}
					}
						// otherwise we should route to next node
					else if ( currentNode._childNodes.ContainsKey( searchString[0] )  )
					{
						// if we have a route, then determine the next node
						currentNode = (Node)currentNode._childNodes[searchString[0]];
						
						// adjust the search string
						searchString = searchString.Substring(1);
						
						// reset the flag so we don't stop the processing cycle
						bEndProcessing = false;
					}
				}
			}

			
			/// <summary>
			/// Returns whether there're whole words stored in this node
			/// </summary>
			public bool ContainsAnyWords
			{
				get { return _hasWords; }
			}

			/// <summary>
			/// Returns the children hashtable
			/// </summary>
			public Hashtable Children
			{
				get { return _childNodes; }
			}

			/// <summary>
			/// Returns the data of this node
			/// </summary>
			public ArrayList Data
			{
				get { return _nodeData; }
			}
		}
		#endregion
	}
}
