using System;
using System.Windows.Forms;
using AspectNet;
using AspectNet.Transformations;
using EnvDTE;

namespace AspectNetAddIn
{


	/// <summary>
	/// Summary description for EventSink.
	/// </summary>
	public class AspectNetEventSink
	{
		
		public const string			PROPERTY_FULLPATH = "FullPath";
		public const string			PROPERTY_OUTPUTPATH = "OutputPath";
		public const string			PROPERTY_OUTPUTFNAME = "OutputFileName";

		private SolutionEvents		_solutionEvents;
		private BuildEvents			_buildEvents;
		private _DTE				_application;

		private AspectWeaver		_aspectWeaver;
		private PointCutManager		_aspectPcutManager;

		private Settings			_settings;

		private bool				_isInitialized = false;

		private bool				_isBuildSuccessful = true;
		
		public bool Initialized
		{
			get { return _isInitialized; }
		}

		/// <summary>
		/// Default ctor
		/// </summary>
		public void Initialize( _DTE application )
		{	
			_isInitialized = true; 

			_application = application;
			// store event providers in instance variables, thus
			// preventing them from being garbage collected
			_solutionEvents = application.Events.SolutionEvents;
			_buildEvents = application.Events.BuildEvents;

			_settings = new Settings(_application);

			// register the handlers
			RegisterEventHandlers( application );

			// create weaver and pointcut manager
			_aspectWeaver = new AspectWeaver();
			_aspectPcutManager = new PointCutManager(_settings.SchemaLocation);

			_aspectPcutManager.Error += new XMLErrorHandler(_aspectPcutManager_Error);		
		}

		/// <summary>
		/// Shows the settings dialog
		/// </summary>
		public bool ShowSettingsDialog()
		{
			SettingsForm frm = new SettingsForm(_settings);
			DialogResult rslt = frm.ShowDialog();
			_settings.SaveSettings();

			return true;
		}

		private void RegisterEventHandlers( _DTE application )
		{
			// bind to solution events
			_solutionEvents.Opened += new _dispSolutionEvents_OpenedEventHandler(SolutionEvents_Opened);
			_solutionEvents.QueryCloseSolution += new _dispSolutionEvents_QueryCloseSolutionEventHandler(SolutionEvents_QueryCloseSolution);
			_solutionEvents.AfterClosing += new _dispSolutionEvents_AfterClosingEventHandler(SolutionEvents_AfterClosing);
			_solutionEvents.BeforeClosing += new _dispSolutionEvents_BeforeClosingEventHandler(SolutionEvents_BeforeClosing);

			// bind to build events
			_buildEvents.OnBuildBegin += new _dispBuildEvents_OnBuildBeginEventHandler(BuildEvents_OnBuildBegin);
			_buildEvents.OnBuildDone += new _dispBuildEvents_OnBuildDoneEventHandler(BuildEvents_OnBuildDone);
			_buildEvents.OnBuildProjConfigBegin += new _dispBuildEvents_OnBuildProjConfigBeginEventHandler(BuildEvents_OnBuildProjConfigBegin);
			_buildEvents.OnBuildProjConfigDone += new _dispBuildEvents_OnBuildProjConfigDoneEventHandler(BuildEvents_OnBuildProjConfigDone);
		}

		#region solution event handlers
		private void SolutionEvents_Opened()
		{
			_settings = new Settings(_application);
		}

		private void SolutionEvents_QueryCloseSolution(ref bool fCancel)
		{
		}

		private void SolutionEvents_AfterClosing()
		{
		}

		private void SolutionEvents_BeforeClosing()
		{
		}
		#endregion solution event handlers

		#region build event handlers
		private void BuildEvents_OnBuildBegin(vsBuildScope Scope, vsBuildAction Action)
		{
			_isBuildSuccessful = true;
		}

		private void BuildEvents_OnBuildProjConfigBegin(string Project, string ProjectConfig, string Platform, string SolutionConfig)
		{
		}

		private void BuildEvents_OnBuildProjConfigDone(string Project, string ProjectConfig, string Platform, string SolutionConfig, bool Success)
		{
			if ( !Success )
			{
				_isBuildSuccessful = false;
			}
		}

		private void BuildEvents_OnBuildDone(vsBuildScope Scope, vsBuildAction Action)
		{
			if ( _settings.Enabled &&
				 _settings.IsConfigurationValid &&
				 _isBuildSuccessful )
			{
				try
				{
					WeaveProjects();	

					_aspectWeaver.Reset();
					_aspectPcutManager.Reset();
					JPFactory.Instance.Reset();

					GC.Collect();
				}
				catch ( Exception e )
				{
					TaskList tlWindow = (TaskList) (_application.Windows.Item( Constants.vsWindowKindTaskList ).Object);

					tlWindow.TaskItems.Add( "Aspect.Net", 
											" ", 
											e.Message,
											vsTaskPriority.vsTaskPriorityHigh,
											vsTaskIcon.vsTaskIconCompile,
											false,
											" ",
											0,
											false,
											true );
				}
			}
		}

		
		#endregion build event handlers

		

		/// <summary>
		/// This method will trigger the weaving process. Called at the
		/// end of the build sequence of events
		/// </summary>
		private void WeaveProjects()
		{
			RailHelper.AspectLibPath = _settings.AspectLibLocation;

			// set the mapping and output locations
			_aspectPcutManager.XmlMappingPath = _settings.XmlMapping;
			_aspectWeaver.OutputFolder = _settings.OutputLocation;

			// now iterate over the projects
			for ( int i = 1; i <= _application.Solution.Projects.Count; i++)
			{
				Project project = _application.Solution.Projects.Item(i);

				if ( project.ConfigurationManager == null )
				{
					continue;
				}

				Configuration actConfig = project.ConfigurationManager.ActiveConfiguration;

				string assemblyFileName = 
					GetPropertyValue( project.Properties, PROPERTY_FULLPATH )		+
					GetPropertyValue( actConfig.Properties, PROPERTY_OUTPUTPATH )	+
					GetPropertyValue( project.Properties, PROPERTY_OUTPUTFNAME );

				// pre-process aspects in current assembly
				_aspectWeaver.LoadAspects( assemblyFileName );

				// identify joinpoints in current assembly
				_aspectPcutManager.ProcessAssembly( assemblyFileName );
			}

			// handle the case where we have both - aspects and base code in the same assembly
			_aspectPcutManager.UpdateAssemblyPaths( _aspectWeaver.AdjustedAssemblies );
			
			// identify the join points and the point cuts
			_aspectPcutManager.BuildJoinPoints();

			// and noooooow..... touche.... weave everything
			_aspectWeaver.ProcessJoinPoints( _aspectPcutManager.PointCuts );
		}

		/// <summary>
		/// Returns the string value of a property as string
		/// </summary>
		private string GetPropertyValue(Properties properties, string propName )
		{
			string result = null;

			for (int k = 1; k <= properties.Count; ++k)
			{
				Property prop = properties.Item( k );
				
				if ( prop.Name == propName )
				{
					result = prop.Value as string;
					break;
				}
			}

			return result;
		}

		/// <summary>
		/// Handles the errors found during join points identification
		/// </summary>
		private void _aspectPcutManager_Error(object sender, ErrorEventArgs vargs)
		{
			TaskList tlWindow = (TaskList) (_application.Windows.Item( Constants.vsWindowKindTaskList ).Object);

			bool isMappingError = vargs.ErrorKind == ErrorEventArgs.ErrKind.ErrMapping;

			string msg = isMappingError ? "Invalid XML file! Check the schema and fix the error." :
													"Errors loading the schema file";

			if ( vargs.Arguments != null )
			{
				msg += " (" + vargs.Arguments.Message + ")";
			}

			tlWindow.TaskItems.Add( "Aspect.Net", 
									" ", 
									msg,
									vsTaskPriority.vsTaskPriorityHigh,
									vsTaskIcon.vsTaskIconCompile,
									false,
									isMappingError ? _settings.XmlMapping : _settings.SchemaLocation,
									0,
									false,
									true );
		}
	}
}
