/*
	The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); 
	you may not use this file except in compliance with the License. You may obtain a copy of the License at 
	http://www.mozilla.org/MPL/ 
	Software distributed under the License is distributed on an "AS IS" basis, 
	WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the specific language 
	governing rights and limitations under the License. 

	The Original Code is RAIL(Runtime Assembly Instrumentation Library) Alpha Version.
	The Initial Developer of the Original Code is University of Coimbra,
	Computer Science Department, Dependable Systems Group. Copyright (C) University of Coimbra. 
	All Rights Reserved.  
*/

using System;
using System.Diagnostics;
using System.Collections;
using System.Reflection;
using System.Reflection.Emit;
using Rail.Reflect;
using Mono.PEToolkit;
using Mono.PEToolkit.Metadata;
using Rail.Exceptions;

namespace Rail.MSIL 
{
	/// <summary>
	/// Represents a method body
	/// </summary>
	public class MethodBody 
	{
		// Should not be null
		/// <summary>
		/// Holds the exception blocks of the method
		/// </summary>
		internal ExceptionTable excpTable;
		/// <summary>
		/// max stack property of the method.
		/// </summary>
		internal int maxStack;

		/// <summary>
		/// True if locals must be initialized with the default constructor.
		/// </summary>
		internal bool initializeLocals; 

		/// <summary>
		/// Return an object able to access the locals signature in the heap.
		/// </summary>
		internal IList locals;

		/// <summary>
		/// The corresponding rmethod.
		/// </summary>
		internal RMethodBase parent;

		/// <summary>
		/// Return a cursor to the stream of IL instructions.
		/// </summary>
		internal Code methodCode;
    
		/// <summary>
		/// MethodBody internal constructor
		/// </summary>
		/// <param name="loadCtx">The MethodLoadContext where to get the Assmbl object</param>
		/// <param name="parent">The MethodBase object that owns this body</param>
		/// <param name="methodStart">The point in memory where to start reading the code of the method</param>
		internal MethodBody(MethodLoadContext loadCtx,MethodIL methodIL, RMethodBase parent,Image pe) 
		{
			if (parent == null)
				throw new InvalidProgramException();

			this.parent = parent;
			this.maxStack = methodIL.MaxStack;
      
			/*********************************
			 * Local variable signatures
			 *********************************/
			// Token to the StandAloneSig table with the Locals signature.
			//TODO: verify it next operation is correct
			int standAloneSigIndex = methodIL.GetLocalSignatureToken() & 0x00FFFFFF;
			MDStream blobStream = pe.MetadataRoot.Streams["#Blob"] as MDStream;
//			if (!methodIL.InitLocals || methodIL.GetLocalSignatureToken()==0) 
			if (methodIL.GetLocalSignatureToken()==0) 
			{
				locals = null;
			} 
			else 
			{
				StandAloneSigRow metaStandAloneSig = 
					(StandAloneSigRow)pe.MetaDataRoot.TablesHeap[TableId.StandAloneSig][standAloneSigIndex];
				initLocals(blobStream.RawData, metaStandAloneSig); 
			}
      
			/*********************************
			 * Read the instruction list
			 *********************************/      
			Hashtable posToIns = new Hashtable();
			methodCode = new Code(loadCtx, this, methodIL, posToIns);

			/*********************************
			 * Exception handling data
			 *********************************/    
			if (!methodIL.HasMoreSections || methodIL.IsMethodTiny()) 
			{
				excpTable = null;
			} 
			else 
			{
				// A method with a fat header does not necessarily have Exceptions.
				// perhaps we should check if it has before creating an ExceptionTable object.
				excpTable = new ExceptionTable(loadCtx, this, methodIL.GetSEHReader() , posToIns);
			}
		}

		/// <summary>
		/// Empty contructor of MethodBody
		/// </summary>
		internal MethodBody()
		{
		}

		/// <summary>
		/// Public contructor for creating a new and empty MethodBody object
		/// </summary>
		/// <param name="rMBase">The MethodBase that owns the MethodBody</param>
		public MethodBody(RMethodBase rMBase)
		{
			this.excpTable = new ExceptionTable();
			this.maxStack = 0;
			this.methodCode = new Code(this);
			this.parent = rMBase;
		}

		/// <summary>
		/// Internal Clone method, useful for copying components between assemblies
		/// </summary>
		/// <param name="mBase">The method that owns the MethodBody to clone</param>
		/// <returns>The newly cloned MethodBody object</returns>
		internal MethodBody Clone(RMethodBase mBase)
		{
			MethodBody mb = new MethodBody();
			mb.maxStack = this.maxStack;
			mb.initializeLocals = this.initializeLocals;; 
			if (this.parent!=null)
				mb.parent = mBase;
			if (this.methodCode!=null)
				this.methodCode.UpdateInstructionIndex();
			if (this.locals!=null)
			{	
				mb.locals = new ArrayList();
				foreach (LocalVariable lV in this.locals)
				{
					//mb.locals.Add(new LocalVariable(mb,lV.Type,lV.pos));
					mb.locals.Add(new LocalVariable(mb,((RTypeDef)mb.parent.DeclaringType).GetCorrectRType(lV.Type),lV.pos));
				}
			}
			if (this.methodCode!=null)
				mb.methodCode = this.methodCode.Clone(mb);	
			if (this.excpTable!=null)
				mb.excpTable = this.excpTable.Clone(mb);
			return mb;
		}

		#region Write method
		/// <summary>
		/// Dumps a method using Reflection.Emit
		/// </summary>
		/// <param name="ig">The ILGenerator object</param>
		/// <param name="tResolver">Objects used to transform Rail objects into Reflection.Emit objects</param>
		internal void Write(ILGenerator ig, TypeResolver tResolver) 
		{
			/****************************************
			 * Phase 1. 
			 ****************************************/ 
			Hashtable lLocals = new Hashtable();
			if (LocalVarsCount>0)
				for (int i = 0; i < LocalVarsCount;i++)
				{
					LocalVariable locVar = GetLocalVar(i);
					LocalBuilder lBuilder = ig.DeclareLocal(tResolver.ResolveRType(locVar.Type));
					// And saves it for future use
					lLocals.Add(locVar, lBuilder);
				}
			
			methodCode.Write(ig, tResolver,lLocals);
      
		}
		#endregion

		/// <summary>
		/// Gets or sets the maximum size of the stack for this method.
		/// </summary>
		public int MaxStack 
		{
			get 
			{
				return maxStack;
			}
			set 
			{
				this.maxStack = value;
			}
		}

		/// <summary>
		/// Returns the number of local variables
		/// </summary>
		public int LocalVarsCount 
		{
			get 
			{
				return locals == null ? 0 : locals.Count;
			}
		}

		/// <summary>
		/// Returns local variable define on this method.
		/// </summary>
		/// <param name="index">The position where the LocalVariable appears</param>
		/// <returns>The request LocalVariable</returns>
		public LocalVariable GetLocalVar(int index) 
		{
			return (LocalVariable) locals[index];
		}

		/// <summary>
		/// Override of the Equals Method
		/// </summary>
		/// <param name="obj">The object to compare with</param>
		/// <returns>True if the objects are a match</returns>
		public override bool Equals(object obj)
		{
			return base.Equals (obj);
		}

		/// <summary>
		/// Override of the GetHashCode method
		/// </summary>
		/// <returns>The HashCode</returns>
		public override int GetHashCode()
		{
			return base.GetHashCode ();
		}



		/// <summary>
		/// Useful to create a new LocalVariable
		/// </summary>
		/// <param name="type">The RType of the LocalVariable to create</param>
		/// <returns>The newly created LocalVariable</returns>
		public LocalVariable DefineLocalVariable(RType type) 
		{
			if (locals == null) 
			{
				locals = new ArrayList();
			}
			LocalVariable loc = new LocalVariable(this, type,locals.Count);
			locals.Add(loc);
			return loc;
		}

		/// <summary>
		/// Removes a local variable. 
		/// </summary>
		/// <remarks>
		/// All indexes higher than the one that is removed will shift one position down.
		/// </remarks>
		/// <param name="index">The index of the LocalVariable to remove</param>
		public void RemoveLocalVariable(int index) 
		{
			localsCcheckValidIndex(index);
			locals.RemoveAt(index);
			UpdateLocalsIndex();
		}

		/// <summary>
		/// Useful for updating the LocalVariables index
		/// </summary>
		public void UpdateLocalsIndex()
		{
			int i = 0;
			if (locals!=null)
				foreach (LocalVariable lV in locals)
				{
					lV.pos = i;
					i++;
				}
		}

		/* 
		 * I think that a set method is not necessary. Users can change the type 
		 * of an existing local variable directly in the LocalVariable class.
		 */
		//    /// <summary>
		//    /// Replaces the local variable at <code>index</code> by a new one with
		//    /// the given type
		//    /// </summary>
		//    /// <param name="index"></param>
		//    /// <param name="type"></param>
		//    /// <returns>The newly created local variable</returns>
		//    public LocalVariable SetLocalVariable(int index, RType type) {
		
		//      locals[index] = new LocalVariable(this, type);      
		//    }

		/// <summary>
		/// Checks if the given index is a valid position in the local variable
		/// table, throwing an exception if it isn't.
		/// </summary>
		/// <param name="index">The index to be checked</param>
		private void localsCcheckValidIndex(int index) 
		{
			if (locals == null || locals.Count <= index) 
			{
				throw new ArgumentException("Local variable index outside of range: " + index);
			}
		}


		/// <summary>
		/// Called during the construction of this object, to initialize the local
		/// varible array
		/// </summary>
		/// <param name="loadCtx">The context object where to find the Assmbl object</param>
		/// <param name="typeSig">The signature</param>
		private void initLocals(byte [] stream, StandAloneSigRow typeSig) 
		{      
			
			int pos = typeSig.Signature;
			MetadataResolver.SetBlobReadPosition(stream,out pos, pos);
			// TODO: Verify why this sometimes doesn't have the 0x07 value
			if (stream[pos]!=0x07) 
			{
				throw new InvalidProgramException("Unexpected value: " + stream[pos] + ". Expecting: " + 0x07);
			}
			
			pos += 1;
			// TODO: check if the Count value is correctly obtained
			int count = MetadataSignatures.ReadCompressedInt(stream, out pos, pos);
			if (count>0) 
			{
				locals = new ArrayList(count);
				for (int i = 0; i< count; i++) 
				{
					locals.Add(new LocalVariable(stream,out pos,pos,this,i));
				}
			}
		}  

		/// <summary>
		/// Returns the table with the exceptions defined in these method
		/// </summary>
		public ExceptionTable ExceptionTable 
		{
			get 
			{
				return this.excpTable;
			}
			set
			{
				this.excpTable = value;
			}
		}

		/// <summary>
		/// Creates an iterator for traversing the method body. 
		/// </summary>
		/// <returns>A <code>Code</code> instance.</returns>
		public Code GetCode() 
		{
			return this.methodCode;
		}

		/// <summary>
		/// Gets the method corresponding to this method body.
		/// </summary>
		public RMethodBase RMethod 
		{
			get 
			{
				return this.parent;
			}
			set 
			{
				this.parent = value;
			}
		}

		/// <summary>
		/// Gets or sets the Code object of the MethodBody
		/// </summary>
		public Code Code
		{
			get
			{
				return this.methodCode;
			}
			set
			{
				this.methodCode = value;
			}
		}

	}

	/// <summary>
	/// Represents the IL instruction stream of a method
	/// </summary>
	public class Code : IEnumerable 
	{
		/// <summary>
		/// The MethodBody owner of this Code object
		/// </summary>
		protected readonly MethodBody mBody;
		/// <summary>
		/// The IList of ILInstruction objects
		/// </summary>
		protected IList instructions = new ArrayList();

		/// <summary>
		/// The internal constructor of the object
		/// </summary>
		/// <param name="mb">The MethodBody that owns this Code object</param>
		internal Code(MethodBody mb)
		{
			this.instructions = new ArrayList();
			this.mBody = mb;
		}
	
		/// <summary>
		/// Internal Constructor of the Code object
		/// </summary>
		internal Code(MethodLoadContext loadCtx, MethodBody mb, MethodIL methodIL, Hashtable posToInst) 
		{
			if (mb == null)
				throw new InvalidProgramException();
			if (posToInst == null || (posToInst.Count!=0))
				throw new InvalidProgramException();
			this.mBody = mb;

			MDStream blobStream = loadCtx.asmLoader.MetaDataRoot.Streams["#Blob"] as MDStream;

			instructions = new ArrayList();

			/**********************************************************
			 * Phase 1. Decodes the OpCodes and partially creates the 
			 * corresponding Instruction objects. During this phase it is
			 * not possible to resolve branch instructions. That's done in 
			 * the second phase. To support the second phase, we create a table
			 * mapping the offset of each instruction (couting from the start
			 * of the method) to the corresponding Instruction object.
			 **********************************************************/
			// This array is filled during the first phase and processed during the second
			ArrayList unresolvedBranches = new ArrayList();
			ArrayList unresolvedSwitches = new ArrayList();

			OpCode op;
			int pos = 0;
			long methodLength = methodIL.CodeSize;
			int toAdvance; // how much should the cursor be advanced
			int toAdvanceSwitch=0; //in the case it's a switch
			int ilCur = 0;
			while (ilCur < methodLength) 
			{
      
				// Decodes the opcode.
				if (methodIL.ByteCode[ilCur] == 0xFE) 
				{
					op = OpCodesList.TwoByteCode[methodIL.ByteCode[ilCur+1]];
					toAdvance = 2;          
				} 
				else 
				{
					op = OpCodesList.OneByteCode[methodIL.ByteCode[ilCur]];
					toAdvance = 1;
				}
				ilCur += toAdvance;

				Instruction newInst;
				// Now the operand
				switch (op.OperandType) 
				{
						#region No operand, strings and InlinePhi
					case OperandType.InlineNone:          
						// No operand. 
						// We need to make a distinction here. Some opcodes have their
						// operands encoded in their name (e.g. ldarg.x, ldloc.x and stloc.x). 
						// We must treat them like if they were of type ILArgument or ILLocalVariable
						// so the user can access the operand as an object. So we need to treat those
						// special cases...
						if (OpCodes.Ldarg_0.Value <= op.Value && op.Value <= OpCodes.Ldarg_3.Value) 
						{
							// It's a ldarg instruction with the operand encoded
							int argIndex = op.Value - OpCodes.Ldarg_0.Value; 
							//BC_CHANGE: changed the argIndexValue in the next lines
							newInst = new ILArgument(op, mBody.RMethod.GetParameter(argIndex));
						} 
						else if (OpCodes.Ldloc_0.Value <= op.Value && op.Value <= OpCodes.Ldloc_3.Value) 
						{
							// It's a ldloc.x instruction, with the operand encoded.
							int locIndex = op.Value - OpCodes.Ldloc_0.Value;
							//BC_CHANGE: changed the locIndex Value in the next lines
							LocalVariable locVar = mBody.GetLocalVar(locIndex);
							newInst = new ILLocalVariable(op, locVar);
						} 
						else if (OpCodes.Stloc_0.Value <= op.Value && op.Value <= OpCodes.Stloc_3.Value) 
						{
							// It's a stloc instruction with the operand encoded.
							int locIndex = op.Value - OpCodes.Stloc_0.Value;
							//BC_CHANGE: changed the locIndex Value in the next lines
							LocalVariable locVar = mBody.GetLocalVar(locIndex);
							newInst = new ILLocalVariable(op, locVar);
						} 
						else 
						{
							// No operand...
							newInst = new ILNone(op);
						}
						toAdvance = 0;
						break;
                        
					case OperandType.InlineString:
						// The operand is a 32-bit metadata string token.
						int stringToken = (int)LEBitConverter.ToInt32(methodIL.ByteCode,ilCur);
						newInst = new ILString(op, loadCtx.paramResolver.ResolveString(stringToken));
						toAdvance = 4;
						break;
					case OperandType.InlinePhi:
						// This member supports the .NET Framework infrastructure 
						// and is not intended to be used directly from your code.          
						throw new Exception("Internal Error! Found InlinePhi opcode.");
					case OperandType.ShortInlineVar: 
						// The operand is an 8-bit integer containing the ordinal of a local variable or an argumenta.
					case OperandType.InlineVar:
						// The operand is 16-bit integer containing the ordinal of a local variable or an argument.
						// There is an error in the specification. This is not 2 bytes, but only one
						int paramIndex = methodIL.ByteCode[ilCur];
						if (op.Value == OpCodes.Starg_S.Value || 
							op.Value == OpCodes.Ldarg_S.Value || 
							op.Value == OpCodes.Ldarga_S.Value ||
							op.Value == OpCodes.Ldarg.Value ) 
						{ 
							// Obtain the parameter corresponding to the operand and create the instruction
//							if ((this.MethodBody.parent.Attributes & MethodAttributes.Static)!=0)
//								paramIndex++;
							RParameter myParam = mBody.RMethod.GetParameter(paramIndex);
							if (myParam==null)
								throw new NullReferenceException();
							newInst = new ILArgument(op, myParam);
						} 
						else if (op.Value == OpCodes.Ldloc_S.Value ||
							op.Value == OpCodes.Stloc_S.Value ||
							op.Value == OpCodes.Ldloca_S.Value) 
						{
							// It's a local variable. Obtain it and create the instruction
							newInst = new ILLocalVariable(op, mBody.GetLocalVar(paramIndex));
						} 
						else 
						{
							throw new InternalErrorException("Invalid value (" + op.Value + ") for ShortInlineVar");
						}
						toAdvance = 1;        
						break;        
						#endregion

						#region Fields, Methods, Types
					case OperandType.InlineField:
						//The operand is a 32-bit metadata token.
						int fieldToken = (int)LEBitConverter.ToInt32(methodIL.ByteCode,ilCur);
						newInst = new ILField(op, loadCtx.paramResolver.ResolveField(fieldToken));
						toAdvance = 4;
						break;
					case OperandType.InlineMethod:
						// The operand is a 32-bit metadata token.
						int methodToken = (int)LEBitConverter.ToInt32(methodIL.ByteCode,ilCur);
						newInst = new ILMethod(op, loadCtx.paramResolver.ResolveMethod(methodToken));
						toAdvance = 4;
						break;
					case OperandType.InlineType:
						// The operand is a 32-bit metadata token.
						int typeToken = (int)LEBitConverter.ToInt32(methodIL.ByteCode,ilCur);
						newInst = new ILType(op, loadCtx.paramResolver.ResolveType(typeToken));
						toAdvance = 4;
						break;
					case OperandType.InlineSig:
						// The operand is a 32-bit metadata signature token.
						int sigToken = (int)LEBitConverter.ToInt32(methodIL.ByteCode,ilCur);
						newInst = new ILSig(op, loadCtx.paramResolver.ResolveSig(sigToken));
						toAdvance = 4;
						break;

					case OperandType.InlineTok: // FIXME: Check this.
						// The operand is a FieldRef, MethodRef, or TypeRef token.

//						TypeDef : 0x02
//						TypeRef : 0x01
//						Field : 0x04
//						Method : 0x06
//						MemberRef : 0x0A

						int token = (int)LEBitConverter.ToInt32(methodIL.ByteCode,ilCur);
						RMember member = loadCtx.paramResolver.ResolveToken(token);
						if (member is RMethodBase) 
						{
							newInst = new ILToken(op, (RMethodBase) member);
						} 
						else if (member is RField) 
						{
							newInst = new ILToken(op, (RField) member);
						} 
						else if (member is RType) 
						{
							newInst = new ILToken(op, (RType) member);
						} 
						else 
						{
							throw new InternalErrorException("Bad member for token: " + member);
						}
						toAdvance = 4;
						break;
						#endregion

						#region Branches
					case OperandType.ShortInlineBrTarget:
						// The operand is an 8-bit integer branch target.
					case OperandType.InlineBrTarget:
						// The operand is a 32-bit integer branch target.            
						UnresolvedBranch uBranch = new UnresolvedBranch();
						// The target will be set during the second phase.
						newInst = uBranch.inst = new ILBranch(op, null);
						int relativeOffset;
						if (op.OperandType == OperandType.ShortInlineBrTarget) 
						{
							relativeOffset = (sbyte)methodIL.ByteCode[ilCur];
							toAdvance = 1;
						} 
						else 
						{
							relativeOffset = (int)LEBitConverter.ToInt32(methodIL.ByteCode,ilCur); 
							toAdvance = 4;
						}
            
						// offset 0 corresponds to the next instruction. So, must add the current position
						// plus the size of this instruction (opcode and operand)  plus the relativeoffset.
						uBranch.offset = pos + (op.Size + toAdvance) + relativeOffset;
						// Stores the branch for later resolution
						unresolvedBranches.Add(uBranch);
						break;
          
					case OperandType.InlineSwitch:
						// The operand is the 32-bit integer argument to a switch instruction.
						int nLabels = (int)LEBitConverter.ToInt32(methodIL.ByteCode,ilCur);
						ilCur += 4;
						toAdvance = 4 + 4*nLabels;
						toAdvanceSwitch = toAdvance;
						int nextInstStart = pos + toAdvance + op.Size;  
						toAdvance=0;

						// Creates a new unresolved switch. It will be resolved in the second phase
						UnresolvedSwitch uSwitch = new UnresolvedSwitch();
						newInst = uSwitch.inst = new ILSwitch(op, null);
						int[] offsets = new int[nLabels];
						for (int i = 0; i < nLabels; i++) 
						{
							// Extracts the branch offset...
							offsets[i] = nextInstStart + (int) LEBitConverter.ToInt32(methodIL.ByteCode,ilCur);
							// and advances
							ilCur +=4;
						}
						uSwitch.offsets = offsets;
						unresolvedSwitches.Add(uSwitch);
						break;
         
					case OperandType.ShortInlineI:
						// The operand is a 16-bit integer.
						// FIXME: Check this: I was thinking that 2 bytes were needed
						
						newInst = new ILOperandI2(op, methodIL.ByteCode[ilCur]);
						toAdvance = 1;
						break;
					case OperandType.InlineI:
						// The operand is a 32-bit integer.
						newInst = new ILOperandI4(op, (int)LEBitConverter.ToInt32(methodIL.ByteCode,ilCur));
						toAdvance = 4;
						break;
					case OperandType.InlineI8:
						// The operand is a 64-bit integer.
						newInst = new ILOperandI8(op, (long)LEBitConverter.ToInt64(methodIL.ByteCode,ilCur));
						toAdvance = 8;
						break;
					case OperandType.ShortInlineR:
						// The operand is a 32-bit IEEE floating point number.
						
						newInst = new ILOperandR4(op, BitConverter.ToSingle(methodIL.ByteCode,ilCur));
						toAdvance = 4;
						break;
					case OperandType.InlineR:
						// The operand is a 64-bit IEEE floating point number.
						newInst = new ILOperandR8(op, BitConverter.ToDouble(methodIL.ByteCode,ilCur));
						toAdvance = 8;
						break;

						#endregion
					default:
						throw new InternalErrorException("Invalid operand type: " + op.OperandType);
				}
				//TODO: verify if this advance is needed
				ilCur += toAdvance;

				// Adds the instruction to the final list of instructions
				instructions.Add(newInst);
				// Keeps the mapping between the position and the instruction
				posToInst[pos] = newInst;
				// Advances pos to the next instruction. toAdvance still holds the size of the operand
				//BC_CHANGE: from "pos = op.Size + toAdvance" to "pos = pos + op.Size + toAdvance"
				pos = pos + op.Size + toAdvance + toAdvanceSwitch;
				toAdvanceSwitch = 0;
			} // while


			/******************************************************
			 * Phase 2. Resolves the switch and branch instructions, and
			 * updates the targeters
			 ******************************************************/ 
			// Resolves the branch and switch instructions.      
			for (int i=0; i<unresolvedBranches.Count; i++) 
			{
				UnresolvedBranch uBranch = (UnresolvedBranch) unresolvedBranches[i];
				// Obtains the target instruction
				Instruction targetInst = (Instruction) posToInst[uBranch.offset];
				targetInst.AddTargeter(uBranch.inst);
				uBranch.inst.Target = targetInst;
			}

			for (int i=0; i<unresolvedSwitches.Count; i++) 
			{
				UnresolvedSwitch uSwitch = (UnresolvedSwitch) unresolvedSwitches[i];
				// This array will hold all the targets of the switch instruction
				Instruction[] targets = new Instruction[uSwitch.offsets.Length];
				// Links each target with its targeter
				for (int j=0; j<uSwitch.offsets.Length; j++) 
				{
					Instruction targetInst = (Instruction) posToInst[uSwitch.offsets[j]];
					targetInst.AddTargeter(uSwitch.inst);
					targets[j] = targetInst;
				}
				uSwitch.inst.targets = targets;
			}      
			UpdateInstructionIndex();
		}

		/// <summary>
		/// Method to clone Code objects, useful to copy components
		/// from an assembly to another
		/// </summary>
		/// <param name="mb">The MethodBody that owns the Code object to clone</param>
		/// <returns>The cloned Code object</returns>
		internal Code Clone(MethodBody mb)
		{
			Code cd = new Code(mb);
			
			ArrayList unresolvedBranches = new ArrayList();
			ArrayList unresolvedSwitches = new ArrayList();

			OpCode op;
			
			foreach (Instruction instr in this.instructions)
			{
      
				Instruction newInst;
				op = instr.OpCode;
				// Now the operand
				switch (op.OperandType) 
				{
						#region No operand, strings and InlinePhi
					case OperandType.InlineNone:          
						if (OpCodes.Ldarg_0.Value <= op.Value && op.Value <= OpCodes.Ldarg_3.Value) 
						{
							// It's a ldarg instruction with the operand encoded
							int argIndex = op.Value - OpCodes.Ldarg_0.Value; 
							newInst = new ILArgument(op, cd.mBody.RMethod.GetParameter(argIndex));

						} 
						else if (OpCodes.Ldloc_0.Value <= op.Value && op.Value <= OpCodes.Ldloc_3.Value) 
						{
							// It's a ldloc.x instruction, with the operand encoded.
							int locIndex = op.Value - OpCodes.Ldloc_0.Value;
							LocalVariable locVar = cd.mBody.GetLocalVar(locIndex);
							newInst = new ILLocalVariable(op, locVar);

						} 
						else if (OpCodes.Stloc_0.Value <= op.Value && op.Value <= OpCodes.Stloc_3.Value) 
						{
							// It's a stloc instruction with the operand encoded.
							int locIndex = op.Value - OpCodes.Stloc_0.Value;
							//BC_CHANGE: changed the locIndex Value in the next lines
							
							
							LocalVariable locVar = cd.mBody.GetLocalVar(locIndex);
							newInst = new ILLocalVariable(op, locVar);

						} 
						else 
						{
							// No operand...
							newInst = new ILNone(op);
						}
						break;
                        
					case OperandType.InlineString:
						newInst = new ILString(op, ((ILString)instr).String);
						break;

					case OperandType.InlinePhi:
						// This member supports the .NET Framework infrastructure 
						// and is not intended to be used directly from your code.          
						throw new Exception("Internal Error! Found InlinePhi opcode.");
						#endregion

						#region Locals and variables
					case OperandType.ShortInlineVar: 
						// The operand is an 8-bit integer containing the ordinal of a local variable or an argumenta.
					case OperandType.InlineVar:
						// The operand is 16-bit integer containing the ordinal of a local variable or an argument.
						// There is an error in the specification. This is not 2 bytes, but only one
						if (op.Value == OpCodes.Starg_S.Value || 
							op.Value == OpCodes.Ldarg_S.Value || 
							op.Value == OpCodes.Ldarga_S.Value ) 
						{ 
							// Obtain the parameter corresponding to the operand and create the instruction
							RParameter myParam = 
								cd.mBody.RMethod.GetParameter(((ILArgument)instr).Parameter.Sequence);
							if (myParam==null)
							{
								int pIndex = ((RMethodDef)cd.mBody.RMethod).AddParameter(
									new RParameter(
									((ILArgument)instr).Parameter.Sequence,
									(int)((ILArgument)instr).Parameter.Attributes,
									((ILArgument)instr).Parameter.Name,
									((ILArgument)instr).Parameter.ParameterType));
								myParam = 
									cd.mBody.RMethod.GetParameter(((ILArgument)instr).Parameter.Sequence);
							}
							newInst = new ILArgument(op, myParam);
						} 
						else if (op.Value == OpCodes.Ldloc_S.Value ||
							op.Value == OpCodes.Stloc_S.Value ||
							op.Value == OpCodes.Ldloca_S.Value) 
						{
							// It's a local variable. Obtain it and create the instruction
							newInst = new ILLocalVariable(op, cd.mBody.GetLocalVar(((ILLocalVariable)instr).Local.pos));
						} 
						else 
						{
							throw new InternalErrorException("Invalid value (" + op.Value + ") for ShortInlineVar");
						}
						break;        
						#endregion

						#region Fields, Methods, Types
					case OperandType.InlineField:
						//HERE
						RField iField = ((ILField)instr).Field;
						RField rf = null;
						if (iField.DeclaringType.FullName.Equals(this.MethodBody.parent.DeclaringType.FullName))
							rf = cd.mBody.RMethod.DeclaringType.GetField(((ILField)instr).Field.Name);
						else
						{
//							if (((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetReferencedAssembly(iField.DeclaringType.Assembly.Name.FullName)==null)
//								((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).AddReferencedAssembly(iField.DeclaringType.Assembly.Name.FullName);
//							RType declareType1 = 
//								((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).RModuleDef.GetType(iField.DeclaringType.NameWithNameSpace);
//							if (declareType1==null)
//								declareType1 =
//									((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetRTypeRef(iField.DeclaringType.NameWithNameSpace);						
							RType declareType1 = ((RTypeDef)cd.MethodBody.parent.DeclaringType).GetCorrectRType(iField.DeclaringType);
							rf = declareType1.GetField(((ILField)instr).Field.Name);
						}
						if (rf!=null)
							newInst = new ILField(op, rf);
						else
							newInst = new ILField(op, ((ILField)instr).Field);
						break;
					case OperandType.InlineMethod:
						RMethodBase iMeth = ((ILMethod)instr).Method;
						RMethodBase rm;

//						if (((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetReferencedAssembly(iMeth.DeclaringType.Assembly.Name.FullName)==null)
//							((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).AddReferencedAssembly(iMeth.DeclaringType.Assembly.Name.FullName);
//						RType declareType = 
//							((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).RModuleDef.GetType(iMeth.DeclaringType.NameWithNameSpace);
//						if (declareType==null)
//							declareType =
//							((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetRTypeRef(iMeth.DeclaringType.NameWithNameSpace);
						RType declareType = ((RTypeDef)cd.MethodBody.parent.DeclaringType).GetCorrectRType(iMeth.DeclaringType);
						if (iMeth is RMethod)
							rm = declareType.GetMethod(iMeth.Name,((RMethod)iMeth).ReturnType.ReturnType,iMeth.GetParameters());
						else if (iMeth is RConstructor)
							rm = declareType.GetConstructor(iMeth.GetParameters());
						else
							throw new InvalidOperationException("Bad Method type");
						if (rm!=null)
							newInst = new ILMethod(op, rm);
						else
							newInst = new ILMethod(op, iMeth);												
						break;
					case OperandType.InlineType:
						RType iType = ((ILType)instr).Type;
						RType rT = null;
						if (iType.FullName.Equals(this.MethodBody.parent.DeclaringType.FullName))
							rT = this.MethodBody.parent.DeclaringType;
						else
						{
//							if (((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetReferencedAssembly(iType.Assembly.Name.FullName)==null)
//								((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).AddReferencedAssembly(iType.Assembly.Name.FullName);
//							rT = 
//								((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).RModuleDef.GetType(iType.NameWithNameSpace);
//							if (rT==null)
//								rT =
//									((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetRTypeRef(iType.NameWithNameSpace);						
							rT = ((RTypeDef)cd.MethodBody.parent.DeclaringType).GetCorrectRType(iType);
						}
						newInst = new ILType(op, rT);
						break;
					case OperandType.InlineSig:
						newInst = new ILSig(op, null);
						break;
					case OperandType.InlineTok: 
						RMember member = ((ILToken)instr).Token;
						if (member is RMethodBase) 
						{
							RMethodBase iM = (RMethodBase)member;
							RMethodBase rmth;
//							if (((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetReferencedAssembly(iM.DeclaringType.Assembly.Name.FullName)==null)
//								((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).AddReferencedAssembly(iM.DeclaringType.Assembly.Name.FullName);
//							RType declareType3 = 
//								((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).RModuleDef.GetType(iM.DeclaringType.NameWithNameSpace);
//							if (declareType3==null)
//								declareType3 =
//									((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetRTypeRef(iM.DeclaringType.NameWithNameSpace);
							RType declareType3 = ((RTypeDef)cd.MethodBody.parent.DeclaringType).GetCorrectRType(iM.DeclaringType);
							if (iM is RMethod)
								rmth = declareType3.GetMethod(iM.Name,((RMethod)iM).ReturnType.ReturnType,iM.GetParameters());
							else if (iM is RConstructor)
								rmth = declareType3.GetConstructor(iM.GetParameters());
							else
								throw new InvalidOperationException("Bad method");
							if (rmth!=null)
								newInst = new ILToken(op, rmth);
							else
								newInst = new ILToken(op, iM);	
						} 
						else if (member is RField) 
						{
							//RField rfld = cd.mBody.RMethod.DeclaringType.GetField(((ILField)instr).Field.Name);
							RField iField2 = ((ILField)instr).Field;
							RField rfld = null;
							if (iField2.DeclaringType.FullName.Equals(this.MethodBody.parent.DeclaringType.FullName))
								rfld = cd.mBody.RMethod.DeclaringType.GetField(iField2.Name);
							else
							{
								//RAssembly declarAssembly1 = iField.DeclaringType.Assembly;
//								if (((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetReferencedAssembly(iField2.DeclaringType.Assembly.Name.FullName)==null)
//									((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).AddReferencedAssembly(iField2.DeclaringType.Assembly.Name.FullName);
//								RType declareType4 = 
//									((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).RModuleDef.GetType(iField2.DeclaringType.NameWithNameSpace);
//								if (declareType4==null)
//									declareType4 =
//										((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetRTypeRef(iField2.DeclaringType.NameWithNameSpace);						
								RType declareType4 = ((RTypeDef)cd.MethodBody.parent.DeclaringType).GetCorrectRType(iField2.DeclaringType);
								rfld = declareType4.GetField(iField2.Name);
							}
							if (rfld!=null)
								newInst = new ILToken(op, rfld);
							else
								newInst = new ILToken(op, (RField)member);
						} 
						else if (member is RType) 
						{
							RType iType2= (RType) member;
							RType rT2 = null;
							if (iType2.FullName.Equals(this.MethodBody.parent.DeclaringType.FullName))
								rT2 = this.MethodBody.parent.DeclaringType;
							else
							{
//								if (((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetReferencedAssembly(iType2.Assembly.Name.FullName)==null)
//									((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).AddReferencedAssembly(iType2.Assembly.Name.FullName);
//								rT2 = 
//									((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).RModuleDef.GetType(iType2.NameWithNameSpace);
//								if (rT2==null)
//									rT2 =
//										((RAssemblyDef)cd.MethodBody.parent.DeclaringType.Assembly).GetRTypeRef(iType2.NameWithNameSpace);						
								rT2 = ((RTypeDef)cd.MethodBody.parent.DeclaringType).GetCorrectRType(iType2);
							}
							newInst = new ILToken(op, rT2);
						} 
						else 
						{
							throw new InternalErrorException("Bad member for token: " + member);
						}
						break;
						#endregion

						#region Branches
					case OperandType.ShortInlineBrTarget:
						// The operand is an 8-bit integer branch target.
					case OperandType.InlineBrTarget:
						// The operand is a 32-bit integer branch target.            
						UnresolvedBranch uBranch = new UnresolvedBranch();
						// The target will be set during the second phase.
						newInst = uBranch.inst = new ILBranch(op, null);
						int index = ((ILBranch)instr).Target.index;
           
						uBranch.offset = index;
						// Stores the branch for later resolution
						unresolvedBranches.Add(uBranch);
						break;
					case OperandType.InlineSwitch:
						UnresolvedSwitch uSwitch = new UnresolvedSwitch();
						newInst = uSwitch.inst = new ILSwitch(op, null);
						int[] offsets = new int[((ILSwitch)instr).targets.Length];
						for (int i = 0; i < ((ILSwitch)instr).targets.Length; i++) 
						{
							offsets[i] = ((ILSwitch)instr).targets[i].index;
						}
						uSwitch.offsets = offsets;
						unresolvedSwitches.Add(uSwitch);
						break;
						#endregion

						#region Constants: Reals and integers          
					case OperandType.ShortInlineI:
						newInst = new ILOperandI2(op, ((ILOperandI2)instr).Operand);
						break;
					case OperandType.InlineI:
						// The operand is a 32-bit integer.
						newInst = new ILOperandI4(op, ((ILOperandI4)instr).Operand);
						break;
					case OperandType.InlineI8:
						// The operand is a 64-bit integer.
						newInst = new ILOperandI8(op, ((ILOperandI8)instr).Operand);
						break;
					case OperandType.ShortInlineR:
						// The operand is a 32-bit IEEE floating point number.
						newInst = new ILOperandR4(op, ((ILOperandR4)instr).Operand);
						break;
					case OperandType.InlineR:
						// The operand is a 64-bit IEEE floating point number.
						newInst = new ILOperandR8(op, ((ILOperandR8)instr).Operand);
						break;

						#endregion
					default:
						throw new InternalErrorException("Invalid operand type: " + op.OperandType);
				}
				cd.instructions.Add(newInst);
			} // while


			/******************************************************
			 * Phase 2. Resolves the switch and branch instructions, and
			 * updates the targeters
			 ******************************************************/ 
			// Resolves the branch and switch instructions.      
			for (int i=0; i<unresolvedBranches.Count; i++) 
			{
				UnresolvedBranch uBranch = (UnresolvedBranch) unresolvedBranches[i];
				// Obtains the target instruction
				Instruction targetInst = (Instruction) cd.instructions[uBranch.offset];
				targetInst.AddTargeter(uBranch.inst);
				uBranch.inst.Target = targetInst;
			}

			for (int i=0; i<unresolvedSwitches.Count; i++) 
			{
				UnresolvedSwitch uSwitch = (UnresolvedSwitch) unresolvedSwitches[i];
				// This array will hold all the targets of the switch instruction
				Instruction[] targets = new Instruction[uSwitch.offsets.Length];
				// Links each target with its targeter
				for (int j=0; j<uSwitch.offsets.Length; j++) 
				{
					Instruction targetInst = (Instruction) cd.instructions[uSwitch.offsets[j]];
					targetInst.AddTargeter(uSwitch.inst);
					targets[j] = targetInst;
				}
				uSwitch.inst.targets = targets;
			}    
			cd.UpdateInstructionIndex();
			return cd;
		}


		#region Writing
		/// <summary>
		/// 
		/// </summary>
		/// <param name="ig"></param>
		/// <param name="tResolver"></param>
		/// <param name="locals"></param>
		internal void Write(ILGenerator ig, TypeResolver tResolver,Hashtable locals) {

			/*****************************************
			 * Some preprocessing is required. 
			 ******************************************/
			// Marks each instruction with its corresponding index.
			UpdateInstructionIndex();
			// Now sorts the exception table, so that the exception blocks appear in the
			// order they must be emitted. This must be done only after the instructions
			// are correctly numbered.
			ExceptionTable exTable = mBody.excpTable;
			if (exTable != null) {
				exTable.Sort();

				// Insert "virtual" instructions to mark exception boundaries
				foreach (ExceptionBlock eb in exTable.excpBlocks) {
					if (eb.start != null) {
						instructions.Insert(instructions.IndexOf(eb.start), ILFactory.BeginExceptionBlock(eb));
					}
					if (eb.end != null) {
						int pos = instructions.IndexOf(eb.GetLastHandler().EndInstruction);
						instructions.Insert(pos, ILFactory.EndExceptionBlock(eb));
					}
					if (eb.FaultHandler != null) {
						instructions.Insert(instructions.IndexOf(eb.FaultHandler.StartInstruction),
							ILFactory.BeginFaultBlock(eb));
					}
					if (eb.FinallyHandler != null) {
						instructions.Insert(instructions.IndexOf(eb.FinallyHandler.StartInstruction),
								  ILFactory.BeginFinallyBlock(eb));
					}
					for (int j = 0; j < eb.GetCatchHandlerCount(); j++) {
						ExceptionHandler ch = eb.GetCatchHandler(j);
						instructions.Insert(instructions.IndexOf(ch.StartInstruction),
							ILFactory.BeginExceptionBlock(eb, ch));
					}
				}
			}

				// Prune out unnecessary leaves
			for (int i = 0; i < (instructions.Count - 1); i++) {
				Instruction a  = (Instruction)instructions[i];
				Instruction b =  (Instruction)instructions[i + 1];

				if ((a.OpCode.Value == OpCodes.Leave.Value) || (a.OpCode.Value == OpCodes.Leave_S.Value)) {

					// If this leave just points to the end of the block, it's not really necessary
					ILBranch leave = a as ILBranch;
					ILVirtualException virt = b as ILVirtualException;
					if (virt != null && virt.WillEmitLeave()) {
						if (leave.Target.Index == virt.LastInstructionIndex) {

							// We want to remove this leave.
							IEnumerator targeters = leave.GetTargeters();

							// Remove this leave from the targets list of targeters
							leave.Target.RemoveTargeter(leave);

							// Remove this leave from the instructions
							instructions.Remove(leave);
							--i;

							if (targeters != null) {
					while (targeters.MoveNext()) {
									ILBranch branch = targeters.Current as ILBranch;
									if (branch != null) {
										branch.Target = virt;
									} else {
										ILSwitch sw = targeters.Current as ILSwitch;
										if (sw != null) {
											Instruction[] switchTargets = sw.Targets;
											for (int k = 0; k < switchTargets.Length; k++) {
												if (switchTargets[k] == leave) {
													switchTargets[k] = virt;
												}
											}
											sw.Targets = switchTargets;
										}
									}
								}
							}
						}
					}
				}
			}
			
			

			WriteContext wCtx = new WriteContext(ig, tResolver, exTable,locals,this);
			foreach (Instruction i in instructions) 
			{
				// does it uses a local variable
				i.Write(wCtx);
			}
		}
		#endregion


		public Code ClonePublic()
		{
			return this.Clone( this.MethodBody );
		}


		/// <summary>
		/// 
		/// </summary>
		/// <param name="originalInstruction"></param>
		/// <param name="newInstructions"></param>
		public void Replace(MSIL.Instruction originalInstruction, ICollection newInstructions) 
		{
this.UpdateInstructionIndex();
int originalInstructionIndex = instructions.IndexOf(originalInstruction);
IEnumerator targeters = originalInstruction.GetTargeters();
if (originalInstruction is ILBranch)
((ILBranch)originalInstruction).Target.RemoveTargeter(originalInstruction);

instructions.Remove(originalInstruction);
bool firstExec = true;
foreach (Instruction i in newInstructions)
{
if (firstExec && targeters!=null)
{
while (targeters.MoveNext())
i.AddTargeter((ITargeter)targeters.Current);
firstExec = false;
}
this.Insert(originalInstructionIndex,i);
++originalInstructionIndex;
this.UpdateInstructionIndex();
}
}

		/// <summary>
		/// Marks each instruction with its sequence number. Starts on 0.
		/// </summary>
		public void UpdateInstructionIndex() 
		{      
			for (int i=0; i<instructions.Count; i++) 
			{
				((Instruction)instructions[i]).index = i;
			}
		}

		/// <summary>
		/// Gets the number of instructions
		/// </summary>
		public int InstructionCount 
		{
			get 
			{
				return instructions.Count;
			}
		}
    
		/// <summary>
		/// Gets the method body corresponding to this iterator.
		/// </summary>
		public MethodBody MethodBody 
		{
			get 
			{
				return this.mBody;
			}
		}

		/// <summary>
		/// Clears all instructions
		/// </summary>
		public void Clear()
		{
			instructions.Clear();
		}
 
		/// <summary>
		/// Adds many instructions at the end of the code.
		/// </summary>
		/// <param name="inst">The instructions collection to add</param>
		public void Add(IEnumerable instructionCollection) 
		{
			foreach (MSIL.Instruction instruction in instructionCollection)
				instructions.Add(instruction);
		}

		#region Auxiliar methods
		/*
		   * Not directly related with instructions. Manage labels,
		   * expcetions, scope, and so on. Currently, they are a copy
		   * of the corresponding methods on ILGenerator. But that will 
		   * change in the future.
		   */
		/// <summary>
		/// Declares a local variable of type <code>t</code>
		/// </summary>
		/// <param name="t"></param>
		/// <returns>The LocalBuilder object for this variable.</returns>
//		public LocalBuilder DeclareLocal(Type t) 
//		{

//		}

		/*
		   * Labels are necessary when the code generation must be done 
		   * sequencially. In that situation a label is used to mark a placeholder
		   * for an instruction that is yet to be defined. But with Rail,
		   * the IL doesn't have to be generated sequencially. One can create an
		   * instruction I that will be the target of other instructions, insert 
		   * those instructions in the stream pointing to I and only then insert 
		   * I.
		   */
		//    public Label DefineLabel() {
		
		//    }
		//
		//    public void MarkLabel(Label l) {
		
		//    }

//		public void BeginCatchBlock(Type exceptionType) 
//		{

//		}
//
//		public void BeginExceptFilterBlock() 
//		{

//		}
//
//		public Label BeginExceptionBlock() 
//		{

//		}
//
//		public void BeginFaultBlock() 
//		{

//		}
//
//		public void BeginFinallyBlock() 
//		{

//		}
//
//		public void BeginScope() 
//		{

//		}
//
//		public void EndExceptionBlock() 
//		{

//		}
//
//		public void EndScope() 
//		{

//		}
//       
//		public void UsingNamespace(String usingNamespace) 
//		{

//		}
		#endregion

		#region Enumerator support
		/* 
		   * An enumerator is read only and is only valid until 
		   * the underlying data is changed. In the current case,
		   * we can simply reuse the enumerator provided by the
		   * collection used to store the instructions.
		   */
		/// <summary>
		/// A Enumerator class for instrucions
		/// </summary>
		public class InstructionEnumerator : IEnumerator 
		{
			/// <summary>
			/// The IEnumerator
			/// </summary>
			private readonly IEnumerator arrayEnum;
			/// <summary>
			/// The public constructor for InstructionEnumerator
			/// </summary>
			/// <param name="arrayEnum">The IEnumerator original object</param>
			public InstructionEnumerator(IEnumerator arrayEnum) 
			{
				this.arrayEnum = arrayEnum;
			}

			/// <summary>
			/// Method to avance to the next enumerated object
			/// </summary>
			/// <returns>True if the avance was possible</returns>
			public bool MoveNext() 
			{
				return arrayEnum.MoveNext();
			}

			/// <summary>
			/// Method to reset the enumerator
			/// </summary>
			public void Reset() 
			{
				arrayEnum.Reset();
			}

			/// <summary>
			/// Method to access the current element in the enumeration
			/// </summary>
			public object Current 
			{
				get 
				{
					return arrayEnum.Current;
				}
			}
		}

		/// <summary>
		/// Method to retunr the IEnumerator associated object
		/// </summary>
		/// <returns>The required IEnumerator object</returns>
		public IEnumerator GetEnumerator() 
		{
			return new InstructionEnumerator(instructions.GetEnumerator());
		}
		#endregion

		#region Manipulating the instruction stream    
		/// <summary>
		/// Gets or sets the instruction at the given position
		/// </summary>
		public Instruction this[int pos] 
		{
			get 
			{
				return (Instruction)instructions[pos];
			}
			set 
			{
				instructions[pos] = value;
			}
		}

		/// <summary>
		/// Inserts an instruction before the one at the given position.
		/// </summary>
		/// <param name="pos">The position where to insert the new instruction</param>
		/// <param name="inst">The new instruction to insert</param>
		public void Insert(int pos, Instruction inst) 
		{
			instructions.Insert(pos, inst);
		}

		/// <summary>
		/// Adds an instruction at the end of the code.
		/// </summary>
		/// <param name="inst">The new instruction to add</param>
		public void Add(Instruction inst) 
		{
			instructions.Add(inst);
		}
    
		/// <summary>
		/// Removes the instruction at the given position.
		/// </summary>
		/// <remarks>
		/// Be careful, this might destroy the integrity of the instruction
		/// stream.
		/// </remarks>
		/// <param name="pos">The position of the instruction to remove</param>
		public void Remove(int pos) 
		{
			instructions.RemoveAt(pos);
		}
		#endregion
    
		/// <summary>
		/// Helper struct to hold the branch instructions for the 1 processing phase.
		/// </summary>
		private sealed class UnresolvedBranch 
		{
			/// <summary>
			/// The Branch instruction
			/// </summary>
			public ILBranch inst;
			/// <summary>
			/// The position of the instruction targeted by the branch instruction.
			/// </summary>
			public int offset;
		}

		/// <summary>
		/// Helper struct to hold the branch instructions for the 1 processing phase.
		/// </summary>
		private sealed class UnresolvedSwitch 
		{
			/// <summary>
			/// The switch instruction
			/// </summary>
			public ILSwitch inst;
			/// <summary>
			/// The offsets of the instructions targeted
			/// </summary>
			public int[] offsets;
		}
	}

	#region Exception handling

	/// <summary>
	/// Contains all structured exception blocks associated with a method
	/// </summary>
	public class ExceptionTable 
	{    
		// Small initial size, since methods normally have few exception blocks.
		/// <summary>
		/// The ArrayList of Exception blocks in the method
		/// </summary>
		internal ArrayList excpBlocks = new ArrayList(2);


		/// <summary>
		/// Internal constructor for a clean ExceptionTable
		/// </summary>
		internal ExceptionTable()
		{
			this.excpBlocks = new ArrayList();
		}
		/// <summary>
		/// Insternal constructor for ExceptionTable
		/// </summary>
		/// <param name="loadCtx">The holder of the Assmbl object that maps the assembly in memory</param>
		/// <param name="method">The MethodBody of the method where to create the ExceptionTable</param>
		/// <param name="sehStart">The position in memory where the ExceptionTable description starts</param>
		/// <param name="posToInst">HashTable that relates index positions with instructions</param>
		internal ExceptionTable(MethodLoadContext loadCtx, MethodBody method, System.IO.BinaryReader reader, Hashtable posToInst) 
		{      
			if (method == null)
				throw new InvalidProgramException();
			if (posToInst == null)
				throw new InvalidProgramException();

			byte startByte = reader.ReadByte();
			bool tinyHeader = !((startByte & 0x40) == 0x40);

			// TODO : Verify that this is an SEH Table!
			if ((startByte & 0x01) != 1) 
			{
				throw new InvalidProgramException("Invalid SEH table. Bad start flag!");
			}
			
			// Cursor to the start of the handling clauses
			
			int rows;
			if (tinyHeader) 
			{
				int size = (int)reader.ReadByte();
				reader.BaseStream.Position += 2;
				
				rows = size / 12;
			} 
			else 
			{
				reader.BaseStream.Position -= 1;
				uint size = ((uint)((int)reader.ReadInt32())) >> 8;
				
				rows = (int)(size / 24);
			}     

			if (rows == 0) 
			{
				return;
			}

			// a map between ProtectedBlock objects and the corresponding ExpectionBlocks
			Hashtable protectedBlocksTable = new Hashtable();

			// Reads the information about the SEH item.
			int flags;
			int tryOffset;
			int tryLength;
			int handlerOffset;
			int handlerLength;
			int classTokenOrFilterOffset;
			for (int i = 0; i<rows ; i++) 
			{
				if (tinyHeader) 
				{
					flags          = (short)reader.ReadInt16();
					tryOffset      = (short)reader.ReadInt16();
					tryLength      = reader.ReadByte();
					handlerOffset  = (short)reader.ReadInt16();
					handlerLength  = reader.ReadByte();
					classTokenOrFilterOffset = (int)reader.ReadInt32();
					
				} 
				else 
				{
					flags          = (int)reader.ReadInt32();
					tryOffset      = (int)reader.ReadInt32();
					tryLength      = (int)reader.ReadInt32();
					handlerOffset  = (int)reader.ReadInt32();
					handlerLength  = (int)reader.ReadInt32();
					classTokenOrFilterOffset = (int)reader.ReadInt32();
				}

				// TODO: I think that the offsets given in the exception handlers are
				// relative to the begining of the IL stream and not to the beginning of 
				// the method. But I am not sure.
				// Gets the instructions start and end instructions protected by this handler
				Instruction tryStart = (Instruction) posToInst[tryOffset];
				Instruction tryEnd = (Instruction) posToInst[tryOffset + tryLength];

				// There might be several handlers for the same protected block. So check
				// if this particular block was already created before.
				
				ProtectedBlock pBlock = new ProtectedBlock(tryStart, tryEnd);
				ExceptionBlock exBlock = (ExceptionBlock) protectedBlocksTable[pBlock];
				bool blockExists = true;
				if (exBlock == null) 
				{
					// This protected block was not found before. Create a new ExceptionBlock object
					exBlock = new ExceptionBlock(tryStart, tryEnd);
					protectedBlocksTable[pBlock] = exBlock;
					blockExists = false;
				}

				// Create the appropriate handler.
				// One ExceptionBlock has only one protected scope, identified by
				// the tryOffSet and tryLength. That information can be used to place
				// the new handler in the correct ExceptionBlock.
				Instruction handlerStart = (Instruction) posToInst[handlerOffset];
				Instruction handlerEnd = (Instruction) posToInst[handlerOffset + handlerLength];              
				switch (flags) 
				{
					case 0x00 : // A typed exception clause 
						RType catchType = loadCtx.paramResolver.ResolveType(classTokenOrFilterOffset);              
						// Creates a new handler and adds it to the exception table
						exBlock.AddCatchHandler(new ExceptionHandler(handlerStart, handlerEnd, catchType));
						break;
					case 0x1: // An exception filter and handler clause 
						throw new NotSupportedException("Found an exception filter and handler clause.");
					case 0x2: // A finally clause              
						// Creates a new handler            
						if (exBlock.FinallyHandler == null) 
						{
							exBlock.FinallyHandler = 
								new ExceptionHandler(ExceptionHandlerType.Finally, handlerStart, handlerEnd);
						} 
						else 
						{
							throw new InternalErrorException("Found two finally handlers for the same protected block");
						}
						break;

					case 0x04: // Fault clause (finally that is called on exception only)               
						// Creates a new handler              
						if (exBlock.FaultHandler == null) 
						{
							exBlock.FaultHandler = 
								new ExceptionHandler(ExceptionHandlerType.Fault, handlerStart, handlerEnd);
						} 
						else 
						{
							throw new InternalErrorException("Found two fault handlers for the same protected block");
						}
						break;
				}
				if (!blockExists)
					this.excpBlocks.Add(exBlock);
			}      
		}    


		/// <summary>
		/// Method useful for cloning ExceptionTable object from one assembly to another,
		/// or from one method to another
		/// </summary>
		/// <param name="method">The original method of the ExceptionTable to clone</param>
		/// <returns>The cloned ExceptionTable</returns>
		internal ExceptionTable Clone(MethodBody method) 
		{      
			if (method == null)
				throw new InvalidProgramException();
			ExceptionTable exTable = new ExceptionTable();

			if (this.excpBlocks==null)
				return exTable;
			if (this.excpBlocks.Count==0)
				return exTable;

			int rows = this.excpBlocks.Count;
			Code mCode = method.GetCode();

			
			// a map between ProtectedBlock objects and the corresponding ExpectionBlocks
			Hashtable protectedBlocksTable = new Hashtable();


			for (int i = 0; i<rows ; i++) 
			{
				Instruction tryStart = (Instruction) mCode[((ExceptionBlock)this.excpBlocks[i]).start.index];
				Instruction tryEnd = (Instruction) mCode[((ExceptionBlock)this.excpBlocks[i]).end.index];
				ProtectedBlock pBlock = new ProtectedBlock(tryStart, tryEnd);
				ExceptionBlock exBlock = (ExceptionBlock) protectedBlocksTable[pBlock];
				if (exBlock == null) 
				{
					exBlock = new ExceptionBlock(tryStart, tryEnd);
					protectedBlocksTable[pBlock] = exBlock;
				}

				foreach(ExceptionHandler eH in ((ExceptionBlock)this.excpBlocks[i]).catchHandlers)          
				{
					exBlock.AddCatchHandler(
						new ExceptionHandler(mCode[eH.start.index], 
						mCode[eH.end.index], 
						eH.ExceptionType));
				}

				if (((ExceptionBlock)this.excpBlocks[i]).finallyHandler != null) 
				{
					exBlock.finallyHandler = 
						new ExceptionHandler(ExceptionHandlerType.Finally, 
						mCode[((ExceptionBlock)this.excpBlocks[i]).finallyHandler.start.index], 
						mCode[((ExceptionBlock)this.excpBlocks[i]).finallyHandler.end.index]);
				} 
				
				if (((ExceptionBlock)this.excpBlocks[i]).faultHandler != null) 
				{
					exBlock.FaultHandler = 
						new ExceptionHandler(ExceptionHandlerType.Fault, 
						mCode[((ExceptionBlock)this.excpBlocks[i]).faultHandler.start.index], 
						mCode[((ExceptionBlock)this.excpBlocks[i]).faultHandler.end.index]);
				} 

				exTable.excpBlocks.Add(exBlock);
			}   
			return exTable;
		}    

		/// <summary>
		/// Number of exception blocks
		/// </summary>
		/// <returns>The number of Exception blocks in the method</returns>
		public int Count()
		{      
			return excpBlocks.Count;      
		}

		/// <summary>
		/// Adds a new exception block
		/// </summary>
		/// <param name="exBlock">The ExceptionBlock to add</param>
		public void Add(ExceptionBlock exBlock)
		{      
			excpBlocks.Add(exBlock);
		}

		/// <summary>
		/// Removes a structured exception block, deleting all the 
		/// handlers and corresponding code.
		/// </summary>
		/// <exception cref="ArgumentException">If the given position is not valid.</exception>
		/// <param name="pos">The index of the Exception block to remove</param>
		public void Remove(int pos) 
		{
			checkValidPosition(pos); 
			excpBlocks.RemoveAt(pos);
		}

		/// <summary>
		/// Private method useful for checking if a index is valid
		/// </summary>
		/// <param name="pos">the index positionj to check</param>
		private void checkValidPosition(int pos) 
		{
			if (excpBlocks.Count <= pos) 
			{
				throw new ArgumentException("Invalid exception block: " + pos);
			}
		}

		/// <summary>
		/// Gets or sets an exception block
		/// </summary>
		/// <exception cref="ArgumentException">If the given position is not valid.</exception>    
		public ExceptionBlock this[int pos] 
		{
			get 
			{
				checkValidPosition(pos);
				return (ExceptionBlock) excpBlocks[pos];
			}
			set 
			{
				checkValidPosition(pos);
				this[pos] = value;
			}
		}

		/// <summary>
		/// Sorts the exception blocks.
		/// </summary>
		public void Sort() 
		{
			excpBlocks.Sort();
		}

		/// <summary>
		/// Represents a region of code protected by a <code>try..catch</code> block
		/// </summary>
		private sealed class ProtectedBlock 
		{
			/// <summary>
			/// The instruction that marks the begining of the protected block
			/// </summary>
			public readonly Instruction start;
			/// <summary>
			/// The instruction that marks the end of the protected block
			/// </summary>
			public readonly Instruction end;

			/// <summary>
			/// Public constructor of the ProtectedBlock
			/// </summary>
			/// <param name="start">The start instruction</param>
			/// <param name="end">The end instruction</param>
			public ProtectedBlock(Instruction start, Instruction end) 
			{
				if (start == null)
					throw new InvalidProgramException();
				if (end == null)
					throw new InvalidProgramException();
				this.start = start;
				this.end = end;
			}

			/// <summary>
			/// Gets the HashCode for the object
			/// </summary>
			/// <returns>The hashcode value</returns>
			public override int GetHashCode() 
			{
				return start.GetHashCode() ^ end.GetHashCode();
			}
    
			/// <summary>
			/// <code>obj</code> must be an instance of <code>ProtectedBlock</code> and 
			/// must refer to the same <code>start</code> and <code>end</code> instructions
			/// instance as this object.
			/// </summary>
			/// <param name="obj">The object to compare</param>
			/// <returns>True if the objects are a match</returns>
			public override bool Equals(object obj) 
			{
				ProtectedBlock p = obj as ProtectedBlock;
				if (p==null) 
				{
					return false;
				} 
				else 
				{
					return Object.ReferenceEquals(p.end, end) && Object.ReferenceEquals(p.start, start);
				}
			}
		}
	}

	/// <summary>
	/// An exception block contains the following:
	/// <list type="bullet">
	/// <item>A protected scope, defined by a start and end instruction (inclusive).</item>
	/// <item>Several handlers, whose type can be: </item>
	/// </list>
	/// 
	/// <list type="bullet">
	///   <item>Catch - Activated based on the type of the exception. Each exception block can
	///   have any number of handlers, starting on 0.</item>
	///   <item>Finally - Executed always at the end of the exception block. Maximum one handler
	///   per block.</item>
	///   <item>Filtered - Activated based on user defined code</item>
	///   <item>Fault - Activated whenever an exception is raised.</item>
	/// </list>
	/// 
	/// <p>In the current version, only catch and finally handlers are supported.</p>
	/// </summary>
	public class ExceptionBlock : IComparable 
	{
		/// <summary>
		/// The instruction where the block starts
		/// </summary>
		internal Instruction start;
		/// <summary>
		/// The instruction where the block ends
		/// </summary>
		internal Instruction end;
		/// <summary>
		/// The Finally handler of the block
		/// </summary>
		internal ExceptionHandler finallyHandler;
		/// <summary>
		/// The Fault handler of the block
		/// </summary>
		internal ExceptionHandler faultHandler;
		// small value, normally there aren't many handlers for a protected block
		/// <summary>
		/// The ArrayList of handlers of the block
		/// </summary>
		internal ArrayList catchHandlers = new ArrayList(2); 

		/// <summary>
		/// Public constructor of the exception block
		/// </summary>
		/// <param name="start">The start instruction of the block</param>
		/// <param name="end">The last instruction of the block</param>
		public ExceptionBlock(Instruction start, Instruction end) 
		{
			this.start = start;
			this.end = end;
		}

		/// <summary>
		/// Gets or sets the first instruction that is protected by this block
		/// </summary>
		public Instruction StartOfProtectedScope 
		{
			get 
			{
				return start;
			}
			set 
			{
				this.start = value;
			}
		}

		/// <summary>
		/// Gets or sets the last instruction of this protected block
		/// </summary>
		public Instruction EndOfProtectedScope 
		{
			get 
			{
				return end;
			}
			set 
			{
				this.end = value;
			}
		}


		/// <summary>
		/// Gets or sets the finally handler. May be null.
		/// </summary>
		public ExceptionHandler FinallyHandler 
		{
			get 
			{
				return finallyHandler;
			}
			set 
			{
				this.finallyHandler = value;        
			}
		}

		/// <summary>
		/// Gets or sets the fault handler. May be null.
		/// </summary>
		public ExceptionHandler FaultHandler 
		{
			get 
			{
				return faultHandler;
			}
			set 
			{
				this.faultHandler = value;        
			}
		}

		#region Catch handlers
		/// <summary>
		/// Method to add a catch handler to the block
		/// </summary>
		/// <param name="handler">The ExceptionHandler to add</param>
		public void AddCatchHandler(ExceptionHandler handler) 
		{
			catchHandlers.Add(handler);      
		}

		/// <summary>
		/// Method to set an ExceptionHandler at a specific position in the
		/// catch handlers list
		/// 
		/// </summary>
		/// <exception cref="ArgumentException">Illegal position</exception>
		/// <param name="pos">The position where to add the handler</param>
		/// <param name="handler">The handler to add</param>
		public void SetCatchHandler(int pos, ExceptionHandler handler) 
		{
			checkValidPos(pos);
			catchHandlers[pos] = handler;      
		}

		/// <summary>
		/// Method to get the Catch Handler at a specif position
		/// </summary>
		/// <param name="pos">The position of the handler</param>
		/// <returns>The handler at position <code>pos</code></returns>
		public ExceptionHandler GetCatchHandler(int pos) 
		{
			checkValidPos(pos);
			return (ExceptionHandler) catchHandlers[pos];
		}

		/// <summary>
		/// Method to remove a Catch Handler at a specific position
		/// </summary>
		/// <exception cref="ArgumentException">Illegal position</exception>
		/// <param name="pos">The position of the handler to remove</param>
		public void RemoveCatchHandler(int pos) 
		{
			checkValidPos(pos);
			catchHandlers.RemoveAt(pos);
		}

		/// <summary>
		/// Get the number of CatchHandlers 
		/// </summary>
		/// <returns>Count of catch handlers</returns>
		public int GetCatchHandlerCount() 
		{      
			return catchHandlers.Count;      
		}

		/// <summary>
		/// Check if a position is valid inside the CatchHandlers ArrayList
		/// </summary>
		/// <param name="pos">The position to check</param>
		private void checkValidPos(int pos) 
		{
			if (catchHandlers.Count <= pos) 
			{
				throw new ArgumentException("Invalid position: " + pos);
			}
		}
		#endregion

		/// <summary>
		/// Returns the index of the last instruction of this exception block.
		/// Necessary to know when to call ILGenerator.EndExceptionBlock(). Assumes
		/// that the instructions are correctly numbered.
		/// </summary>
		/// <returns>The index of the last instruction of the block</returns>
		public int LastInstruction() 
		{      
			int max = -1;
			if (faultHandler != null) 
			{
				max = faultHandler.end.index;
			}
			if (finallyHandler != null) 
			{
				max = Math.Max(max, finallyHandler.end.index);
			}
			for (int i=0; i<catchHandlers.Count; i++) 
			{
				//BC_CHANGE
				
				int handlerEnd = ((ExceptionHandler)catchHandlers[i]).EndInstruction.index;
				max = Math.Max(max, handlerEnd);
			}
			//Debug.Assert(max > 0, "Found expcetion block with invalid or no handlers");      
			if (max <= 0)
				throw new InvalidProgramException("Found expcetion block with invalid or no handlers");
			return max;
		}
  
		public ExceptionHandler GetLastHandler()
		{
			ExceptionHandler handler = null;
			int max = -1;
			if (faultHandler != null) 
			{
				max = faultHandler.end.index;
				handler = faultHandler;
			}
			if (finallyHandler != null) 
			{
				if (max<finallyHandler.end.index)
					handler = finallyHandler;
				max = Math.Max(max, finallyHandler.end.index);
			}
			for (int i=0; i<catchHandlers.Count; i++) 
			{
				//BC_CHANGE
				
				int handlerEnd = ((ExceptionHandler)catchHandlers[i]).EndInstruction.index;
				if (max<handlerEnd)
					handler = ((ExceptionHandler)catchHandlers[i]);
				max = Math.Max(max, handlerEnd);
			}
			//Debug.Assert(max > 0, "Found expcetion block with invalid or no handlers");      
			if (max <= 0)
				throw new InvalidProgramException("Found expcetion block with invalid or no handlers");
			return handler;
		}
		/// <summary>
		/// Useful to compare exception blocks
		/// </summary>
		/// <remarks>
		/// An exception block is smaller than another if:
		/// 1. Its protected block starts before the protected block of the other.
		/// 2. If they start at the same instruction, them the one who finishes first
		///  is the bigger.
		///  
		///  Informally, this corresponds to sorting the blocks by the order that 
		///  they must be opened and closed using Reflection emit.
		///  
		///  This assumes that the instructions are correctly numbered.
		/// </remarks>
		/// <param name="obj">The Exception block to compare with</param>
		/// <returns>1 if this is greater that obj</returns>
		public int CompareTo(object obj) 
		{
			if (this.start.index == -1)
				throw new InvalidProgramException("Instructions must be numbered.");
			if (!(obj is ExceptionBlock)) 
			{
				throw new ArgumentException("Cannot compare with type " + obj.GetType());
			}
			ExceptionBlock other = (ExceptionBlock)obj;
			if (other == null) 
			{
				// Any object is greater than null. Definition.
				return 1;
			}
			if (start.index != other.start.index) 
			{
				return start.index.CompareTo(other.start.index);
			}

			// They have the same start. Check each one of them finishes first. 
			// That will be the bigger one
			return other.LastInstruction().CompareTo(this.LastInstruction());
		}
	}


	/// <summary>
	/// Represents an exception handler.
	/// </summary>
	public class ExceptionHandler 
	{
		/// <summary>
		/// The ExceptionHandlerType
		/// </summary>
		internal ExceptionHandlerType type;
		/// <summary>
		/// The start of the handler
		/// </summary>
		internal Instruction start;
		/// <summary>
		/// The last instruction on the handler
		/// </summary>
		internal Instruction end;
		/// <summary>
		/// The Exception type
		/// </summary>
		internal RType exType;

		/// <summary>
		/// Creates an exception handler for <code>fault</code> and <code>finally</code> 
		/// handlers.
		/// </summary>    
		/// <param name="type">The ExceptionHandlerType</param>
		/// <param name="start">The first instruction of this handler. May be initally null.</param>
		/// <param name="end">The last instruction of this handler. May be initially null.</param>    
		public ExceptionHandler(ExceptionHandlerType type, Instruction start, Instruction end) 
		{
			if (type == ExceptionHandlerType.Catch || type == ExceptionHandlerType.Filter) 
			{                
				throw new ArgumentException("Only fault and finally handler types are supported by this constructor.");
			}
			this.exType = null;
			this.type = type;
			this.start = start;
			this.end = end;
		}

		/// <summary>
		/// Creates a new catch handler
		/// </summary>
		/// <param name="catchStart">The star instruction of the handler</param>
		/// <param name="catchEnd">The last instruction of the handler</param>
		/// <param name="exType">The type od the exception</param>
		public ExceptionHandler(Instruction catchStart, Instruction catchEnd, RType exType) 
		{
			if (exType == null) 
			{
				throw new ArgumentException("Handler type cannot be null.");
			}
			this.type = ExceptionHandlerType.Catch;
			this.start = catchStart;
			this.end = catchEnd;
			this.exType = exType;
		}

		/// <summary>
		/// Gets the type of this handler
		/// </summary>
		public ExceptionHandlerType HandlerType 
		{
			get 
			{
				return type;
			}
		}

		/// <summary>
		/// Gets or sets the first instruction of this handler. Cannot be null.
		/// </summary>
		public Instruction StartInstruction 
		{
			get 
			{
				return start;
			}
			set 
			{
				this.start = value; 
			}
		}

		/// <summary>
		/// Gets or sets the last instruction of this handler. Cannot be null.
		/// </summary>
		//TODO: The last instruction must always be leave. Perhaps the setter should
		//check this
		public Instruction EndInstruction 
		{
			get 
			{
				return end;
			}
			set 
			{
				this.end = value; 
			}
		}

		/// <summary>
		/// Gets or sets the type of exception that fire this handler. May be null, if 
		/// the handler type is not a catch.
		/// </summary>
		public RType ExceptionType 
		{
			get 
			{
				return exType;
			}
			set 
			{
				this.exType = value; 
			}
		}
	}

	/// <summary>
	/// The type of an exception handler
	/// </summary>
	public enum ExceptionHandlerType 
	{
		/// <summary>
		/// An handler with a type check as filter
		/// </summary>
		Catch,
		/// <summary>
		/// An handler with user defined code as filter. Not supported.
		/// </summary>
		Filter,
		/// <summary>
		/// Executes always at the end of an exception block, with or without an exception.
		/// </summary>
		Finally,
		/// <summary>
		/// Executes whenever an exception is thrown. A finally that only executes with exceptions. 
		/// </summary>
		Fault
	}
	#endregion

	/// <summary>
	/// Represents a local variable of a method
	/// </summary>
	public class LocalVariable 
	{
		// I don't know if this information is required. If not, it should
		// be removed.
		/// <summary>
		/// DebugByRef value
		/// </summary>
		internal bool debugByRef;
		/// <summary>
		/// Debug Constraint value
		/// </summary>
		internal bool debugConstraint;
		/// <summary>
		/// The index of the local variable
		/// </summary>
		internal int pos;
		/// <summary>
		/// The MethodBody of the LocalVariable
		/// </summary>
		readonly private MethodBody methodBody;
		/// <summary>
		/// The type of the LocalVariable
		/// </summary>
		private RType varType;

		/// <summary>
		/// Creates a new LocalVariable
		/// </summary>
		/// <param name="method">The MethodBody where it should be created</param>
		/// <param name="type">The RType of the LocalVariable</param>
		/// <param name="pos">The index of the LocalVariable</param>
		internal LocalVariable(MethodBody method, RType type,int pos) 
		{
			this.methodBody = method;
			this.varType = type;
			this.pos = pos;
			debugByRef = false;
			debugConstraint = false;
		}
		/// <summary>
		/// Create a new LocalVariable
		/// </summary>
		internal LocalVariable(byte [] stream,out int index, int indexValue, MethodBody method, int pos) 
		{
			index = indexValue;
			this.pos = pos;
			this.methodBody = method;

			int sizeRead = index;
			debugConstraint = false;
			//The only value currently used is ELEMENT_TYPE_PINNED. There might be several constraints
			while (MetadataSignatures.ReadCompressedInt(stream, out sizeRead,sizeRead) 
				== (int)ElementType.Pinned) 
			{
				debugConstraint = true;
				index = sizeRead;
			}
			
			sizeRead = index;
			if (MetadataSignatures.ReadCompressedInt(stream, out sizeRead, sizeRead) 
				== (int)ElementType.ByRef) 
			{
				debugByRef = true;
				index = sizeRead;
			} 
			else 
			{
				debugByRef = false;
			}
			
			//BC_CHANGE: Changed the next lines
			LocalVarSig.LocalSigType st = new LocalVarSig.LocalSigType(stream,out index,index);
			// Obtains the type of the local variable
			if (method.parent.DeclaringType!=null)
			this.varType = RType.GetRType(st.type,
				(RAssemblyDef)method.parent.DeclaringType.Assembly);
			else
				this.varType = RType.GetRType(st.type,
					(RAssemblyDef)method.parent.GlobalSetModule.Assembly);
		}

		/// <summary>
		/// Gets the method to which this local variable belongs to.
		/// </summary>
		public MethodBody Method 
		{
			get 
			{
				return methodBody;
			}
		}

		/// <summary>
		/// Gets or sets the type of this local variable. Changing the type 
		/// without fixing the instruction stream will most likely corrupt it.
		/// </summary>
		public RType Type 
		{
			get 
			{
				return varType;
			}
			set 
			{
				this.varType = value;
			}
		}
	}  
}