/*
	The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); 
	you may not use this file except in compliance with the License. You may obtain a copy of the License at 
	http://www.mozilla.org/MPL/ 
	Software distributed under the License is distributed on an "AS IS" basis, 
	WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the specific language 
	governing rights and limitations under the License. 

	The Original Code is RAIL(Runtime Assembly Instrumentation Library) Alpha Version.
	The Initial Developer of the Original Code is University of Coimbra,
	Computer Science Department, Dependable Systems Group. Copyright (C) University of Coimbra. 
	All Rights Reserved.  
*/
using System;
using System.Reflection;
using System.Reflection.Emit;
using System.Collections;
using System.IO;
using Mono.PEToolkit;
using Mono.PEToolkit.Metadata;
using Rail.MSIL;
using Rail.Transformation;
using System.Threading;

namespace Rail.Reflect 
{
	/// <summary>
	/// 
	/// </summary>
	public abstract class RAssembly : CustomAttributeOwner 
	{    
		/// <summary>
		/// 
		/// </summary>
		protected RAssemblyName name;
		/// <summary>
		/// 
		/// </summary>
		protected string fileName;
		/// <summary>
		/// 
		/// </summary>
		protected RMethod entryPoint;
		/// <summary>
		/// 
		/// </summary>
		protected RModuleDef rModuleDef;
		/// <summary>
		/// 
		/// </summary>
		protected Hashtable rModuleRefs;
	
		/// <summary>
		/// Accesses the name of the assembly.
		/// </summary>
		/// <remarks>
		/// An assembly has always a name defined, so this property is never null.
		/// Also, it's not possible to copy the returned object into another assembly.
		/// </remarks>
		public virtual RAssemblyName Name	
		{
			get 
			{
				return this.name;
			}
			set 
			{
				this.name = value;
			}
		}    

		/// <summary>
		/// Accesses the filename of the assembly.
		/// </summary>
		public virtual string FileName	
		{
			get 
			{
				return this.fileName;
			}
			set 
			{
				this.fileName = value;
			}
		} 

		/// <summary>
		/// Gets the entry point of this assembly
		/// </summary>    
		/// <remarks>
		/// Calling the setter will result in an <code>NotSupportedException</code>
		/// </remarks>
		public virtual RMethod EntryPoint 
		{
			get 
			{
				return this.entryPoint;
			}
			set 
			{
				this.entryPoint = value;
			}
		}
    
		#region Module managing
		
		/// <summary>
		/// 
		/// </summary>
		/// <param name="name"></param>
		/// <returns></returns>
		public virtual RModule GetModule(String name) 
		{
			if (this.rModuleDef!=null)
				if (this.rModuleDef.FullyQualifiedName.Equals(name))
					return (RModule)this.rModuleDef;
			if (this.rModuleRefs!=null)
			{
				if (this.rModuleRefs.Contains(name))
					return (RModule)this.rModuleRefs[name];
			}
			return null;
		}
		
		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public virtual RModule [] GetModules() 
		{
			int rmrSize = 0;
			RModule [] rmrs = null;
			if (this.rModuleRefs!=null)
				rmrSize = this.rModuleRefs.Count;
			if (this.rModuleDef!=null)
				rmrs = new RModule[rmrSize+1];
			if (rmrs != null)
			{
				rmrs[0] = this.rModuleDef;
				if (this.rModuleRefs!=null && this.rModuleRefs.Count > 0) 
				{
					int i = 1;
					IDictionaryEnumerator e =
						this.rModuleRefs.GetEnumerator();
					while (e.MoveNext()) 
					{
						rmrs[i++] = (RModule)e.Value;
					}
				}
                                
			}
			return rmrs;
		}

		#endregion    
	}

  
	/// <summary>
	/// Represents a reference to an external assembly.
	/// </summary>
	/// <remarks>
	/// This type is read-only and serves only to support the buiding and 
	/// manipulation of the assembly loaded in an RAssemblyDef.
	/// 
	/// Some properties defined in the base class are supposed to be read-only
	/// here and read-write in <code>RAssemblyDef</code>. Unfortunately, 
	/// <code>C#</code> does not allow a property to be defined as read-only in a 
	/// base class and extended to read-write in a derived class. Therefore, 
	/// all the properties that are read-write in the <code>RAssemblyDef</code> 
	/// class are defined like that in <code>RAssembly</code> and in the 
	/// <code>RAssemblyRef</code> the setter method throws a 
	/// <code>NotSupportedException</code>.
	/// </remarks>
	public sealed class RAssemblyRef : RAssembly 
	{
		/// <summary>
		/// Creates a reference to the given assembly. 
		/// </summary>
		/// <param name="asm"></param>
		public RAssemblyRef(Assembly asm) 
		{
			this.name = new RAssemblyNameRef(asm.GetName());
		}

		/// <summary>
		/// Creates a reference to the given assembly. 
		/// </summary>
		/// <param name="asmName"></param>
		public RAssemblyRef(RAssemblyName asmName) 
		{
			this.name = asmName;
		}

		public RAssemblyRef(string fullName) 
		{
			this.name = new RAssemblyNameRef(fullName);
		}

		/// <summary>
		/// Creates a new empty assembly ref. To be used internally.
		/// </summary>
		internal RAssemblyRef() 
		{
		}    
	}


	/// <summary>
	/// Represents an assembly that is being manipulated.
	/// </summary>
	public sealed class RAssemblyDef : RAssembly, IVisitable, ICustomAttributeOwnerDef 
	{
		#region Fields
		/// <summary>
		/// 
		/// </summary>
		internal ArrayList rAssemblyRefs;
		/// <summary>
		/// 
		/// </summary>
		private ArrayList rTypeRefs;
		/// <summary>
		/// 
		/// </summary>
		internal ArrayList rFieldDefs;
		/// <summary>
		/// 
		/// </summary>
		internal ArrayList rMethodDefs;
		/// <summary>
		/// 
		/// </summary>
		internal ArrayList rParameterDefs;
		/// <summary>
		/// 
		/// </summary>
		internal ArrayList resources;

		internal AppDomain LoadDomain;

		//internal byte[] CurrentRawAssembly;
		
		#endregion

		#region Constructors
		/// <summary>
		/// A private constructor prevents external creation. Instances of this
		/// object are to be created using one of the factory methods.
		/// </summary>
		internal RAssemblyDef()
		{
			this.rAssemblyRefs = new ArrayList();
			this.rModuleDef = null;
			this.rModuleRefs = new Hashtable();
			this.rTypeRefs = new ArrayList();
			this.rFieldDefs = new ArrayList();
			this.rParameterDefs = new ArrayList();
		}

		/// <summary>
		/// Create a new RAssemblyDef
		/// </summary>
		/// <param name="name"></param>
		public RAssemblyDef(RAssemblyNameDef name)
		{
			this.name = name;
			this.rAssemblyRefs = new ArrayList();
			this.rModuleDef = null;
			this.rModuleRefs = new Hashtable();
			this.rTypeRefs = new ArrayList();
			this.rFieldDefs = new ArrayList();
			this.rParameterDefs = new ArrayList();
		}

		#endregion
 
		#region Dealing with Nested types
		/// <summary>
		/// Finds out the index to the TypeDef Table of the parent of the neste class
		/// </summary>
		/// <param name="TypeDefIndex">The index of the NestedClass in TypeDefTable (beggining in 0)</param>
		/// <param name="tabs">The TablesHeap object</param>
		///  the first entry is not valid and the real index starts at 1 not 0)</returns>
		internal int GetParentTypeBuilderIndex(int TypeDefIndex,TablesHeap tabs) 
		{
			int index = -1;
			for (int nestedClassIndex = 1; nestedClassIndex <= tabs[TableId.NestedClass].NumberOfRows; nestedClassIndex++) 
			{
				NestedClassRow tnc = tabs[TableId.NestedClass][nestedClassIndex] as NestedClassRow;
				if (tnc.NestedClass==TypeDefIndex) 
				{
					index = tnc.EnclosingClass - 2;
					break;
				}
			}
			return index;
		}

		/// <summary>
		/// Verifies if the Class is a NesteClass
		/// </summary>
		/// <param name="TypeDefIndex">The index of the NestedClass in TypeDefTable (beggining in 0)</param>
		/// <param name="tabs">The TablesHeap object</param>
		/// <returns>true if the class is nested, false otherwise</returns>
		internal bool IsNested(int TypeDefIndex,TablesHeap tabs) 
		{
			bool returnVal = false;
			if (tabs[TableId.NestedClass]!=null && tabs[TableId.NestedClass].NumberOfRows > 0)
				for (int i = 1; i <= tabs[TableId.NestedClass].NumberOfRows; i++) 
				{
					NestedClassRow tnc = tabs[TableId.NestedClass][i] as NestedClassRow;
					if (tnc.NestedClass==TypeDefIndex) 
					{
						returnVal = true;
						break;
					}
				}
			return returnVal;
		}
		#endregion

		#region RModuleRefList access
		/// <summary>
		/// Removes a RModuleRef object from the Assembly
		/// </summary>
		/// <param name="name">The name of the module</param>
		/// <returns>The removed module, null if the remove did not occurr</returns>
		public RModuleRef RemoveModuleRef(String name) 
		{
			RModuleRef r = (RModuleRef)rModuleRefs[name];
			rModuleRefs.Remove(name);
			return r;
		}

		/// <summary>
		/// Creates a new module ref, by copying the contents of another module. 
		/// The new module is forever associated with this assembly.
		///     
		/// TODO: Think about all the details
		/// 
		/// </summary>
		/// <param name="original">The module ref to be used as source</param>
		/// <param name="newName">The name for the new module</param>
		/// <returns>The added module</returns>
		public RModuleRef CopyModuleRef(RModuleRef original, String newName) 
		{
			RModuleRef r = new RModuleRef(this,original);
			r.Name = newName;
			//Attention, if already exists a moduleref with newName it will be replaced
			if (rModuleRefs.Contains(original.Name))
			{
				return null;
			}
			else if (rModuleRefs.Contains(r.Name))
			{
				return null;
			}
			else
			{
				rModuleRefs.Add(newName,r);
				return (RModuleRef)rModuleRefs[r.Name];
			}
		}
		
		/// <summary>
		/// Adds a new RTypeRef object to the RAssemblyDef
		/// </summary>
		/// <param name="t">The Type object</param>
		/// <returns>The added RTypeRef</returns>
		public RTypeRef DefineTypeRef(Type t)
		{
			if (t==null)
				throw new NullReferenceException("Null type");
//			RTypeRef rtr = new RTypeRef(t.Name,t.Namespace,this);
//			if (rtr==null)
//				throw new NullReferenceException("Bad RTypeRef Creation");
//			//rtr.CompleteRTypeRef(t,this);
//			if (this.rTypeRefs==null)
//				throw new NullReferenceException("Array of RTypeRef objects is null");
//			this.rTypeRefs.Add(rtr);
			return GetRTypeRef(t);
		}

		/// <summary>
		/// Creates a new empty module ref.
		/// </summary>
		/// <param name="name">Name of the new module</param>
		/// <param name="filename">Probably the same as name</param>
		/// <returns></returns>
		public RModuleRef DefineModuleRef(String name, String filename) 
		{      
			RModuleRef r = new RModuleRef(this,name,filename);
			if (r!=null)
			{
				rModuleRefs.Add(r.Name,r);
				return (RModuleRef)rModuleRefs[r.Name];
			}
			return null;
		} 
		#endregion

		#region RModuleDef managing
		/// <summary>
		/// 
		/// </summary>
		public RModuleDef RModuleDef 
		{
			get 
			{
				return this.rModuleDef;
			}
			set 
			{
				this.rModuleDef = value;
			}
		}

		/// <summary>
		/// Creates a RModuleDef.
		/// </summary>
		/// <param name="name">Name of the new module</param>
		/// <param name="filename">Probably the same as name</param>
		/// <returns></returns>
		public RModuleDef DefineModuleDef(String name, String filename) 
		{      
			RModuleDef r = new RModuleDef(this,name,filename);
			if (r!=null) 
			{
				this.RModuleDef = r;
				return this.RModuleDef;
			}
			else 
				return null;
		}
		#endregion

		#region RTypeRefList managing

		/// <summary>
		/// 
		/// </summary>
		/// <param name="type"></param>
		/// <returns></returns>
		public RTypeRef GetRTypeRef(string type)
		{
			RTypeRef rtr = null;
			if (this.rTypeRefs!=null)
				for (int i = 0; i < this.rTypeRefs.Count; i++)
				{
					rtr = (RTypeRef)this.rTypeRefs[i];

					if (rtr!=null)
						if (rtr.NameWithNameSpace!=null)
							if (rtr.NameWithNameSpace.Equals(type))
								return rtr;
				}
			Type t = Type.GetType(type);
			if (t==null && this.rAssemblyRefs!=null)
			{
				if (LoadDomain==null)
					LoadDomain = AppDomain.CreateDomain("RAILLoadDomain",AppDomain.CurrentDomain.Evidence);
				Assembly [] asms = LoadDomain.GetAssemblies();
				if (asms==null)
					asms = new Assembly[0];
				foreach (RAssemblyRef r in this.rAssemblyRefs)
				{
					Assembly asm = null;
					foreach (Assembly a in asms)
						if (a.FullName==r.Name.FullName ||
							a.GetName().Name.Equals(r.Name.Name))
						{
							asm = a;
							break;
						}
					
					try
					{
						if (asm==null)
							asm = LoadDomain.Load(r.Name.FullName);
					}
					catch (System.Exception)
					{
						//asm = LoadDomain.Load(r.Name.Name);
						AssemblyLoader objc = (AssemblyLoader) LoadDomain.CreateInstanceAndUnwrap(
								"Rail",
								"Rail.Reflect.AssemblyLoader");
						asm = objc.LoadAssemblyByPatialName(r.Name.Name);						
					}
					t = asm.GetType(type);
					if (t!=null)
						return GetRTypeRef(t);
				}
			}
			if (t==null)
				return null;
				//throw new NullReferenceException("Type not Found");
			return GetRTypeRef(t);
		}
		/// <summary>
		/// Get the RTypeRef from the Type object
		/// </summary>
		/// <param name="type">The Type to look for</param>
		/// <returns>The associated RTypeRef</returns>
		public RTypeRef GetRTypeRef(Type type)
		{
			if (type==null)
				return null;
			RTypeRef r = null;
			if (this.rTypeRefs!=null)
				for (int i = 0; i < this.rTypeRefs.Count; i++)
				{
					r = (RTypeRef)this.rTypeRefs[i];

					if (r!=null)
						if (r.NameWithNameSpace!=null)
							if (r.NameWithNameSpace.Equals(type.FullName))
								return r;
				}

			RTypeRef rtr = new RTypeRef(type.Name,type.Namespace,this);
			rtr.ReflectionType = type;
			if (type.DeclaringType!=null)
				rtr.DeclaringType = GetRTypeRef(type.DeclaringType);

			#region SOLO-HOT-FIX: ValueTypes and Array resolve
			// Since type do not have a assembly ref, search and add it.
			foreach(RAssemblyRef asmref in this.rAssemblyRefs)
			{
				if(asmref.Name.FullName.Equals(type.Assembly.FullName))
				{
					rtr.Assembly = asmref;
					break;
				}
			}
			#endregion SOLO-HOT-FIX: ValueTypes and Array resolve

			this.rTypeRefs.Add(rtr);
			return rtr;
		}
		
		/// <summary>
		/// Creates a new typeref in this assembly.
		/// </summary>
		/// <param name="r">The RTypeRef to add</param>
		public RTypeRef AddTypeRef(RTypeRef r) 
		{
			this.rTypeRefs.Add(r);
			return r;
		}

		/// <summary>
		/// Removes a type from this module.
		/// </summary>
		/// <param name="typeName">The name of the type to be removed</param>
		public void RemoveTypeRef(String typeName) 
		{
			if (this.rTypeRefs.Contains(typeName))
				this.rTypeRefs.Remove(typeName);     
		}    

		/// <summary>
		/// To get a RTypeRef from it's token value in other tables
		/// </summary>
		/// <param name="pos">The index in the table of the type in the token</param>
		/// <returns></returns>
		internal RTypeRef GetTypeByToken(int pos) 
		{
			
			Object rT = null;
			if (this.rTypeRefs==null) return null;
			if (this.rTypeRefs.Count>pos-1)
				rT = (Object)this.rTypeRefs[pos-1];
			return (RTypeRef)rT;

		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="name"></param>
		/// <returns></returns>
		public RTypeRef GetType(String name) 
		{
			if (this.rTypeRefs==null) return null;
			if (this.rTypeRefs.Count==0) return null;
			RTypeRef rt = null;
			bool found = false;
			for (int i=0;i<this.rTypeRefs.Count;i++)
			{
				rt = (RTypeRef)this.rTypeRefs[i];
				if (rt.NameWithNameSpace==name)
				{
					found = true;
					break;
				}
			}
			if (found)
				return rt;
			return null;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <returns></returns>
		public RTypeRef [] GetTypes()
		{
			if (this.rTypeRefs==null) return new RTypeRef[0];
			if (this.rTypeRefs.Count==0) return new RTypeRef[0];
			int cnt = this.rTypeRefs.Count;
			RTypeRef [] typeRefs = new RTypeRef[cnt];
			for (int i = 0; i < cnt; i++)
			{
				typeRefs[i] = (RTypeRef)this.rTypeRefs[i];
			}
			return typeRefs;
		}
		#endregion

		#region Loading, saving and running assemblies

		public static RAssemblyDef LoadAssembly( String fileName, bool bLoadSystemAssembly )
		{
			//			FileStream assemblyFileStream = null;
			//			byte [] currentRawAssembly = null;
			//
			//			try
			//			{
			//				assemblyFileStream = new FileStream(fileName, FileMode.Open);
			//				currentRawAssembly = new byte[(int) assemblyFileStream.Length];
			//				assemblyFileStream.Read(currentRawAssembly, 0, currentRawAssembly.Length);
			//				assemblyFileStream.Close();
			//			}
			//			catch (Exception)
			//			{
			//				Console.WriteLine("The assembly is in use by another application");
			//			}


			if (fileName.StartsWith("file:///")/*URI*/)
				fileName = fileName.Substring(8);

			using (Image pe = new Image (fileName)) 
			{
				pe.Open ();
				pe.ReadHeaders ();

				if (pe.IsCLI) 
				{
					//Get the tables heap
					TablesHeap tabs = pe.MetaDataRoot.TablesHeap;
					//Check if the tables heap is null
					if (tabs==null)
						throw new NullReferenceException("TablesHeap is null");
					//Get the Strings heap
					StringsHeap stringsHeap = (pe.MetadataRoot.Streams["#Strings"] as MDStream).Heap as StringsHeap;
					//Check if the strings heap is null
					if (stringsHeap==null)
						throw new NullReferenceException("StringsHeap is null");
					//Get the blob stream
					MDStream blobStream = pe.MetadataRoot.Streams["#Blob"] as MDStream;
					//Get GUID stream
					MDStream guidHeap = pe.MetaDataRoot.Streams["#GUID"] as MDStream;
					//Check if GUIDHeap exists
					if (guidHeap==null)
						throw new NullReferenceException("GUIDHeap is null");

					//Check if Assembly metadata table exists
					if (!tabs.HasAssembly)
						throw new InvalidProgramException("Assembly metadata table doesnt exists");
					
					//Get the Assembly table metadata row
					AssemblyRow assemblyRow = (AssemblyRow)tabs[TableId.Assembly][1];
					if (assemblyRow == null)
						throw new NullReferenceException();
					//Create the Assembly name
					RAssemblyNameDef assemblyNameDef = new RAssemblyNameDef();
					//Set the name read from the assembly table
					assemblyNameDef.Name = stringsHeap[assemblyRow.Name];
					//Create a RVersion 
					assemblyNameDef.Version = new RVersion(
						assemblyRow.MajorVersion,
						assemblyRow.MinorVersion,
						assemblyRow.BuildNumber,
						assemblyRow.RevisionNumber);
					//TODO: Not able to convert the string "neutral" that is read from the assembly
					//to a correct int or string CultureInfo object. Because of this the culture is always set 
					//to InvariantCulture
					assemblyNameDef.CultureInfo = System.Globalization.CultureInfo.InvariantCulture;
					//Set the flags for the assembly
					assemblyNameDef.Flags = (AssemblyNameFlags)assemblyRow.Flags;
					//Set the publicKeyToken
					int blobReadPos = 0;
					if (assemblyRow.PublicKey != 0 ) 
					{
						blobReadPos = assemblyRow.PublicKey;
						MetadataResolver.SetBlobReadPosition(blobStream.RawData,out blobReadPos,blobReadPos);
						byte [] pkt = new byte[8];
						pkt[0] = blobStream.RawData[blobReadPos];
						pkt[1] = blobStream.RawData[blobReadPos + 1];
						pkt[2] = blobStream.RawData[blobReadPos + 2];
						pkt[3] = blobStream.RawData[blobReadPos + 3];
						pkt[4] = blobStream.RawData[blobReadPos + 4];
						pkt[5] = blobStream.RawData[blobReadPos + 5];
						pkt[6] = blobStream.RawData[blobReadPos + 6];
						pkt[7] = blobStream.RawData[blobReadPos + 7];
						assemblyNameDef.PublicKeyToken = pkt;
					}

					//Create an RAssemblyDef instance
					RAssemblyDef assemblyDef = new RAssemblyDef();
					//					assemblyDef.CurrentRawAssembly = currentRawAssembly;
					//TODO: Eliminate this GeralLoader
					
					//Set the name for the RAssemblyDef
					assemblyDef.Name = assemblyNameDef;
					//TODO : Set the Filename
					//assemblyDef.fileName = -----
					assemblyDef.FileName = Path.GetFullPath(fileName);
					//**************//
					//Set up the list of RAssemblyRefs
					
					if (tabs[TableId.AssemblyRef]!=null && tabs[TableId.AssemblyRef].NumberOfRows > 0) 
					{
						RAssemblyRef assemblyRef = null;
						for (int i = 1; i <= tabs[TableId.AssemblyRef].NumberOfRows; i++) 
						{
							AssemblyRefRow metaAssemblyRef = (AssemblyRefRow)tabs[TableId.AssemblyRef][i];
							string assemblyNameRefName = stringsHeap[metaAssemblyRef.Name];
							//Create a RVersion 
							RVersion assemblyNameRefVersion = new RVersion(
								metaAssemblyRef.MajorVersion,
								metaAssemblyRef.MinorVersion,
								metaAssemblyRef.BuildNumber,
								metaAssemblyRef.RevisionNumber);
							//TODO: Not able to convert the string "neutral" that is read from the assembly
							//to a correct int or string CultureInfo object. Because of this the culture is always set 
							//to InvariantCulture
							System.Globalization.CultureInfo assemblyNameRefCultureInfo = System.Globalization.CultureInfo.InvariantCulture;
							//Set the flags for the assembly
							AssemblyNameFlags assemblyNameRefFlags = (AssemblyNameFlags)metaAssemblyRef.Flags;
							//Set the publicKeyToken
							byte [] pkt = null;
							if (metaAssemblyRef.PublicKeyOrToken != 0 ) 
							{
								blobReadPos = metaAssemblyRef.PublicKeyOrToken;
								MetadataResolver.SetBlobReadPosition(blobStream.RawData,out blobReadPos,blobReadPos);
								pkt = new byte[8];
								pkt[0] = blobStream.RawData[blobReadPos];
								pkt[1] = blobStream.RawData[blobReadPos + 1];
								pkt[2] = blobStream.RawData[blobReadPos + 2];
								pkt[3] = blobStream.RawData[blobReadPos + 3];
								pkt[4] = blobStream.RawData[blobReadPos + 4];
								pkt[5] = blobStream.RawData[blobReadPos + 5];
								pkt[6] = blobStream.RawData[blobReadPos + 6];
								pkt[7] = blobStream.RawData[blobReadPos + 7];
							}
							RAssemblyNameRef assemblyNameRef = new RAssemblyNameRef(
								assemblyNameRefName,
								assemblyNameRefVersion,
								assemblyNameRefCultureInfo,
								assemblyNameRefFlags,
								pkt);
							//Create an instance of RAssemblyRef
							assemblyRef = new RAssemblyRef(assemblyNameRef);
							assemblyDef.rAssemblyRefs.Add(assemblyRef);
						}
					}
					//******************************//
					//Set the Modules of the Assembly

					//Get the Module table from the metadata
					
					if (tabs[TableId.Module]!=null && tabs[TableId.Module].NumberOfRows > 0)
					{
						ModuleRow mr = (ModuleRow)tabs[TableId.Module][1];
						//Define a new ModuleDef by the entry in metadata Module table
						RModuleDef r = assemblyDef.DefineModuleDef(stringsHeap[mr.Name],stringsHeap[mr.Name]);
					}
					else
						throw new NullReferenceException("Metadata Module table is empty");

					//Get the ModuleRef table from the metadata to add the RModuleRefs to the 
					//RAssemblyDef object, each entry in the ModuleRef table is a module that belongs to
					//the assembly but isn't the one with the manifest
					//ModuleRef table has the reference to the modules defined in the assembly
					///but that do not hold the manifest data, each entry as a match in the file table
					///TODO: Support to assemblies with more than one module will be implemented
					///in Version 2 of RAIL
					
					if (tabs[TableId.ModuleRef]!=null && tabs[TableId.ModuleRef].NumberOfRows > 0)
						for (int i = 1; i <= tabs[TableId.ModuleRef].NumberOfRows; i++) 
						{
							ModuleRefRow t = (ModuleRefRow)tabs[TableId.ModuleRef][i];
							//Define a new ModuleDef by the entry in metadata Module table
							RModuleRef r = assemblyDef.DefineModuleRef(stringsHeap[t.Name],stringsHeap[t.Name]);
							if (r == null)
								throw new NullReferenceException("Bad RModuleRef creation");
						}
					//*********************************//
					//Read Types and create RTypeDef
		
					
					if (tabs[TableId.TypeDef]!=null && tabs[TableId.TypeDef].NumberOfRows > 0)
						//The first entry in TypeDef table is invalid then i starts at 2
						for (int i = 2; i <= tabs[TableId.TypeDef].NumberOfRows; i++) 
						{
							TypeDefRow t = (TypeDefRow)tabs[TableId.TypeDef][i];
							if (t == null)
								throw new AssemblyReadException("BAD TypeDef row reading");
							//TODO : How can we know to which Module the type belongs??? Probably it
							//depends of the file where it's written
							//We will use de module with the manifest because it's the only one that is loaded in V.1

							//Create the RTypeDef
							RTypeDef r = assemblyDef.rModuleDef.DefineType(stringsHeap[t.Name],stringsHeap[t.Namespace]);
							//Set the TypeAttributes
							r.Attributes = (TypeAttributes)t.Flags;
				
							//Nested types
							if (assemblyDef.IsNested(i,tabs))
							{	
								int index = assemblyDef.GetParentTypeBuilderIndex(i,tabs);
								r.DeclaringType = assemblyDef.rModuleDef.GetTypeAt(index);
							}
						}
					else
						throw new AssemblyReadException("Not able to obtain type definitions from the assembly");

					//Read TypeRefs and create RTypeRefs
					
					if (tabs[TableId.TypeRef]!=null && tabs[TableId.TypeRef].NumberOfRows > 0)
						for (int i = 1; i <= tabs[TableId.TypeRef].NumberOfRows; i++) 
						{
							TypeRefRow t = (TypeRefRow)tabs[TableId.TypeRef][i];
							if (t == null)
								throw new AssemblyReadException("BAD TypeRef row reading");
							//Create the RTypeRef
							RTypeRef r = new RTypeRef(stringsHeap[t.Name],
								stringsHeap[t.Namespace],
								assemblyDef,
								t.ResolutionScope);
							if (assemblyDef.AddTypeRef(r)==null)
								throw new NullReferenceException("BAD RTypeRef creation");
						}
					else
						if ( !bLoadSystemAssembly )
							throw new AssemblyReadException("Not able to obtain any type ref from the assembly");

					//Set the base type for TypeDefs
					if (tabs[TableId.TypeDef]!=null && tabs[TableId.TypeDef].NumberOfRows > 0)
						//The first entry in TypeDef table is invalid then i starts at 1
						for (int i = 2; i <= tabs[TableId.TypeDef].NumberOfRows; i++) 
						{
							TypeDefRow t = (TypeDefRow)tabs[TableId.TypeDef][i];
							if (t == null)
								throw new AssemblyReadException("BAD TypeDef row reading");
							if (t.Extends.RID == 0)
								continue;
							RType ExtendsType;
							if (t.Extends.Type == TokenType.TypeDef) 
							{
								ExtendsType = (RTypeDef)assemblyDef.rModuleDef.GetTypeByToken(t.Extends.RID);
							}
							else if (t.Extends.Type == TokenType.TypeRef) 
							{ //Token = 1, it's a TypeRef
								ExtendsType = assemblyDef.GetTypeByToken(t.Extends.RID);
							}
							else 
							{ //it's a TypeSpec 
								//TODO: Deal width TypeSpec references
								throw new Exception("ExtendsType not found");
							}
							//Set the base type for the RTypeDef
							assemblyDef.rModuleDef.GetTypeAt(i-2).BaseType = ExtendsType;
						}
					else
						throw new AssemblyReadException("Not able to obtain type definitions from the assembly");


					//Set interfaces for types
					
					if (tabs[TableId.InterfaceImpl]!=null && tabs[TableId.InterfaceImpl].NumberOfRows > 0)
						for (int i = 1 ;i <= tabs[TableId.InterfaceImpl].NumberOfRows;i++) 
						{
							InterfaceImplRow MetaInterfaceImpl = (InterfaceImplRow)tabs[TableId.InterfaceImpl][i];
							int classIndex = -1;
							int interfaceIndex = -1;
							if (tabs[TableId.TypeDef].NumberOfRows > 65536) 
							{
								classIndex = MetaInterfaceImpl.Class & 0x00FFFFFF;
							}
							else
							{
								classIndex = MetaInterfaceImpl.Class & 0x00000FFF;
							}

							interfaceIndex = MetaInterfaceImpl.Interface.RID;
							
							switch(MetaInterfaceImpl.Interface.Type)
							{
								case TokenType.TypeDef:
									RTypeDef rt = assemblyDef.rModuleDef.GetTypeByToken(interfaceIndex);
									assemblyDef.rModuleDef.GetTypeByToken(classIndex).AddInterface(rt);
									break;
								case TokenType.TypeRef:
									RTypeRef r = assemblyDef.GetTypeByToken(interfaceIndex);
									assemblyDef.rModuleDef.GetTypeByToken(classIndex).AddInterface(r);
									break;
								default:
									throw new Exception("Bad Interface token");
							}				
						}


					//******************************************//
					//Set the layout of the classes
					//******************************************//
					if (tabs[TableId.ClassLayout]!=null && 
						tabs[TableId.ClassLayout].NumberOfRows > 0)
					{
						for (int i = 1; i <= tabs[TableId.ClassLayout].NumberOfRows; i++)
						{
							ClassLayoutRow clr = tabs[TableId.ClassLayout][i] as ClassLayoutRow;
							RTypeDef rtd = assemblyDef.RModuleDef.GetTypeByToken(clr.Parent) as RTypeDef;
							rtd.PackingSize = (PackingSize)clr.PackingSize;
							rtd.TypeSize = clr.Size;
						}
						
					}


					//**************************************//
					//Read Fields and build RFields
					
					if (tabs[TableId.Field]!=null && 
						tabs[TableId.Field].NumberOfRows > 0) 
					{
						for (int i = 2; i <= tabs[TableId.TypeDef].NumberOfRows; i++) 
						{
							TypeDefRow t = (TypeDefRow)tabs[TableId.TypeDef][i];
							if (t == null)
								throw new AssemblyReadException("BAD TypeDef row reading when creatinf RFields");
							if (t.FieldList > 0)
							{
								int fieldListStart = t.FieldList;
								int fieldListEnd = t.FieldList + 1;
								int j = 1;
								if (i+j <= tabs[TableId.TypeDef].NumberOfRows)
								{
									t = (TypeDefRow)tabs[TableId.TypeDef][i+j];
									while (t.FieldList<1)
									{
										++j;
										t = (TypeDefRow)tabs[TableId.TypeDef][i+j];
									}
									fieldListEnd = t.FieldList;
								}
								else
								{
									//TODO: Check this
									fieldListEnd = tabs[TableId.Field].NumberOfRows + 1;
								}
								RTypeDef r = assemblyDef.rModuleDef.GetTypeAt(i-2);
								if (fieldListStart!=0)
									for (int x = fieldListStart; x < fieldListEnd; x++) 
									{
										FieldRow metaField = (FieldRow)tabs[TableId.Field][x];
										string fieldName = stringsHeap[metaField.Name];
										RType sigRType = null;
										FieldSig fs = new FieldSig(blobStream.RawData,metaField.Signature);
										if (fs==null)
											throw new NullReferenceException("FieldSig not valid");
										if (fs.Type==null)
											throw new NullReferenceException("Not valid SigType in FieldSig");
										sigRType = RType.GetRType(fs.Type,assemblyDef);
										RFieldDef rfd = r.DefineField(fieldName,sigRType);
										rfd.Attributes = (FieldAttributes)metaField.Flags;
										//Set the fields values

										if ((((FieldAttributes)metaField.Flags) & FieldAttributes.HasDefault) !=0) 
										{
											rfd.Attributes = rfd.Attributes | FieldAttributes.HasDefault;
											if (tabs[TableId.Constant]!=null && tabs[TableId.Constant].NumberOfRows > 0)
												for (int w=1;w<=tabs[TableId.Constant].NumberOfRows;w++) 
												{
													ConstantRow metaConstant = (ConstantRow)tabs[TableId.Constant][w];
													//TODO: verifiy if this is correct


													if (metaConstant.Parent.Type == TokenType.FieldDef)
														if (metaConstant.Parent.RID == x) 
														{
															int readPosition = metaConstant.Value;
															int sizeToRead = MetadataResolver.GetBlobSizeToRead(blobStream.RawData,out readPosition,readPosition);
															switch (metaConstant.Type) 
															{
																case ElementType.Boolean: 
																	rfd.Value = Convert.ToBoolean(blobStream.RawData[readPosition]);
																	break;
																	
																case ElementType.Char: 
																	rfd.Value = (char)(blobStream.RawData[readPosition]);
																	break;
																	
																case ElementType.I1: 
																	rfd.Value = (sbyte)(blobStream.RawData[readPosition]);
																	break;
																	
																case ElementType.I2: 
																	rfd.Value = LEBitConverter.ToInt16(blobStream.RawData,readPosition);
																	break;
																	
																case ElementType.I4:  
																	rfd.Value = LEBitConverter.ToInt32(blobStream.RawData,readPosition);
																	break;
																	
																case ElementType.I8:  
																	rfd.Value = LEBitConverter.ToInt64(blobStream.RawData,readPosition);
																	break;
																	
																case ElementType.R4:  
																	rfd.Value = BitConverter.ToSingle(blobStream.RawData,readPosition);
																	break;
																	
																case ElementType.R8:   
																	rfd.Value = BitConverter.ToDouble(blobStream.RawData,readPosition);
																	break;
																	
																case ElementType.String:    
																	rfd.Value = BitConverter.ToString(blobStream.RawData,readPosition,sizeToRead);
																	break;
																	
																case ElementType.Class: 
																	rfd.Value = (int)0;
																	break;
																case ElementType.U1:
																	rfd.Value = (byte)blobStream.RawData[readPosition];
																	break;
																default: 
																	rfd.Value = null;
																	break;
															}	
															
															break;
														}
												}
										}
										assemblyDef.rFieldDefs.Add(rfd);
									}
							}
						}
					}

					//*************************************************//
					// Read the FieldRVA table and fields initial values
					//*************************************************//
					
					if (tabs[TableId.FieldRVA]!=null)
					{
						for (int i = 1; i <= tabs[TableId.FieldRVA].NumberOfRows; i++)
						{
							//throw new NotSupportedException();
						}
					}

					//****************************************************//
					// Read the FieldLayout table and fields initial offset
					//****************************************************//
					
					if (tabs[TableId.FieldLayout]!=null)
					{
						for (int i = 1; i <= tabs[TableId.FieldLayout].NumberOfRows; i++)
						{
							FieldLayoutRow flr = tabs[TableId.FieldLayout][i] as FieldLayoutRow;
							RFieldDef rfd = assemblyDef.rFieldDefs[flr.Field - 1] as RFieldDef;
							rfd.Offset = flr.Offset;
						}
					}

				
					//****************************************///
					//First pass through the methods list
					//Create the RMethods
					//****************************************///
					assemblyDef.rMethodDefs = new ArrayList();
					if (tabs[TableId.Method]==null)
						throw new NullReferenceException("Wasn't possible to read the methods table");
					if (tabs[TableId.Method].NumberOfRows < 1)
						throw new AssemblyReadException("BAD methods metadata reading");

					int typesTableIndex = 1; //TODO: Check if this index is correct
					int methodsArrayIndex = 0;
					MethodRow metaMethod = null;
					TypeDefRow metaTypeDef = null;
					RTypeDef currentRTypeDef = null;
					int ImplMapTableIndex = 1;
					bool isPInvoke = false;

					AppDomain AppDomainForTypeDefs = AppDomain.CreateDomain("TempDForRRtypes");

					for (int i=1;i<=tabs[TableId.Method].NumberOfRows;i++) 
					{
						metaMethod = (MethodRow)tabs[TableId.Method][i];
						
						TypeDefRow nextMetaTypeDef = null;
						if (tabs[TableId.Method].NumberOfRows>1)
						{
							metaTypeDef = (TypeDefRow)tabs[TableId.TypeDef][typesTableIndex+1];						
							while (tabs[TableId.TypeDef].NumberOfRows >= (typesTableIndex+2)) 
							{
								nextMetaTypeDef = (TypeDefRow)tabs[TableId.TypeDef][typesTableIndex+2];
								if (i==(nextMetaTypeDef.MethodList - metaTypeDef.MethodList + metaTypeDef.MethodList))
								{
									++typesTableIndex;
									metaTypeDef = (TypeDefRow)tabs[TableId.TypeDef][typesTableIndex+1];
								}
								else if (0==(nextMetaTypeDef.MethodList - metaTypeDef.MethodList + metaTypeDef.MethodList))
								{
									++typesTableIndex;
									metaTypeDef = (TypeDefRow)tabs[TableId.TypeDef][typesTableIndex+1];
								}
								else
									break;
							}
							currentRTypeDef = assemblyDef.rModuleDef.GetTypeAt(typesTableIndex-1); //TODO: check if this index is correct
						}

						MethodDefSig mds = new MethodDefSig(blobStream.RawData,metaMethod.Signature);
						string methodName = stringsHeap[metaMethod.Name];

						//Build params array
						RParameter [] parameters = null;
						if (tabs[TableId.Param]!=null)
						{
							if (metaMethod.ParamList != 0 && 
								metaMethod.ParamList <= tabs[TableId.Param].NumberOfRows &&
								mds.ParamCount>0)
							{
								int currentMethodParamsIndex = metaMethod.ParamList;
								int nextMethodParamsIndex = 1;

								if (i<tabs[TableId.Method].NumberOfRows) 
								{
									nextMethodParamsIndex = ((MethodRow)tabs[TableId.Method][i+1]).ParamList;
								}
								else 
								{
									if ( tabs[TableId.Param]!=null)
										nextMethodParamsIndex = tabs[TableId.Param].NumberOfRows + 1;
								}

								if (nextMethodParamsIndex-currentMethodParamsIndex>0)
									parameters = new RParameter[nextMethodParamsIndex-currentMethodParamsIndex];
								else if (nextMethodParamsIndex-currentMethodParamsIndex == 0)
								{
									parameters = new RParameter[1];
									nextMethodParamsIndex += 1;
								}
								int paramsArrayIndex = 0;
								for ( int paramIndex = currentMethodParamsIndex; paramIndex < nextMethodParamsIndex; paramIndex++) 
								{
									ParamRow metaParam = (ParamRow)tabs[TableId.Param][paramIndex];
									int	sequence = metaParam.Sequence;
									if (sequence == 0)
									{
										++currentMethodParamsIndex;
										parameters = new RParameter[parameters.Length-1];
										continue;
									}
									
									if ((metaMethod.Flags & MethodAttributes.Static)!=0)
										--sequence; 
									int flags = (int)metaParam.Flags;
									string paramName = stringsHeap[metaParam.Name];
									SigType st = mds.Params[paramIndex-currentMethodParamsIndex].Type; //TODO: Check if paramIndex-currentMethodParamsIndex is correct
									RType pRType = RType.GetRType(st,assemblyDef);
									RType pRTypeTemp =pRType;
									if (mds.Params[paramsArrayIndex].ByRef)
									{
										pRType = assemblyDef.GetRTypeRef(pRType.NameWithNameSpace+"&");
										if (pRType==null)
										{
											if (pRTypeTemp.DeclaringType!=null)
												pRType = new RTypeRef(pRTypeTemp.NameWithNameSpace+"&","",assemblyDef);	
											else
												pRType = new RTypeRef(pRTypeTemp.Name+"&",pRTypeTemp.NameSpace,assemblyDef);
											((RTypeRef)pRType).SetAssembly(assemblyDef);
											if (pRType!=null)
												assemblyDef.rTypeRefs.Add(pRType);
										}
									}
									if (pRType==null)
										Console.WriteLine("Stop");
									RParameter rp = new RParameter(sequence,flags,paramName,pRType);						
									parameters[paramsArrayIndex] = rp;
									assemblyDef.rParameterDefs.Add(rp);
									paramsArrayIndex++;
								}
							}
							else if (metaMethod.ParamList > tabs[TableId.Param].NumberOfRows &&
								metaMethod.ParamList != 0 &&
								mds.ParamCount == 1)
							{
								parameters = new RParameter[1];
								int	sequence = 1;
								if ((metaMethod.Flags & MethodAttributes.Static)!=0)
									sequence = 0;
								int flags = (int)ParameterAttributes.In;
								string paramName = "A_0";
								SigType st = mds.Params[0].Type; //TODO: Check if paramIndex-currentMethodParamsIndex is correct
								RType pRType = RType.GetRType(st,assemblyDef);
								if (pRType==null)
									Console.WriteLine("Stop");
								RParameter rp = new RParameter(sequence,flags,paramName,pRType);						
								parameters[0] = rp;
								assemblyDef.rParameterDefs.Add(rp);
							}
						} 
						else
						{
							if (mds.ParamCount>0)
							{
								parameters = new RParameter[mds.ParamCount];
								for ( int paramIndex = 0; paramIndex < mds.ParamCount; paramIndex++) 
								{
									int	sequence = paramIndex + 1;
									if ((metaMethod.Flags & MethodAttributes.Static)!=0)
										sequence = paramIndex;
									string paramName = "A_" + sequence;
									SigType st = mds.Params[paramIndex].Type; //TODO: Check if paramIndex-currentMethodParamsIndex is correct
									RType pRType = RType.GetRType(st,assemblyDef);
									int flags = 0;
									if (mds.Params[paramIndex].ByRef)
									{
										RType pRTypeTemp = pRType;
										pRType = assemblyDef.GetType(pRType.NameWithNameSpace+"&");
										if (pRType == null && pRTypeTemp is RTypeDef)
										{
											pRType = assemblyDef.GetRTypeRef(pRType.NameWithNameSpace+"&");
											if (pRType==null)
											{
												if (pRTypeTemp.DeclaringType!=null)
													pRType = new RTypeRef(pRTypeTemp.NameWithNameSpace+"&","",assemblyDef);	
												else
													pRType = new RTypeRef(pRTypeTemp.Name+"&",pRTypeTemp.NameSpace,assemblyDef);
												((RTypeRef)pRType).SetAssembly(assemblyDef);
												if (pRType!=null)
													assemblyDef.rTypeRefs.Add(pRType);
											}
										}
									}
									RParameter rp = new RParameter(sequence,flags,paramName,pRType);						
									parameters[paramIndex] = rp;
									assemblyDef.rParameterDefs.Add(rp);
								}
							}
						}
						//Get the RType of the ReturnType
						SigType rtst = mds.retType.Type;
						RType returnRType = RType.GetRType(rtst,assemblyDef);
						
						//TODO: verify this entry ppoint token
						MDToken EntryPoint = pe.GetEntryPointToken();
						int entryPointMethodIndex = 0;
						entryPointMethodIndex = EntryPoint.RID;
						if (currentRTypeDef!=null && metaTypeDef!=null && i>=metaTypeDef.MethodList)
						{
							if (0 != (metaMethod.Flags & System.Reflection.MethodAttributes.RTSpecialName) 
								&& (methodName == ".ctor" || methodName == ".cctor")) 
							{
					
								RConstructorDef rcd = currentRTypeDef.DefineConstructor(parameters,methodName);
								rcd.Attributes = (MethodAttributes)metaMethod.Flags;
								rcd.MethodImplementationFlags = (MethodImplAttributes)metaMethod.ImplFlags;
								rcd.CallingConventions = (CallingConventions)mds.CallingConvention;
								assemblyDef.rMethodDefs.Add(rcd);
								if ((rcd.Attributes & MethodAttributes.PinvokeImpl) != 0)
									isPInvoke = true;
							}						
							else 
							{
								RMethodDef rmd = currentRTypeDef.DefineMethod(methodName,returnRType,parameters, (MethodAttributes)metaMethod.Flags);							
								//rmd.Attributes = (MethodAttributes)metaMethod.Flags;
								rmd.MethodImplementationFlags = (MethodImplAttributes)metaMethod.ImplFlags;
								rmd.CallingConventions = (CallingConventions)mds.CallingConvention;
								assemblyDef.rMethodDefs.Add(rmd);
								if (i==entryPointMethodIndex)
									assemblyDef.entryPoint = (RMethod)rmd;
								if ((rmd.Attributes & MethodAttributes.PinvokeImpl) != 0)
									isPInvoke = true;
							}	
							if (isPInvoke)
							{
								RMethodDef rm = assemblyDef.rMethodDefs[assemblyDef.rMethodDefs.Count - 1] as RMethodDef;
								ImplMapRow implMapRow = tabs[TableId.ImplMap][ImplMapTableIndex] as ImplMapRow;
								if (i != (implMapRow.MemberForwarded & 0x00FFFFFF))
									throw new InvalidProgramException("Bad index");
								ModuleRefRow mrr = tabs[TableId.ModuleRef][implMapRow.ImportScope] as ModuleRefRow;
								string dllName = stringsHeap[mrr.Name];
								string entryName = stringsHeap[implMapRow.ImportName];
								rm.DefinePInvokeInfo(dllName, entryName, (int)implMapRow.MappingFlags);
								++ImplMapTableIndex;
							}

							isPInvoke = false;
							methodsArrayIndex++;
							assemblyDef.rModuleDef.RTypes[typesTableIndex-1] = currentRTypeDef;
						}
						else
						{
							RMethodBase rmb = assemblyDef.RModuleDef.DefineGlobalMethod(
								methodName,
								new RReturnType(returnRType),
								parameters,
								(MethodAttributes)metaMethod.Flags);
							if (rmb is RConstructorDef)
							{
								((RConstructorDef)rmb).MethodImplementationFlags = (MethodImplAttributes)metaMethod.ImplFlags;
								((RConstructorDef)rmb).CallingConventions = (CallingConventions)mds.CallingConvention;
							}
							else
							{
								((RMethodDef)rmb).MethodImplementationFlags = (MethodImplAttributes)metaMethod.ImplFlags;
								((RMethodDef)rmb).CallingConventions = (CallingConventions)mds.CallingConvention;
							}
							assemblyDef.rMethodDefs.Add(rmb);
						}
					}
					AppDomain.Unload(AppDomainForTypeDefs);			
	
			

					//*************************************//
					//Read Properties and Creates RProperties
					//*************************************//
					ArrayList rPropertyDefs = new ArrayList();
							
					if (tabs[TableId.Property]!=null) 
					{
						int propertyMapIndex = 1;
						PropertyMapRow metaPropertyMap;
						PropertyMapRow nextMetaPropertyMap = null;
						for (int propertyIndex = 1;propertyIndex<=tabs[TableId.Property].NumberOfRows;propertyIndex++) 
						{
							PropertyRow metaProperty = (PropertyRow)tabs[TableId.Property][propertyIndex];
							metaPropertyMap = (PropertyMapRow)tabs[TableId.PropertyMap][propertyMapIndex];
							if (propertyMapIndex + 1 <=tabs[TableId.PropertyMap].NumberOfRows)
								nextMetaPropertyMap = (PropertyMapRow)tabs[TableId.PropertyMap][propertyMapIndex+1];
							if (nextMetaPropertyMap!=null)
								if (propertyIndex == nextMetaPropertyMap.PropertyList) 
								{	
									if (propertyMapIndex<tabs[TableId.PropertyMap].NumberOfRows)
										propertyMapIndex++;
									metaPropertyMap = nextMetaPropertyMap;
								}
							string propertyName = stringsHeap[metaProperty.Name];					
							PropertySig propertySignature = new PropertySig(blobStream.RawData,metaProperty.Type);
							RType [] propertyParams = new RType[propertySignature.ParamCount];
							for (int i = 0; i<propertySignature.ParamCount; i++) 
							{
								SigType st = propertySignature.Params[i].Type;
								propertyParams[i] = RType.GetRType(st,assemblyDef);
							}
							SigType pt = propertySignature.Type;
							RType propertyType = RType.GetRType(pt,assemblyDef);
							RTypeDef rtd = assemblyDef.rModuleDef.GetTypeByToken(metaPropertyMap.Parent);
							RPropertyDef rp = rtd.DefineProperty(propertyName,propertyType);
							rp.Attributes = (PropertyAttributes)metaProperty.Flags;
							rp.SetIndexParameters(propertyParams);
							rPropertyDefs.Add(rp);
						}
						if (tabs[TableId.MethodSemantics]!=null)
							for ( int methSmntsIndex = 1; methSmntsIndex <= tabs[TableId.MethodSemantics].NumberOfRows; methSmntsIndex++) 
							{
								MethodSemanticsRow metaMethodSemantics = (MethodSemanticsRow)tabs[TableId.MethodSemantics][methSmntsIndex];
								if (metaMethodSemantics.Association.Type == TokenType.Event) //Events don't count :)
									continue;
								RMethodDef methodDef = (RMethodDef)assemblyDef.rMethodDefs[metaMethodSemantics.Method-1];

								if ( ((int)metaMethodSemantics.Semantics & 0x0001) != 0)
									((RPropertyDef)rPropertyDefs[metaMethodSemantics.Association.RID-1]).SetMethod = methodDef;
								else if ( ((int)metaMethodSemantics.Semantics & 0x0002) != 0)
									((RPropertyDef)rPropertyDefs[metaMethodSemantics.Association.RID-1]).GetMethod = methodDef;
								else
									throw new AssemblyReadException("Problem looking up property method type");
							}
					}

					//*******************************************************///
					//Read Events and build REvents
					ArrayList rEventDefs = new ArrayList();
					
					
					if (tabs[TableId.Event]!=null) 
					{
						int eventMapIndex = 1;
						EventMapRow metaEventMap;
						EventMapRow nextMetaEventMap = null;
						for (int eventIndex = 1;eventIndex<=tabs[TableId.Event].NumberOfRows;eventIndex++) 
						{
							EventRow metaEvent = (EventRow)tabs[TableId.Event][eventIndex];
							metaEventMap = (EventMapRow)tabs[TableId.EventMap][eventMapIndex];
							if (eventMapIndex + 1 <=tabs[TableId.EventMap].NumberOfRows)
								nextMetaEventMap = (EventMapRow)tabs[TableId.EventMap][eventMapIndex+1];
							if (nextMetaEventMap!=null)
								if (eventIndex == nextMetaEventMap.EventList) 
								{	
									if (eventMapIndex<tabs[TableId.EventMap].NumberOfRows)
										eventMapIndex++;
									metaEventMap = nextMetaEventMap;
								}
							string eventName = stringsHeap[metaEvent.Name];
							// Depending on the token type, it either references
							// a TypeRef Table entry or a TypeDef Table entry token.
							RType entryType = null;
							bool validType = false;


							if (metaEvent.EventType.Type == TokenType.TypeDef)
							{
								entryType = assemblyDef.rModuleDef.GetTypeByToken(metaEvent.EventType.RID);
								validType = true;
							}
							else if (metaEvent.EventType.Type == TokenType.TypeRef)
							{
								entryType = assemblyDef.GetTypeByToken(metaEvent.EventType.RID);
								validType = true;
							}
							if (!validType)
								throw new NullReferenceException("Not a valid entry type in events creation");
							RTypeDef rtd = assemblyDef.rModuleDef.GetTypeByToken(metaEventMap.Parent);
							REventDef red = rtd.DefineEvent(eventName,entryType);
							if (red == null)
								throw new NullReferenceException("BAD REventDef creation");
							red.Attributes = (EventAttributes)metaEvent.EventFlags;
							rEventDefs.Add(red);
						}
						if (tabs[TableId.MethodSemantics]!=null)
							for ( int methodSemanticsIndex = 1; methodSemanticsIndex <= tabs[TableId.MethodSemantics].NumberOfRows; methodSemanticsIndex++) 
							{
								MethodSemanticsRow metaMethodSemantics = (MethodSemanticsRow)tabs[TableId.MethodSemantics][methodSemanticsIndex];
								if (metaMethodSemantics.Association.Type == TokenType.Property) // Properties don't count
									continue;
								RMethodDef methodDef = (RMethodDef)assemblyDef.rMethodDefs[metaMethodSemantics.Method-1];
								if ( ((int)metaMethodSemantics.Semantics & 0x0020) != 0)
									((REventDef)rEventDefs[metaMethodSemantics.Association.RID-1]).RaiseMethod = methodDef;
								else if ( ((int)metaMethodSemantics.Semantics & 0x0008) != 0)
									((REventDef)rEventDefs[metaMethodSemantics.Association.RID-1]).AddOnMethod = methodDef;
								else if ( ((int)metaMethodSemantics.Semantics & 0x0010) != 0)
									((REventDef)rEventDefs[metaMethodSemantics.Association.RID-1]).RemoveOnMethod = methodDef;
								else
									throw new AssemblyReadException("Problem looking up event method type");
							}
					}

					//****************************************************
					// Read the FieldMarshal info and create MarshalSpec
					//****************************************************

					if (tabs[TableId.FieldMarshal]!=null)
					{
						for (int i = 1; i <= tabs[TableId.FieldMarshal].NumberOfRows; i++)
						{
							FieldMarshalRow fmr = tabs[TableId.FieldMarshal][i] as FieldMarshalRow;
							MarshalSpec ms = new MarshalSpec(blobStream.RawData, fmr.NativeType);
							if (fmr.Parent.Type == TokenType.FieldDef)
							{
								RFieldDef fd = assemblyDef.rFieldDefs[fmr.Parent.RID - 1] as RFieldDef;
								fd.MarshalSpec = ms;
							}
							else if (fmr.Parent.Type == TokenType.ParamDef)
							{
								RParameter rp = assemblyDef.rParameterDefs[fmr.Parent.RID - 1] as RParameter;
								rp.MarshalSpec = ms;
							}
							else
								throw new InvalidProgramException("Unknown token type");
						}
					}

					///*****************************
					///Create MemberRefs
					///*****************************
					if (tabs[TableId.MemberRef]!=null)
						for (int memberRefIndex = 1; memberRefIndex <= tabs[TableId.MemberRef].NumberOfRows;memberRefIndex++)
						{
							MemberRefRow tmr = (MemberRefRow)tabs[TableId.MemberRef][memberRefIndex];
							string memberName = stringsHeap[tmr.Name];
							RTypeRef rtr = null;
							RMethodBase rmb = null;
							switch (tmr.Class.Type)// & 0x00000007) 
							{
								case TokenType.TypeRef: 
									rtr = assemblyDef.GetTypeByToken(tmr.Class.RID);
									break;
									///TODO: see if the next case is ok?
								case TokenType.TypeDef: 
									continue;
									//break;
								case TokenType.MethodDef:
									rmb = (RMethodBase)assemblyDef.rMethodDefs[(tmr.Class.RID - 1)];
									break;
								case TokenType.TypeSpec:
									
									if (tabs[TableId.TypeSpec]!=null)
									{
										if (tabs[TableId.TypeSpec]!=null)
											if (tmr.Class.RID<=tabs[TableId.TypeSpec].NumberOfRows)
											{
												TypeSpecRow tts = (TypeSpecRow)tabs[TableId.TypeSpec][tmr.Class.RID];
												TypeSpecSig tss = new TypeSpecSig(blobStream.RawData,tts.Signature);
												rtr = (RTypeRef)RType.GetRType(tss.Type,assemblyDef);
											}
									}
									break;
								case TokenType.ModuleRef:
								default:
									throw new Exception("Unsupported token type");
							}
							int tempPos = tmr.Signature;
							if ((blobStream.RawData[tempPos] & 0x80) == 0)
								tempPos += 1;
							else if ((blobStream.RawData[tempPos] & 0x40) == 0)
								tempPos += 2;
							else
								tempPos += 4;
							bool testeField = (blobStream.RawData[tempPos] == 0x06 ||
								blobStream.RawData[tempPos] == 0x6);
							if (testeField)
							{
								FieldSig fs = new FieldSig(blobStream.RawData,tmr.Signature);
								RType fieldType = RType.GetRType(fs.Type,assemblyDef);
								CallingConventions cc = (CallingConventions)fs.CallingConvention;
								RFieldRef rfr = null;
								if (rtr!=null)
								{
									rfr = new RFieldRef(memberName,rtr,fieldType,assemblyDef);
									rtr.AddField(rfr);
								}
								else if (rmb!=null)
								{
									rfr = new RFieldRef(memberName,rmb,fieldType,assemblyDef);
									((RTypeRef)rmb.DeclaringType).AddField(rfr);
								}
							}
							else
							{
								MethodRefSig mrs = new MethodRefSig(blobStream.RawData,tmr.Signature);
								if (mrs!=null)
								{
									RType retType = RType.GetRType(mrs.retType.Type,assemblyDef);
									RParameter [] mrParams = new RParameter[mrs.ParamCount];

									for (int mrParamsIndex = 0; mrParamsIndex<mrs.ParamCount;mrParamsIndex++)
									{
										Param inParam = mrs.Params[mrParamsIndex];						
										RType pType = RType.GetRType(inParam.Type,assemblyDef);
										if (inParam.ByRef)
											pType = assemblyDef.GetRTypeRef(pType.NameWithNameSpace+"&");
										mrParams[mrParamsIndex] = 
											new RParameter(!mrs.HasThis?mrParamsIndex:mrParamsIndex+1,pType);
									}
									if (rtr!=null) 
									{
										if (memberName == ".ctor" || memberName == ".cctor") 
										{
											RConstructorRef rcr = new RConstructorRef(memberName,retType,rtr,mrParams,assemblyDef);
											if (!mrs.HasThis)
												rcr.Attributes = MethodAttributes.Static;
											rtr.AddConstructor(rcr);
										}
										else
										{
											RMethodRef rmr2 = new RMethodRef(memberName,retType,rtr,mrParams,assemblyDef);
											if (!mrs.HasThis)
												rmr2.SetAttributeStatic();
											rtr.AddMethod(rmr2);
										}
									}
								}
							}
						}

					//***************************************//
					//Read Override info and set the overrides
					//***************************************//
					
					if (tabs[TableId.MethodImpl]!=null &&
						tabs[TableId.MethodImpl].NumberOfRows > 0)
					{
						MetadataResolver mDataRes = new MetadataResolver(pe,assemblyDef);
						for (int p = 1; p<= tabs[TableId.MethodImpl].NumberOfRows;p++)
						{
							MethodImplRow mir = tabs[TableId.MethodImpl][p] as MethodImplRow;
							RMethodBase methodDeclar = 
								mDataRes.ResolveMethod(mir.MethodDeclaration.Type,
								mir.MethodDeclaration.RID);
							RMethodBase methodBody = 
								mDataRes.ResolveMethod(mir.MethodBody.Type,
								mir.MethodBody.RID);
							methodBody.AddOverride(methodDeclar);
						}
					}

					//**********************************//
					// Create the methods body

					int methodTableIndex = 1;
					if (tabs[TableId.Method]!=null)
						foreach (RMethodBase r in assemblyDef.rMethodDefs)
						{
							MethodRow tm = (MethodRow)tabs[TableId.Method][methodTableIndex];
							if (tm == null)
								throw new NullReferenceException("BAD TypeMethod reading from metadata");
							if (r.DeclaringType!=null)
							{
								if ((r.Attributes & MethodAttributes.Abstract) == 0 &&
									(r.DeclaringType.Attributes & TypeAttributes.Interface) == 0 &&
									(r.DeclaringType.BaseType!=null &&
									r.DeclaringType.BaseType.NameWithNameSpace != "System.MulticastDelegate") &&
									!r.IsPInvoke)
								{
									if (r is RMethodDef)
									{
										RMethodDef rmd = r as RMethodDef;
									
										rmd.DefineMethodBody(pe.MetaDataRoot.GetMethodBody(methodTableIndex),pe);
									}
									else if (r is RConstructorDef)
									{
										RConstructorDef rcd = r as RConstructorDef;
										rcd.DefineMethodBody(pe.MetaDataRoot.GetMethodBody(methodTableIndex),pe);
									}
									else
										throw new InvalidOperationException("Not the right type of RMethod");
								}
							}
							else
							{
								//Is a Global Method
								if (r is RMethodDef)
								{
									RMethodDef rmd = r as RMethodDef;									
									rmd.DefineMethodBody(pe.MetaDataRoot.GetMethodBody(methodTableIndex),pe);
								}
								else if (r is RConstructorDef)
								{
									RConstructorDef rcd = r as RConstructorDef;
									rcd.DefineMethodBody(pe.MetaDataRoot.GetMethodBody(methodTableIndex),pe);
								}
								else
									throw new InvalidOperationException("Not the right type of RMethod");

							}
							++methodTableIndex;
						}

					//**************************************
					//Set the Custom Attributes
					//**************************************
					

					#region CustomAttributes
					if (tabs[TableId.CustomAttribute]!=null) 
					{
						for( int customAttributeIndex = 1; customAttributeIndex <= tabs[TableId.CustomAttribute].NumberOfRows; customAttributeIndex++) 
						{
							CustomAttributeRow metaCustomAttribute = 
								(CustomAttributeRow)tabs[TableId.CustomAttribute][customAttributeIndex];

							RConstructor ctor = null;
							if (metaCustomAttribute.Type.Type == TokenType.MethodDef) //tag == 0x2 ) 
							{
								RMethodBase methBase = (RMethodBase)assemblyDef.rMethodDefs[metaCustomAttribute.Type.RID-1];
								ctor = methBase as RConstructorDef;
								if (ctor==null)
									throw new InvalidOperationException("Error looking up Ctor");
							} 
							else if (metaCustomAttribute.Type.Type == TokenType.MemberRef)
							{
								
								MemberRefRow metaMemberRef = 
									(MemberRefRow)tabs[TableId.MemberRef][metaCustomAttribute.Type.RID];
								ctor = 
									(RConstructor)MetadataResolver.GetReflectionMethod(pe,blobStream.RawData,assemblyDef,metaMemberRef);
								if (ctor==null)
									throw new InvalidOperationException("Error looking up Ctor");
							}
							else
								if (ctor==null)
								throw new InvalidOperationException("Error looking up Ctor, not supported type");
				
							byte[] custAttrData;
							int offset = metaCustomAttribute.Value;
							int size = 0;
							byte [] temp = blobStream.RawData;
							
							if ((temp[offset] & 0x80) == 0)
							{
								size = temp[offset];
								offset += 1;
							}
							else if ((temp[offset] & 0x40) == 0)
							{
								size = (temp[offset] & ~0x80) << 8 | temp[offset+1];
								offset += 2;
							} 
							else
							{
								size = (temp[offset] & ~0xC0) << 24 | 
									temp[offset + 1] << 16 | 
									temp[offset + 2] << 8 | 
									temp[offset + 3];
								offset += 4;
							}
							int finalPos = offset + size;
							int newPos = 0;
							custAttrData = new byte[size];
							do
							{
								custAttrData[newPos] = temp [offset];
								++offset;
								++newPos;
							}while (offset<finalPos);
							
							MDToken nParent = metaCustomAttribute.Parent; 
							if (customAttributeIndex==223)
								Console.WriteLine("Stop");
							RCustomAttrib rCA = new RCustomAttrib(ctor,custAttrData,assemblyDef);
							switch (nParent.Type) 
							{
								case TokenType.MethodDef:
									RMethodBase rMethodBase = (RMethodBase)assemblyDef.rMethodDefs[nParent.RID-1];
									if (rMethodBase is RConstructor)
									{
										RConstructorDef rCtor = rMethodBase as RConstructorDef;
										rCtor.InternalAddCustomAttributtes(rCA);
										assemblyDef.rMethodDefs[nParent.RID-1] = rCtor;
									} 
									else 
									{
										RMethodDef rMethodDef = rMethodBase as RMethodDef;
										rMethodDef.InternalAddCustomAttributtes(rCA);
										assemblyDef.rMethodDefs[nParent.RID-1] = rMethodDef;
									}
									break;
								case TokenType.FieldDef:
									((RFieldDef)assemblyDef.rFieldDefs[nParent.RID-1]).InternalAddCustomAttributtes(rCA);
									break;
								case TokenType.TypeDef:
									RTypeDef rtd = assemblyDef.rModuleDef.GetTypeByToken(nParent.RID);
									rtd.InternalAddCustomAttributtes(rCA);
									break;
								case TokenType.Assembly:
									assemblyDef.InternalAddCustomAttributtes(rCA);
									break;
								case TokenType.Module:
									assemblyDef.rModuleDef.InternalAddCustomAttributtes(rCA);
									break;
								case TokenType.ParamDef:
									RParameter rp = (RParameter)assemblyDef.rParameterDefs[nParent.RID-1];
									rp.InternalAddCustomAttributtes(rCA);
									break;
								case TokenType.Property:
									RPropertyDef rpp = (RPropertyDef)rPropertyDefs[nParent.RID-1];
									rpp.InternalAddCustomAttributtes(rCA);
									break;
								case TokenType.Event:
									REventDef redd = (REventDef)rEventDefs[nParent.RID-1];
									redd.InternalAddCustomAttributtes(rCA);
									break;
								case TokenType.TypeRef:
								case TokenType.InterfaceImpl:
								case TokenType.MemberRef:
								case TokenType.Permission:
								case TokenType.ModuleRef:
								case TokenType.Signature:
								case TokenType.TypeSpec:
								case TokenType.AssemblyRef:
								case TokenType.File:
								case TokenType.ExportedType:
								case TokenType.ManifestResource:
								default:
									throw new InvalidOperationException("Custom Attribute not supported");
							}
						}
					}

					#endregion

					//*************************************//
					//Copy Manifest Resources
					//*************************************//

				
					if (tabs[TableId.ManifestResource]!=null &&
						tabs[TableId.ManifestResource].NumberOfRows > 0) 
					{
						assemblyDef.resources = new ArrayList();//new RResource[tabs[TableId.ManifestResource].NumberOfRows];
						for (int manifestResourceIndex = tabs[TableId.ManifestResource].NumberOfRows; manifestResourceIndex>=1; manifestResourceIndex--) 
						{
							ManifestResourceRow metaManisfestResource = tabs[TableId.ManifestResource][manifestResourceIndex] as ManifestResourceRow;
							RResource rResource = new RResource(stringsHeap[metaManisfestResource.Name],null,(ResourceAttributes)metaManisfestResource.Flags,null);
							if (metaManisfestResource.Implementation.IsNilToken)
							{
								BinaryReader be = pe.GetResourcesReader();
								be.BaseStream.Position += metaManisfestResource.Offset; 
								int length = be.ReadInt32();
								rResource.Data = be.ReadBytes(length);
							}
							else 
							{
								if (metaManisfestResource.Implementation.Type==TokenType.File)
								{
									if (tabs[TableId.File]!=null)
									{
										FileRow fr = 
											tabs[TableId.File][metaManisfestResource.Implementation.RID] as FileRow;
										if (fr!=null)
											rResource.FileName = stringsHeap[fr.Name];
									}
								}
								else if (metaManisfestResource.Implementation.Type==TokenType.AssemblyRef)
								{
									if (tabs[TableId.AssemblyRef]!=null)
									{
										AssemblyRefRow arr = 
											tabs[TableId.AssemblyRef][metaManisfestResource.Implementation.RID] as AssemblyRefRow;
										if (arr!=null)
											rResource.FileName = stringsHeap[arr.Name];
									}								
								}
								else
									throw new InvalidProgramException("Unknown token type");
							}
							assemblyDef.resources.Add(rResource);
						}
					}

					return assemblyDef;
					
				}
			}
			return null;			
		}

		/// <summary>
		/// Loads an existing assembly from disk.
		/// </summary>
		/// 
		/// TODO: Overload this method so it can accept generic BinaryReaders
		/// 
		/// <param name="fileName">The assembly to be loaded</param>
		/// <returns>A new object representing the assembly</returns>
		public static RAssemblyDef LoadAssembly(String fileName) 
		{
			return LoadAssembly( fileName, false );
		}


		/// <summary>
		/// Saves this dynamic assembly to disk.
		/// </summary>
		/// <param name="assemblyFileName">The file name of the assembly,</param>
		public void SaveAssembly(String assemblyFileName, string asmPath)
		{
			TypeBuildersHolder.Clear();

			AssemblyBuilder ab = this.ReturnAssembly(assemblyFileName, asmPath);
			ab.Save(assemblyFileName);
		}


		/// <summary>
		/// Resolves re-locations that might occur while instrumenting constructors, methods or properties.
		/// </summary>
		/// <param name="code">The code body.</param>
		/// <exception cref="System.Exception">If re-location could not be done.</exception>
		/// <remarks>
		/// If <see cref="Rail.MSIL.Code"/> is null or 
		/// <see cref="Rail.MSIL.Code.InstructionCount"/> is less than one it will
		/// silently exit.
		/// </remarks>
		private void ResolveRelocations(Code code)
		{
			if (null == code || code.InstructionCount <= 0)
				return;

			for(int i = 0;i < code.InstructionCount;i++)
			{
				Instruction instr = code[i];
				if (instr is ILBranch)
				{
					ILBranch ilBranch = instr as ILBranch;
					if (false == ilBranch.Target.Equals(code[ilBranch.Target.index]))
					{
						int iRelBr = ilBranch.Target.index - ilBranch.index;
						if (false == ilBranch.Target.Equals(code[ilBranch.index + iRelBr]))
							throw new Exception("Could not relocate branch, instruction index: " + i);

						ilBranch.Target = code[ilBranch.index + iRelBr];
					}
				}
			}
		}
		/*foreach (Instruction brInstruction in brMethodDef.MethodBody.Code)
								if (brInstruction is ILBranch)
								{
									ILBranch ilBranch = brInstruction as ILBranch;
									if (!ilBranch.Target.Equals(brMethodDef.MethodBody.Code[ilBranch.Target.index]))
									{
										ilBranch.Target = brMethodDef.MethodBody.Code[ilBranch.Target.index];
										ilBranch.Target.AddTargeter(ilBranch);
									}else if (!ilBranch.Target.IsTargetOf(ilBranch))
										ilBranch.Target.AddTargeter(ilBranch);
								}
								*/

		/// <summary>
		/// Return dynamic assembly
		/// </summary>
		/// <param name="assemblyFileName">The file name of the assembly,</param>
		public AssemblyBuilder ReturnAssembly(String assemblyFileName, string asmPath)
		{
			//TODO: Global methods writing
			
			//Verify if branch target instructions are correctly marked
			if (this.RModuleDef!=null)
				foreach (RTypeDef brTypeDef in this.RModuleDef.GetTypes())
				{
					foreach (RMethodDef brMethodDef in brTypeDef.GetMethods())
						if (brMethodDef.MethodBody != null)
							this.ResolveRelocations(brMethodDef.MethodBody.Code);

					foreach (RConstructorDef brConstructorDef in brTypeDef.GetConstructors())
						if (brConstructorDef.MethodBody != null)
							this.ResolveRelocations(brConstructorDef.MethodBody.Code);

					foreach(RPropertyDef pd in brTypeDef.GetProperties())
					{
						RMethodDef rmdp = (RMethodDef)pd.GetMethod;
						if (null != rmdp && null != rmdp.MethodBody)
							this.ResolveRelocations(rmdp.MethodBody.Code);

						rmdp = (RMethodDef)pd.SetMethod;
						if (null != rmdp && null != rmdp.MethodBody)
							this.ResolveRelocations(rmdp.MethodBody.Code);
					}
				}
			
			RCustomAttrib rca = null;

			// The filename must be the same as the main module. Do not
			// allow overwriting an existing assembly.
			AssemblyBuilder ab = TypeResolver.ResolveRAssembly(this,AssemblyBuilderAccess.RunAndSave, asmPath);


			TypeResolver tr = new TypeResolver(this,ab);
			if (this.CustomAttributeCount>0)
				while (this.MoveToNextCustomAttribute())
				{
					rca = null;
					rca = this.CurrentCustomAttribute;
					if (rca!=null)
						ab.SetCustomAttribute(tr.ResolveRCustomAttrib(rca));
				}
			//TODO: Just one Module limitation should be altered to support several
			RModuleDef rmd = this.rModuleDef;
			if (rmd==null)
				throw new NullReferenceException("There isnt a valid module");
			ModuleBuilder mb = tr.ResolveRModule(rmd,assemblyFileName) as ModuleBuilder;
			if (rmd.CustomAttributeCount>0)
				while (rmd.MoveToNextCustomAttribute())
				{
					rca = null;
					rca = rmd.CurrentCustomAttribute;
					if (rca!=null && rca.GetAttributeConstructorArguments()!=null)
						mb.SetCustomAttribute(tr.ResolveRCustomAttrib(rca));
					else if (rca!=null)
					{
						ConstructorInfo cInfo = tr.ResolveRConstructor(rca.AttributeConstructor) as ConstructorInfo;
						mb.SetCustomAttribute(cInfo,rca.AttributeData);
					}
				}
			RTypeDef [] rtdefs;
			if (rmd.RTypes.Count < 1)
				rtdefs = new RTypeDef[0];
			else
				rtdefs = (RTypeDef [])rmd.GetTypes();
			if (rtdefs==null)
				throw new NullReferenceException("There arent valid types");
			Type [] tbs = new Type[rtdefs.Length];
			for (int i = 0; i< rtdefs.Length;i++)
			{
				Type tempBuilder = tr.ResolveRTypeDef(rtdefs[i]);
				tbs[i] = tempBuilder;
				if (rtdefs[i].CustomAttributeCount>0)
					while (rtdefs[i].MoveToNextCustomAttribute())
					{
						rca = null;
						rca = rtdefs[i].CurrentCustomAttribute;
						if (rca!=null && rca.GetAttributeConstructorArguments()!=null)
							((TypeBuilder)tbs[i]).SetCustomAttribute(tr.ResolveRCustomAttrib(rca));
						else if (rca!=null)
						{
							ConstructorInfo cInfo = tr.ResolveRConstructor(rca.AttributeConstructor) as ConstructorInfo;
							((TypeBuilder)tbs[i]).SetCustomAttribute(cInfo,rca.AttributeData);
						}
					}
			}
			for (int i = 0; i< rtdefs.Length;i++)
			{
				if (rtdefs[i].BaseType!=null && rtdefs[i].BaseType.NameWithNameSpace.Equals("System.Enum"))
					continue;
				RField [] rfs = rtdefs[i].GetFields();
				if (rfs!=null)
				{
					FieldBuilder [] fbs = new FieldBuilder[rfs.Length];
					for (int j=0; j<rfs.Length;j++)
					{
						fbs[j] = tr.ResolveRField(rfs[j]) as FieldBuilder;
						if (rfs[j].CustomAttributeCount>0)
							while (rfs[j].MoveToNextCustomAttribute())
							{
								rca = null;
								rca = rfs[j].CurrentCustomAttribute;
								if (rca!=null && rca.GetAttributeConstructorArguments()!=null)
									fbs[j].SetCustomAttribute(tr.ResolveRCustomAttrib(rca));
								else if (rca!=null)
								{
									ConstructorInfo cInfo = tr.ResolveRConstructor(rca.AttributeConstructor) as ConstructorInfo;
									fbs[j].SetCustomAttribute(cInfo,rca.AttributeData);
								}
							}
					}
				}
			} 
			
			for (int i = 0; i< rtdefs.Length;i++)
			{
				RConstructor [] rcs = rtdefs[i].GetConstructors();
				if (rcs!=null)
				{
					ConstructorBuilder [] cbs = new ConstructorBuilder[rcs.Length];
					for (int j=0; j<rcs.Length;j++)
					{
						cbs[j] = tr.ResolveRConstructor(rcs[j]) as ConstructorBuilder;
						if (rcs[j].CustomAttributeCount>0)
							while (rcs[j].MoveToNextCustomAttribute())
							{
								rca = null;
								rca = rcs[j].CurrentCustomAttribute;
								if (rca!=null && rca.GetAttributeConstructorArguments()!=null)
									cbs[j].SetCustomAttribute(tr.ResolveRCustomAttrib(rca));
								else if (rca!=null)
								{
									ConstructorInfo cInfo = tr.ResolveRConstructor(rca.AttributeConstructor) as ConstructorInfo;
									cbs[j].SetCustomAttribute(cInfo,rca.AttributeData);
								}
							}
					}
				}
			} 
			
			
			for (int i = 0; i< rtdefs.Length;i++)
			{
				RMethod [] rms = rtdefs[i].GetMethods();
				if (rms!=null)
				{
					MethodBuilder [] mbs = new MethodBuilder[rms.Length];
					for (int j=0; j<rms.Length;j++)
					{
						mbs[j] = tr.ResolveRMethod(rms[j]) as MethodBuilder;
						if (rms[j].IsOverrideBody)
						{
							foreach (RMethodBase overMeth in rms[j].OverridedMethod)
							{
								MethodInfo mi = 
									tr.ResolveRMethod((RMethod)overMeth);
								tr.DefineMethodOverride(mi,(MethodInfo)mbs[j]);
							}
						}
						if (rms[j].CustomAttributeCount>0)
							while (rms[j].MoveToNextCustomAttribute())
							{
								rca = null;
								rca = rms[j].CurrentCustomAttribute;
								if (rca!=null && rca.GetAttributeConstructorArguments()!=null)
									mbs[j].SetCustomAttribute(tr.ResolveRCustomAttrib(rca));
								else if (rca!=null)
								{
									ConstructorInfo cInfo = tr.ResolveRConstructor(rca.AttributeConstructor) as ConstructorInfo;
									mbs[j].SetCustomAttribute(cInfo,rca.AttributeData);
								}
							}
					}
				}
			} 

			for (int i = 0; i< rtdefs.Length;i++)
			{
				RProperty [] rps = rtdefs[i].GetProperties();
				if (rps!=null)
				{
					PropertyBuilder [] pbs = new PropertyBuilder[rps.Length];
					for (int j=0; j<rps.Length;j++)
					{
						pbs[j] = tr.ResolveRProperty(rps[j]) as PropertyBuilder;
						if (rps[j].CustomAttributeCount>0)
							while (rps[j].MoveToNextCustomAttribute())
							{
								rca = null;
								rca = rps[j].CurrentCustomAttribute;
								if (rca!=null && rca.GetAttributeConstructorArguments()!=null)
									((PropertyBuilder)pbs[j]).SetCustomAttribute(tr.ResolveRCustomAttrib(rca));
								else if (rca!=null)
								{
									ConstructorInfo cInfo = tr.ResolveRConstructor(rca.AttributeConstructor) as ConstructorInfo;
									((PropertyBuilder)pbs[j]).SetCustomAttribute(cInfo,rca.AttributeData);
								}
							}
					}
				}
			} 

			for (int i = 0; i< rtdefs.Length;i++)
			{
				REvent [] res = rtdefs[i].GetEvents();
				if (res!=null)
				{
					EventBuilder [] ebs = new EventBuilder[res.Length];
					for (int j=0; j<res.Length;j++)
					{
						ebs[j] = tr.ResolveREvent((REventDef)res[j]) as EventBuilder;
						if (res[j].CustomAttributeCount>0)
							while (res[j].MoveToNextCustomAttribute())
							{
								rca = null;
								rca = res[j].CurrentCustomAttribute;
								if (rca!=null && rca.GetAttributeConstructorArguments()!=null)
									((EventBuilder)ebs[j]).SetCustomAttribute(tr.ResolveRCustomAttrib(rca));
								else if (rca!=null)
								{
									ConstructorInfo cInfo = tr.ResolveRConstructor(rca.AttributeConstructor) as ConstructorInfo;
									((EventBuilder)ebs[j]).SetCustomAttribute(cInfo,rca.AttributeData);
								}
							}
					}
				}
			}
			
			if (this.resources!=null)
				for (int i = this.resources.Count-1; i>=0;i--)
				{
					System.Resources.IResourceWriter resourceWriter;
					resourceWriter = 
						((ModuleBuilder)tr.asmREmit.GetModule(tr.asmRail.RModuleDef.Name)).DefineResource(
						((RResource)this.resources[i]).Name,
						((RResource)this.resources[i]).FileName,
						((RResource)this.resources[i]).Attributes);
					resourceWriter.AddResource(((RResource)this.resources[i]).Name, ((RResource)this.resources[i]).Data);
					//resourceWriter.Close();
				}

			for (int i = 0; i< rtdefs.Length;i++)
			{
				RConstructor [] rcs = rtdefs[i].GetConstructors();
				if (rcs!=null)
				{
					for (int j=0;j<rcs.Length;j++)
					{
						//BC_CHANGE: changed from rcs[i] to rcs[j]
						RConstructorDef rcdef = rcs[j] as RConstructorDef;
						rcdef.Write(tr);
					}				
				}
				
				RMethod [] rms = rtdefs[i].GetMethods();
				if (rms!=null)
				{
					for (int j=0;j<rms.Length;j++)
					{
						RMethodDef rmdef = rms[j] as RMethodDef;
						rmdef.Write(tr);
						if (this.entryPoint == rmdef)
							tr.asmREmit.SetEntryPoint((MethodInfo)tr.ResolveRMethod(rmdef));
					}				
				}
			}

			// Hook up the event listening.
			TypeResolveHandler typeResolveHandler = new TypeResolveHandler(tr.asmREmit);
			// Add a listener for the type resolve events.
			AppDomain currentDomain = Thread.GetDomain();
			ResolveEventHandler resolveHandler = new ResolveEventHandler(typeResolveHandler.ResolveEvent);
			currentDomain.TypeResolve += resolveHandler;

			TypeBuildersHolder.typeBuilders = new ArrayList(tbs.Length);
			
			#region SOLO-HOT-FIX: Interface must come first
			ArrayList lstTypesDummy = new ArrayList(tbs.Length);
			foreach(Type taa in tbs)
			{
				if (taa.IsInterface)
					lstTypesDummy.Insert(0,taa);
				else
					lstTypesDummy.Add(taa);
			}
			
			tbs = (Type[])lstTypesDummy.ToArray(typeof(Type));
			#endregion SOLO-HOT-FIX: Interface must come first

			for (int i = 0; i< tbs.Length;i++)
			{
				TypeBuildersHolder.typeBuilders.Add((TypeBuilder)tbs[i]);
			}
			for (int i = 0; i< tbs.Length;i++)
			{
				if (TypeBuildersHolder.IsCreated(i))
					continue;

				TypeBuildersHolder.CurrentType = (TypeBuilder)tbs[i];
				((TypeBuilder)tbs[i]).CreateType();
				TypeBuildersHolder.SetCreated(i);
			}
			return tr.asmREmit;
		}


		// Helper class called when a resolve type event is raised.
		/// <summary>
		/// 
		/// </summary>
		class TypeResolveHandler 
		{
			/// <summary>
			/// 
			/// </summary>
			private Assembly m_assembly;

			/// <summary>
			/// 
			/// </summary>
			/// <param name="asm"></param>
			public TypeResolveHandler(Assembly asm)
			{
				m_assembly = asm;
			}

			/// <summary>
			/// 
			/// </summary>
			/// <param name="sender"></param>
			/// <param name="args"></param>
			/// <returns></returns>
			public Assembly ResolveEvent(Object sender, ResolveEventArgs args)
			{
				// Use args.Name to look up the type name. In this case, you are getting AnEnum.
				try 
				{
					string current = TypeBuildersHolder.CurrentType.FullName;
					string nestedParent = "";
					if (TypeBuildersHolder.typeBuilders!=null)
						for (int i = 0; i < TypeBuildersHolder.typeBuilders.Count; i++)
							if (!TypeBuildersHolder.IsCreated(i))
							{
								if (((TypeBuilder)TypeBuildersHolder.typeBuilders[i]).DeclaringType != null)
								 nestedParent = ((TypeBuilder)TypeBuildersHolder.typeBuilders[i]).DeclaringType.FullName;
								if (nestedParent == current)
								{
									((TypeBuilder)TypeBuildersHolder.typeBuilders[i]).CreateType();
									TypeBuildersHolder.SetCreated(i);
								}
								else if (((TypeBuilder)TypeBuildersHolder.typeBuilders[i]).BaseType!=null &&
									((TypeBuilder)TypeBuildersHolder.typeBuilders[i]).BaseType.FullName.Equals("System.Enum"))
								{	
									((TypeBuilder)TypeBuildersHolder.typeBuilders[i]).CreateType();
									TypeBuildersHolder.SetCreated(i);
								}
								else if (((TypeBuilder)TypeBuildersHolder.typeBuilders[i]).BaseType!=null &&
									((TypeBuilder)TypeBuildersHolder.typeBuilders[i]).BaseType.FullName.Equals("System.ValueType"))
								{	
									((TypeBuilder)TypeBuildersHolder.typeBuilders[i]).CreateType();
									TypeBuildersHolder.SetCreated(i);
								}
								nestedParent = "";
							}
				}
				catch 
				{
					// This is needed to throw away InvalidOperationException.
					// Loader might send the TypeResolve event more than once
					// and the type might be complete already.
				}

				// Complete the type.            
				return m_assembly;
			}
		}


		/// <summary>
		/// 
		/// </summary>
		internal class TypeBuildersHolder
		{
			/// <summary>
			/// 
			/// </summary>
			public static ArrayList typeBuilders;
			/// <summary>
			/// 
			/// </summary>
			private static ArrayList typeChecks;

			/// <summary>
			/// 
			/// </summary>
			public static TypeBuilder CurrentType;


			public static void Clear()
			{
				typeBuilders = null;
				typeChecks = null;
				CurrentType = null;
			}
			
			/// <summary>
			/// 
			/// </summary>
			/// <param name="i"></param>
			/// <returns></returns>
			public static bool IsCreated(int i)
			{
				if (TypeBuildersHolder.typeChecks == null)
				{
					TypeBuildersHolder.typeChecks = new ArrayList(TypeBuildersHolder.typeBuilders.Capacity);
					for (int j = 0; j < TypeBuildersHolder.typeBuilders.Capacity ; j++)
						TypeBuildersHolder.typeChecks.Add(0);
					return false;
				}
				if (TypeBuildersHolder.typeChecks.Count > i )
				{
					if(((int)TypeBuildersHolder.typeChecks[i])==i)
						return true;
				}
//				else
//				{
//					int cnt = TypeBuildersHolder.typeChecks.Count - i;
//
//					for ( int j = 0; j <= cnt; ++j )
//						TypeBuildersHolder.typeChecks.Add(0);
//				}
					
				return false;
			}

			/// <summary>
			/// 
			/// </summary>
			/// <param name="i"></param>
			public static void SetCreated(int i)
			{
				if (TypeBuildersHolder.typeChecks == null)
				{
					TypeBuildersHolder.typeChecks = new ArrayList(TypeBuildersHolder.typeBuilders.Capacity);
					for (int j = 0; j < TypeBuildersHolder.typeBuilders.Capacity ; j++)
						TypeBuildersHolder.typeChecks.Add(0);
				}
				TypeBuildersHolder.typeChecks[i] = i;
			}
		}

	
		/// <summary>
		/// Defines this assembly in the CLR and prepares it to be executed.
		/// </summary>
		/// <returns>The newly generated assembly.</returns>
		public void RunAssembly(string[] arguments)
		{
			AppDomain runDomain = AppDomain.CreateDomain("RunDomain");
			//Assembly [] assms = runDomain.GetAssemblies();
			AssemblyBuilder ab = this.ReturnAssembly(this.RModuleDef.FileName, "..\\");
			ab.Save(this.RModuleDef.FileName);
			//runDomain.CreateInstance("..\\"+this.RModuleDef.FileName,ab.EntryPoint.DeclaringType.Name);
			runDomain.ExecuteAssembly("..\\"+this.RModuleDef.FileName,null, arguments);
			//assms = runDomain.GetAssemblies();
			AppDomain.Unload(runDomain);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="arguments"></param>
		public void RunAndSaveAssembly(object[] arguments)
		{
			AssemblyBuilder ab = this.ReturnAssembly(this.RModuleDef.FileName, "..\\");
			object ibaseObject = Activator.CreateInstance(ab.EntryPoint.DeclaringType);
			ab.Save(this.RModuleDef.FileName);
		}
		#endregion

		#region Custom Attributes
		/// <summary>
		/// 
		/// </summary>
		/// <param name="attribute"></param>
		public void AddCustomAttributtes(RCustomAttrib attribute) 
		{
			InternalAddCustomAttributtes(attribute);
		}

		/// <summary>
		/// TODO: How should the attribute be specified?
		/// </summary>
		/// <param name="attribute"></param>
		public void RemoveCustomAttribute(RCustomAttrib attribute) 
		{
			InternalRemoveCustomAttribute(attribute);
		}

		#endregion

		#region Assembly references 
		/// <summary>
		/// Obtains a list of all the assemblies referenced by this 
		/// assembly.
		/// </summary>
		/// <returns>
		/// An array of the assemblies referenced. It never returns 
		/// null, if no external assemblies are referenced, the array has 
		/// 0 length.
		/// </returns>    
		// Perhaps this should go into the base class. Does it make sense to
		// obtain the references of an assembly that is only a reference itself?
		public RAssemblyName[] GetReferencedAssemblies() 
		{
			RAssemblyName [] ran;
			if (this.rAssemblyRefs==null) return new RAssemblyName[0];
			ran = new RAssemblyName[this.rAssemblyRefs.Count];
		

			for (int i = 0 ;i < this.rAssemblyRefs.Count;i++)
			{
				ran[i] = ((RAssembly)this.rAssemblyRefs[i]).Name;
			}
			return ran;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="assem"></param>
		/// <returns></returns>
		public RAssemblyRef GetReferencedAssembly(Assembly assem)
		{
			if (this.rAssemblyRefs==null)
				return null;
			RAssemblyRef rar = null;
			AssemblyName an = assem.GetName();
			for (int i = 0; i< this.rAssemblyRefs.Count; i++)
			{
				rar = (RAssemblyRef)this.rAssemblyRefs[i];
				if (rar.Name.Name.Equals(an.Name))
					return rar;
				
			}
			return null;
		}

		public RAssemblyRef GetReferencedAssembly(string assem)
		{
			if (this.rAssemblyRefs==null)
				return null;
			RAssemblyRef rar = null;
			for (int i = 0; i< this.rAssemblyRefs.Count; i++)
			{
				rar = (RAssemblyRef)this.rAssemblyRefs[i];
				if (assem.Equals(rar.Name.FullName))
					return rar;
				
			}
			return null;
		}

		/// <summary>
		/// Get the resource list
		/// </summary>
		/// <returns></returns>
		public RResource [] GetResources() 
		{
			if (resources == null) 
			{
				return new RResource[0];
			} 
			else 
			{
				RResource [] temp =  new RResource[this.resources.Count];
				int i = 0;
				foreach (RResource r in this.resources)
				{
					temp[i] = r;
					++i;
				}
				return temp;
			}
		}
		/// <summary>
		/// Adds a new referenced assembly.
		/// </summary>
		/// <param name="assemblyName">The name of the assembly to be referenced</param>
		public void AddReferencedAssembly(RAssemblyName assemblyName) 
		{
			// Check if the assembly is already referenced.
			if (!this.rAssemblyRefs.Contains(assemblyName))
			{
				RAssemblyRef rar = new RAssemblyRef(assemblyName);
				this.rAssemblyRefs.Add(rar);
			}
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="assemblyFullName"></param>
		public void AddReferencedAssembly(string assemblyFullName) 
		{
			// Check if the assembly is already referenced.
			RAssemblyNameRef ranr = new RAssemblyNameRef(assemblyFullName);
			if (!this.rAssemblyRefs.Contains(ranr))
			{
				RAssemblyRef rar = new RAssemblyRef(ranr);
				this.rAssemblyRefs.Add(rar);
			}
		}

		/// <summary>
		/// Removes one assembly reference. If the assembly that is being 
		/// unreferenced is still being used, an exception is thrown.
		/// </summary>
		/// <param name="assemblyName"></param>
		public void RemoveReferencedAssembly(RAssemblyName assemblyName) 
		{
			// Check if removing this assembly reference breaks any 
			// dependency inside this assembly and throw an exception
			foreach (RTypeRef rtr in this.rAssemblyRefs)
			{
				if (rtr.Assembly.Name.Equals(assemblyName))
					throw new InvalidOperationException("Dependency exception");
			}
			if (this.rAssemblyRefs.Contains(assemblyName))
				this.rAssemblyRefs.Remove(assemblyName);
			else
				throw new NullReferenceException("Assembly refrence not found");
		}
		#endregion

		/// <summary>
		/// Visits the assembly.
		/// </summary>
		/// <remarks>
		/// Visits the assembly itself and the modules. The order by which these
		/// elements are visited, depends on what is set in the Visitor.
		/// </remarks>
		/// <param name="v">The visitor to be used.</param>
		public void Accept(Visitor v) 
		{
			v.GetWalker().WalkAssembly(this);
		}    
	}

}
