/*
	The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); 
	you may not use this file except in compliance with the License. You may obtain a copy of the License at 
	http://www.mozilla.org/MPL/ 
	Software distributed under the License is distributed on an "AS IS" basis, 
	WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the specific language 
	governing rights and limitations under the License. 

	The Original Code is RAIL(Runtime Assembly Instrumentation Library) Alpha Version.
	The Initial Developer of the Original Code is University of Coimbra,
	Computer Science Department, Dependable Systems Group. Copyright (C) University of Coimbra. 
	All Rights Reserved.  
*/
using System;
using System.Reflection;
using Rail.Transformation;

namespace Rail.Reflect
{
	/// <summary>
	/// When an event is created, three methods are generated:
	///  - add_EventName
	///  - remove_EventName
	///  - OnEventName
	///  
	/// An event has the following classes associated with it:
	///  /<EventName/>Args derived from System.EventArgs
	///  /<EventName/>Handler derived from System.Delegate (or MulticastDelegate)
	/// </summary>
	public abstract class REvent : RMember
	{
		#region fields
		/// <summary>
		/// 
		/// </summary>
		protected EventAttributes attributes;
		/// <summary>
		/// 
		/// </summary>
		protected RMethod raiseMethod;
		/// <summary>
		/// 
		/// </summary>
		protected RMethod addOnMethod;
		/// <summary>
		/// 
		/// </summary>
		protected RMethod removeOnMethod;
		/// <summary>
		/// 
		/// </summary>
		protected RType eventHandlerType;
		/// <summary>
		/// 
		/// </summary>
		protected RType eventType;
		/// <summary>
		/// 
		/// </summary>
		protected RType eventArgsType;
		#endregion
		
		/// <summary>
		/// 
		/// </summary>
		public virtual EventAttributes Attributes 
		{
			get 
			{
				return this.attributes;
			}
			set 
			{
				this.attributes = value;
			}
		}

		/// <summary>
		/// Gets or sets the method used to raise this event. The naming convention
		/// is On/<EventName/>. If the name is changed, an exception will be raised 
		/// when the assembly is generated.
		/// </summary>
		public virtual RMethod RaiseMethod 
		{
			get 
			{
				return this.raiseMethod;
			}
			set 
			{
				this.raiseMethod = value;
			}
		}

		/// <summary>
		/// Gets or sets the method used to subscribe to this event. The 
		/// naming convention is add_/<EventName/>. If the name is changed, an
		/// exception will be raised when the assembly is generated.
		/// </summary>
		public virtual RMethod AddOnMethod 
		{
			get 
			{
				return this.addOnMethod;
			}
			set 
			{
				this.addOnMethod = value;
			}
		}

		/// <summary>
		/// Gets or sets the method used to unsubscribe to this event. The 
		/// naming convention is remove_/<EventName/> and it must be instance and 
		/// public. If the name is changed, an exception will be raised when 
		/// the assembly is generated.
		/// </summary>
		public virtual RMethod RemoveOnMethod 
		{
			get 
			{
				return this.removeOnMethod;
			}
			set 
			{
				this.removeOnMethod = value;
			}
		}

		/// <summary>
		/// The type of the event handler of this event. This type always derives
		/// from System.Delegate (and more commonly, from System.MulticastDelegate).
		/// </summary>
		public virtual RType EventHandlerType 
		{
			get 
			{
				return this.eventHandlerType;
			}
			set 
			{
				this.eventHandlerType = value;
			}
		}


		/// <summary>
		/// 
		/// </summary>
		public virtual RType EventType 
		{
			get 
			{
				return this.eventType;
			}
			set 
			{
				this.eventType = value;
			}
		}
		/// <summary>
		/// The type of the event args of this event. This type always derives 
		/// from System.EventArgs
		/// </summary>
		public virtual RType EventArgsType 
		{
			get 
			{
				return this.eventArgsType;
			}
			set 
			{
				this.eventArgsType = value;
			}
		}    
	}

	/// <summary>
	/// 
	/// </summary>
	public class REventRef : REvent 
	{
		#region fields
		/// <summary>
		/// 
		/// </summary>
		private RAssemblyDef rad;
		#endregion

		/// <summary>
		/// 
		/// </summary>
		/// <param name="ei"></param>
		/// <param name="rad"></param>
		internal REventRef(EventInfo ei, RAssemblyDef rad) 
		{
			this.rad = rad;
			MethodInfo mi = ei.GetAddMethod();
			if (mi!=null)
                this.addOnMethod = new RMethodRef(mi,rad);
			mi = null;
			mi = ei.GetRaiseMethod();
			if (mi!=null)
				this.raiseMethod = new RMethodRef(mi,rad);
			mi = null;
			mi = ei.GetRemoveMethod();
			if (mi!=null)
				this.removeOnMethod = new RMethodRef(mi,rad);
			this.attributes = ei.Attributes;
			this.declaringType = rad.GetRTypeRef(ei.DeclaringType);
			this.eventHandlerType = rad.GetRTypeRef(ei.EventHandlerType);
			//TODO: this.eventArgsType = ???
			this.eventType = rad.GetRTypeRef(ei.ReflectedType);
			this.memberType = ei.MemberType;
			this.name = ei.Name;
		}
	}

	/// <summary>
	/// 
	/// </summary>
	public class REventDef : REvent, IVisitable, ICustomAttributeOwnerDef 
	{
		/// <summary>
		/// 
		/// </summary>
		internal REventDef() 
		{      
		}

		/// <summary>
		/// 
		/// </summary>
		public override EventAttributes Attributes 
		{
			set 
			{
				this.attributes = value;
			}
		}

		/// <summary>
		/// 
		/// </summary>
		public override String Name 
		{
			set 
			{
				this.name = value;
			}
		}

		/// <summary>
		/// Gets or sets the method used to raise this event. The naming convention
		/// is On/<EventName/>. If the name is changed, an exception will be raised 
		/// when the assembly is generated.
		/// </summary>
		public override RMethod RaiseMethod 
		{
			set 
			{
				this.raiseMethod = value;
			}
		}

		/// <summary>
		/// Gets or sets the method used to subscribe to this event. The 
		/// naming convention is add_/<EventName/>. If the name is changed, an
		/// exception will be raised when the assembly is generated.
		/// </summary>
		public override RMethod AddOnMethod 
		{
			set 
			{
				this.addOnMethod = value;
			}
		}

		/// <summary>
		/// Gets or sets the method used to unsubscribe to this event. The 
		/// naming convention is remove_/<EventName/> and it must be instance and 
		/// public. If the name is changed, an exception will be raised when 
		/// the assembly is generated.
		/// </summary>
		public override RMethod RemoveOnMethod 
		{
			set 
			{
				this.removeOnMethod = value;
			}
		}

		/// <summary>
		/// The type of the event handler of this event. This type always derives
		/// from System.Delegate (and more commonly, from System.MulticastDelegate).
		/// </summary>
		public override RType EventHandlerType 
		{
			set 
			{
				this.eventHandlerType = value;
			}      
		}

		/// <summary>
		/// 
		/// </summary>
		public override RType EventType 
		{
			set 
			{
				this.eventType =  value;
			}      
		}
		/// <summary>
		/// The type of the event args of this event. This type always derives 
		/// from System.EventArgs
		/// </summary>
		public override RType EventArgsType 
		{
			set 
			{
				this.eventArgsType = value;
			}      
		}

		#region Custom Attributes
		/// <summary>
		/// 
		/// </summary>
		/// <param name="attribute"></param>
		public void AddCustomAttributtes(RCustomAttrib attribute) 
		{
			InternalAddCustomAttributtes(attribute);
		}

		/// <summary>
		/// TODO: How should the attribute be specified?
		/// </summary>
		/// <param name="attribute"></param>
		public void RemoveCustomAttribute(RCustomAttrib attribute) 
		{
			InternalRemoveCustomAttribute(attribute);
		}

		#endregion

		/// <summary>
		/// Visits this event.
		/// </summary>
		/// <remarks>
		/// An event has no subelements    
		/// </remarks>
		/// <param name="v">The visitor to be used.</param>
		public void Accept(Visitor v) 
		{      
			v.GetWalker().WalkEvent(this);
		}
	}
}
